/*
 * Click nbfs://nbhost/SystemFileSystem/Templates/Licenses/license-default.txt to change this license
 * Click nbfs://nbhost/SystemFileSystem/Templates/Classes/Class.java to edit this template
 */
package com.tobee.opencv.detector;

import org.opencv.core.*;
import org.opencv.dnn.*;

import java.util.List;
import java.util.ArrayList;

public class MaskRCNNAutoDetection {
    public static void main(String[] args) {
        System.loadLibrary(Core.NATIVE_LIBRARY_NAME);

        // ✅ 1. Mask R-CNN 모델 로드
        String modelWeights = "C:\\DEV\\Experimental\\Vision\\pretrained_model\\mrcnn\\frozen_inference_graph.pb";  // TensorFlow 모델 사용
        String modelConfig = "C:\\DEV\\Experimental\\Vision\\pretrained_model\\mrcnn\\mask_rcnn_inception_v2_coco_2018_01_28.pbtxt";

        Net net = Dnn.readNetFromTensorflow(modelWeights, modelConfig);

        if (net.empty()) {
            System.out.println("❌ Mask R-CNN 네트워크 로드 실패!");
            return;
        }
        System.out.println("✅ Mask R-CNN 네트워크 로드 성공!");

        // ✅ 2. CPU 강제 사용 (CUDA 오류 방지)
        net.setPreferableBackend(Dnn.DNN_BACKEND_OPENCV);
        net.setPreferableTarget(Dnn.DNN_TARGET_CPU);

        // ✅ 3. 입력 이미지 준비
        Mat image = new Mat(800, 800, CvType.CV_8UC3, new Scalar(0, 0, 0));  // 예제용 빈 이미지
        Mat blob = Dnn.blobFromImage(image, 1.0, new Size(800, 800), new Scalar(0, 0, 0), true, false);

        if (blob.empty()) {
            System.out.println("❌ Blob 생성 실패!");
            return;
        }
        System.out.println("✅ Blob 생성 성공!");

        net.setInput(blob);

        // ✅ 4. 출력 레이어 자동 탐색
        List<String> layerNames = net.getLayerNames();
        MatOfInt outLayers = net.getUnconnectedOutLayers();
        int[] outLayersArray = outLayers.toArray();

        List<String> outputLayerNames = new ArrayList<>();
        for (int i : outLayersArray) {
            outputLayerNames.add(layerNames.get(i - 1)); // OpenCV는 1-based index 사용
        }

        // 🔍 Mask R-CNN은 보통 `detection_out_final`과 `detection_masks`를 사용함
        boolean foundDetectionOutFinal = false;
        for (String layer : outputLayerNames) {
            if (layer.contains("detection_out_final")) {
                foundDetectionOutFinal = true;
                break;
            }
        }

        // ✅ 강제로 `detection_out_final` 추가 (자동 감지 실패 시)
        if (!foundDetectionOutFinal) {
            outputLayerNames.add("detection_out_final");
            System.out.println("⚠️ 자동 감지 실패! detection_out_final을 강제로 추가합니다.");
        }

        // ✅ 감지된 출력 레이어 출력
        System.out.println("📌 감지된 Mask R-CNN 출력 레이어:");
        for (String layer : outputLayerNames) {
            System.out.println(layer);
        }

        // ✅ 5. Forward 실행
        List<Mat> outputs = new ArrayList<>();
        net.forward(outputs, outputLayerNames);

        // ✅ 6. 결과 확인
        for (int i = 0; i < outputs.size(); i++) {
            Mat result = outputs.get(i);
            System.out.println("📌 " + outputLayerNames.get(i) + " 결과 크기: " + result.size());
        }

        System.out.println("✅ Mask R-CNN Forward 실행 완료!");
    }
}


