/*
 * Click nbfs://nbhost/SystemFileSystem/Templates/Licenses/license-default.txt to change this license
 * Click nbfs://nbhost/SystemFileSystem/Templates/Classes/Class.java to edit this template
 */
package com.tobee.opencv.detector;

import com.tobee.opencv.detector.tests.input.DNNOption;
import conf.CocoLabel;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import org.opencv.core.Core;
import org.opencv.core.Mat;
import org.opencv.core.MatOfInt;
import org.opencv.core.Rect;
import org.opencv.core.Scalar;
import org.opencv.core.Size;
import org.opencv.dnn.Dnn;
import org.opencv.dnn.Net;
import org.opencv.imgproc.Imgproc;

/**
 *
 * @author
 */
public abstract class ObjectDetector implements IObjectDetector {
    protected String winName;
    protected String modelName;
    protected String mediaType;
    protected String deviceType;
    protected String mediaFile;
    protected String modelPath; 
    protected String modelConfiguration;
    protected boolean isSaveOuput;
    public boolean isMaskShading;
    protected double threshold;
        
    private final DNNOption dnnOption;
    
    static{
        System.loadLibrary(Core.NATIVE_LIBRARY_NAME);
    }
    
    protected ObjectDetector(final DNNOption dnnOption)
    {
        this.dnnOption = dnnOption;
        setupOption();
        setupLabelSet();
        
    }
    
    private void setupLabelSet()
    {
        if(modelName.equals(MODEL_MASK_RCNN))
        {
            CocoLabel.setupCocoLabelSet(MASK_RCNN_LABEL_FILE);
        }
        else if(modelName.equals(MODEL_YOLO))
        {
            CocoLabel.setupCocoLabelSet(YOLO_LABEL_FILE);
        }
    }
    
    private void mrcnnImageTest()
    {
        modelName = MODEL_MASK_RCNN;
        deviceType=TYPE_CPU;
        mediaType = TYPE_IMAGE;
        
//        modelPath = "D:\\DEV\\WORK\\Vision\\pretrained_model\\mrcnn\\frozen_inference_graph.pb";
//        modelConfiguration = "D:\\DEV\\WORK\\Vision\\pretrained_model\\mrcnn\\mask_rcnn_inception_v2_coco_2018_01_28.pbtxt";
        modelPath = "C:\\DEV\\Experimental\\Vision\\pretrained_model\\mrcnn\\frozen_inference_graph.pb";
        modelConfiguration = "C:\\DEV\\Experimental\\Vision\\pretrained_model\\mrcnn\\mask_rcnn_inception_v2_coco_2018_01_28.pbtxt";

	mediaFile = "C:\\DEV\\Experimental\\Vision\\data\\images\\2516944023_d00345997d_z.jpg";
//        mediaFile = "D:\\DEV\\WORK\\Vision\\data\\images\\2516944023_d00345997d_z.jpg";
        isSaveOuput= false;
        isMaskShading= true;
    }
    
    private void mrcnnVideoTest()
    {
        modelName = MODEL_MASK_RCNN;
        deviceType=TYPE_CPU;
        mediaType = TYPE_VIDEO;
        
//        modelPath = "D:\\DEV\\WORK\\Vision\\pretrained_model\\mrcnn\\frozen_inference_graph.pb";
//        modelConfiguration = "D:\\DEV\\WORK\\Vision\\pretrained_model\\mrcnn\\mask_rcnn_inception_v2_coco_2018_01_28.pbtxt";
        modelPath = "C:\\DEV\\Experimental\\Vision\\pretrained_model\\mrcnn\\frozen_inference_graph.pb";
        modelConfiguration = "C:\\DEV\\Experimental\\Vision\\pretrained_model\\mrcnn\\mask_rcnn_inception_v2_coco_2018_01_28.pbtxt";

//        mediaFile = "D:\\DEV\\WORK\\Vision\\data\\videos\\vtest.avi";
        mediaFile = "C:\\DEV\\Experimental\\Vision\\data\\videos\\vtest.avi";
        mediaFile = "C:\\DEV\\Experimental\\Vision\\data\\videos\\cars.mp4";
        isSaveOuput= false;
        isMaskShading= true;
        threshold = 0.65D;
    }
    
    private void yoloImageTest()
    {
        modelName = MODEL_YOLO;
        deviceType=TYPE_CPU;
        mediaType = TYPE_IMAGE;
        
        modelPath = "D:\\DEV\\WORK\\Vision\\pretrained_model\\yolo\\v4\\yolov4.weights";
        modelConfiguration = "D:\\DEV\\WORK\\Vision\\pretrained_model\\yolo\\v4\\yolov4.cfg";
//        modelPath = "C:\\DEV\\Experimental\\Vision\\pretrained_model\\yolo\\v4\\yolov4.weights";
//        modelConfiguration = "C:\\DEV\\Experimental\\Vision\\pretrained_model\\yolo\\v4\\yolov4.cfg";

//	mediaFile = "C:\\DEV\\Experimental\\Vision\\data\\images\\2516944023_d00345997d_z.jpg";
        mediaFile = "D:\\DEV\\WORK\\Vision\\data\\images\\2516944023_d00345997d_z.jpg";
    }
    
    private void yoloVideoTest()
    {
        modelName = MODEL_YOLO;
        deviceType=TYPE_CPU;
        mediaType = TYPE_VIDEO;
        
        modelPath = "D:\\DEV\\WORK\\Vision\\pretrained_model\\yolo\\v4\\yolov4.weights";
        modelConfiguration = "D:\\DEV\\WORK\\Vision\\pretrained_model\\yolo\\v4\\yolov4.cfg";
//        modelPath = "C:\\DEV\\Experimental\\Vision\\pretrained_model\\yolo\\v4\\yolov4.weights";
//        modelConfiguration = "C:\\DEV\\Experimental\\Vision\\pretrained_model\\yolo\\v4\\yolov4.cfg";

        mediaFile = "D:\\DEV\\WORK\\Vision\\data\\videos\\vtest.avi";
        mediaFile = "C:\\DEV\\Experimental\\Vision\\data\\videos\\vtest.avi";
    }
    
    private void setupOption()
    {
        winName = dnnOption.WinName.replace("${Model}",dnnOption.modelName);
        modelName = dnnOption.modelName;
        mediaType = dnnOption.mediaType;
        deviceType = dnnOption.deviceType;
        mediaFile = dnnOption.mediaFile;
        modelPath = dnnOption.modelPath; 
        modelConfiguration = dnnOption.modelConfiguration;
        isSaveOuput = dnnOption.isSaveOutput;
        isMaskShading = dnnOption.isMaskShading;
        threshold = dnnOption.threshold;
        
        modelName = MODEL_MASK_RCNN;
        deviceType=TYPE_CPU;
        mediaType = TYPE_IMAGE;
        isSaveOuput= false;
        isMaskShading= true;
        
        mrcnnImageTest();
//        mrcnnVideoTest();
//        yoloImageTest();
//        yoloVideoTest();
    }
    
    protected Net readDNNetByModelName()
    {
        Net dnnNet = null;
        
        if(modelName.equals(MODEL_YOLO))
        {
            dnnNet = Dnn.readNetFromDarknet(modelConfiguration, modelPath);
        }
        else if(modelName.equals(MODEL_MASK_RCNN))
        {
            dnnNet = Dnn.readNetFromTensorflow(modelPath, modelConfiguration);
        }
        
        if(dnnNet == null) return null;
        
        switch(deviceType)
        {
        case "cpu" -> { 
            dnnNet.setPreferableTarget(org.opencv.dnn.Dnn.DNN_TARGET_CPU);
            }
        case "gpu" -> { 
            dnnNet.setPreferableBackend(Dnn.DNN_BACKEND_CUDA);
            dnnNet.setPreferableTarget(Dnn.DNN_TARGET_CUDA);
            }
        }
        
        return dnnNet;
    }
    
    protected Map<String, Mat> getMapOutputLayer(final Net net)
    {
        if(modelName.equals(MODEL_YOLO))
        {
            return getYoloOutputLayer(net);
        }
        else if(modelName.equals(MODEL_MASK_RCNN))
        {
            return getMaskRCnnOutputLayer(net);
        }
        
        return null;
    }
    
    protected Map<String, Mat> getMaskRCnnOutputLayer(final Net net) {
        Map<String, Mat> retMap = new HashMap<>();
        
        // 1. Auto detection of the ouput layer.
        List<String> layerNames = net.getLayerNames();
        MatOfInt outLayers = net.getUnconnectedOutLayers();
        int[] outLayersArray = outLayers.toArray();

        List<String> outputLayerNames = new ArrayList<>();
        for (int i : outLayersArray) {
            retMap.put(layerNames.get(i - 1), null);
            outputLayerNames.add(layerNames.get(i - 1)); // OpenCV uses 1-based index 
        }

        // Generally, Both `detection_out_final` and `detection_masks` layers are used by Mask R-CNN
        boolean foundDetectionOutFinal = false;
        for (String layer : outputLayerNames) {
            if (layer.contains("detection_out_final")) {
                foundDetectionOutFinal = true;
                break;
            }
        }

        // Add `detection_out_final` layer.
        if (!foundDetectionOutFinal) {
            outputLayerNames.add("detection_out_final");
            retMap.put("detection_out_final", null);
        }
        
        return retMap;
    }
    
    public static Mat getSafeROI(Mat frame, Rect bbox) {
        int frameHeight = frame.rows();
        int frameWidth = frame.cols();

        // 좌표가 이미지 범위를 벗어나지 않도록 보정
        int top = Math.max(0, Math.min(bbox.y, frameHeight - 1));
        int bottom = Math.max(0, Math.min(bbox.y + bbox.height, frameHeight));
        int left = Math.max(0, Math.min(bbox.x, frameWidth - 1));
        int right = Math.max(0, Math.min(bbox.x + bbox.width, frameWidth));

        // 유효한 ROI인지 추가 체크 (너비/높이가 0보다 커야 함)
        if (top >= bottom || left >= right) {
            System.err.println("⚠️ 잘못된 ROI 영역입니다!");
            return new Mat();  // 빈 Mat 반환
        }

        // 안전한 ROI 추출
        return frame.submat(top, bottom, left, right);
    }
    
    protected Map<String, Mat>  getYoloOutputLayer(final Net net) {
        Map<String, Mat> retMap = new HashMap<>();
        
        // 1. Retrieve all layer names in the Net.
        List<String> layerNames = net.getLayerNames();

        // 2. Get indexes for output layers
        MatOfInt outLayers = net.getUnconnectedOutLayers();
        int[] outLayersArray = outLayers.toArray();

        // 3. Auto detection of the ouput layer and add the index to the map
        for (int i : outLayersArray) {
            retMap.put(layerNames.get(i - 1), null);
        }

        
        return retMap;
    }
    
    /**
     * 
     * @param orgMat
     * @param targetSize
     * @return java.util.Map
     * 
     */
    protected Map<String, Object> resizeMatImage(Mat orgMat, Size targetSize) {
        Map<String, Object> retMap = new HashMap<>();
        Rect padding =  new Rect();
        int targetWidth = (int)targetSize.width;
        int targetHeight = (int)targetSize.height;
        
        //float[] scale = null;
        int originalWidth = orgMat.cols();
        int originalHeight = orgMat.rows();

        // 비율 유지한 채 리사이징할 크기 계산
        float scaleX = (float) targetWidth / originalWidth;
        float scaleY = (float) targetHeight / originalHeight;
        float minScale = Math.min(scaleX, scaleY);

        int newWidth = (int) (originalWidth * minScale);
        int newHeight = (int) (originalHeight * minScale);

        // 패딩 값 계산
        int left = (targetWidth - newWidth) / 2;
        int top = (targetHeight - newHeight) / 2;
        int right = targetWidth - newWidth - left;
        int bottom = targetHeight - newHeight - top;

        // 리사이징 및 패딩 적용
        Mat resizedMat = new Mat();
        Imgproc.resize(orgMat, resizedMat, new Size(newWidth, newHeight));

        Mat paddedMat = new Mat();
        Core.copyMakeBorder(resizedMat, paddedMat, top, bottom, left, right, Core.BORDER_CONSTANT, new Scalar(0, 0, 0));

        // 패딩 정보 저장
        padding.x = left;
        padding.y = top;
        padding.width = right;
        padding.height = bottom;

        // 스케일 정보 저장
        //scale[0] = minScale;
        retMap.put("mat", paddedMat);
        retMap.put("padding", padding);
        retMap.put("scale", minScale);
        return retMap;
    }
    
    static class DetectedBBox
    {
        private final double confidenceValue;
        private final int classId;
        private final int left;
        private final int top;
        private final int right;
        private final int bottom;
        
        DetectedBBox( final double confidence, final int classId, final int left, final int top, final int right, final int bottom)
        {
            this.confidenceValue = confidence;
            this.classId = classId;
            this.left = left;
            this.right = right;
            this.top = top;
            this.bottom = bottom;
        }
        
        public int getClassId() {
            return classId;
        }

        public int getLeft() {
            return left;
        }

        public int getTop() {
            return top;
        }

        public int getRight() {
            return right;
        }

        public int getBottom() {
            return bottom;
        }
        
        public double getConfidence() {
            return confidenceValue;
        }
    }
}
