/*
 * Click nbfs://nbhost/SystemFileSystem/Templates/Licenses/license-default.txt to change this license
 * Click nbfs://nbhost/SystemFileSystem/Templates/Classes/Class.java to edit this template
 */
package com.tobee.opencv.detector;

import org.opencv.core.*;
import org.opencv.dnn.*;

import java.util.List;
import java.util.ArrayList;

public class YOLOAutoDetection {
    public static void main(String[] args) {
        System.loadLibrary(Core.NATIVE_LIBRARY_NAME);

        // ✅ YOLO 모델 로드 (파일 경로 확인!)
        String modelWeights = "C:\\DEV\\Experimental\\Vision\\pretrained_model\\yolo\\v4\\yolov4.weights";
        String modelConfig = "C:\\DEV\\Experimental\\Vision\\pretrained_model\\yolo\\v4\\yolov4.cfg";
        Net net = Dnn.readNetFromDarknet(modelConfig, modelWeights);

        if (net.empty()) {
            System.out.println("❌ YOLO 네트워크를 로드할 수 없습니다!");
            return;
        }
        System.out.println("✅ YOLO 네트워크 로드 성공!");

        // ✅ 입력 이미지 준비
        Mat image = new Mat(416, 416, CvType.CV_8UC3, new Scalar(0, 0, 0)); // 예제용 빈 이미지
        Mat blob = Dnn.blobFromImage(image, 1 / 255.0, new Size(416, 416), new Scalar(0, 0, 0), true, false);

        // 네트워크에 입력 설정
        net.setInput(blob);

        // ✅ 모든 레이어 이름 가져오기
        List<String> layerNames = net.getLayerNames();

        // ✅ 출력 레이어 인덱스 가져오기
        MatOfInt outLayers = net.getUnconnectedOutLayers();
        int[] outLayersArray = outLayers.toArray();

        // ✅ 출력 레이어 자동 탐색
        List<String> outputLayerNames = new ArrayList<>();
        for (int i : outLayersArray) {
            outputLayerNames.add(layerNames.get(i - 1)); // OpenCV는 1-based index 사용
        }

        // ✅ 감지된 출력 레이어 출력
        System.out.println("📌 감지된 YOLO 출력 레이어:");
        for (String layer : outputLayerNames) {
            System.out.println(layer);
        }

        // ✅ 올바른 출력 레이어 자동 선택 (형식 분석)
        List<Mat> outputs = new ArrayList<>();
        net.forward(outputs, outputLayerNames);

        // ✅ 결과 확인
        for (int i = 0; i < outputs.size(); i++) {
            Mat result = outputs.get(i);
            System.out.println("📌 " + outputLayerNames.get(i) + " 결과 크기: " + result.size());
        }

        System.out.println("✅ YOLO Forward 실행 완료!");
    }
}


