package com.tobee.opencv.detector.history;

import com.tobee.opencv.detector.ObjectDetector;
import com.tobee.opencv.detector.tests.CvUtils;
import com.tobee.opencv.detector.tests.input.DNNOption;
import conf.CocoLabel;
import java.io.File;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import javax.swing.JOptionPane;

import org.opencv.core.Core;
import org.opencv.core.CvType;
import org.opencv.core.Mat;
import org.opencv.core.MatOfDouble;
import org.opencv.core.MatOfPoint;
import org.opencv.core.Point;
import org.opencv.core.Rect;
import org.opencv.core.Scalar;
import org.opencv.core.Size;
import org.opencv.dnn.Dnn;
import org.opencv.dnn.Net;
import org.opencv.highgui.HighGui;
import org.opencv.imgcodecs.Imgcodecs;
import org.opencv.imgproc.Imgproc;
import org.opencv.videoio.VideoCapture;
import org.opencv.videoio.VideoWriter;
import org.opencv.videoio.Videoio;

public class MaskRCNNDetector_v1 extends ObjectDetector {
    private static final double maskThreshold = 0.3; // Mask threshold
    
     public MaskRCNNDetector_v1(final DNNOption dnnOption)
     {
         super(dnnOption);
     }
     
    @Override
    public void detectObject()
    {
        if(!new File(mediaFile).exists()) {
            JOptionPane.showMessageDialog(null, String.format("File9[%s] Not Found", mediaFile), "File Error", JOptionPane.WARNING_MESSAGE);
            return;
        }
        if(!new File(modelPath).exists() || !new File(modelConfiguration).exists()) {
            JOptionPane.showMessageDialog(null, String.format("Can not open Model[%s] \n config[%s]", 
                    modelPath,modelConfiguration), "Model Error", JOptionPane.WARNING_MESSAGE);
            return;
        }
        
        // create an object for label information.
        Map<Integer, String> cocoLabels =  CocoLabel.getCocoLabel();
        // generate radnom color in order to draw bounding boxes
        List<Scalar> colors = CocoLabel.getColorsByLabel();
        
        Net dnnNet = readDNNetByModelName();
        Map<String, Mat> outputLayers = getMapOutputLayer(dnnNet);
                
        if (mediaType.equals("Image"))
        {
            detectObjectOnImageTest(winName, dnnNet, outputLayers, mediaFile, threshold, isSaveOuput, isMaskShading);
        }
        else if (mediaType.equals("Video"))
        {
            detectObjectOnVideoTest(winName, dnnNet, outputLayers, mediaFile, threshold, isSaveOuput, isMaskShading);
//            List<String> outNames = new ArrayList<>(2);
//            for(Iterator<String> iter = outputLayers.keySet().iterator(); iter.hasNext();)
//            {
//                outNames.add(iter.next());
//            }
        }
    }
    
    public void detectObjectOnImageTest(
        String winName, Net dnnNet, Map<String, Mat> outputLayers, 
        String mediaFile, double threshold, boolean isSaveOuput, boolean isMaskShading
    ) {
       HighGui.namedWindow(winName, HighGui.WINDOW_NORMAL);
       
        Mat frame = Imgcodecs.imread(mediaFile);
        
        int idx = new File(mediaFile).getName().split("\\.").length - 2;
        
        String outputFile 
            = new File(mediaFile).getParent() + File.separator + "out" +  File.separator + 
                new File(mediaFile).getName().split("\\.")[idx] + "_mrcnn_out.jpg";
        
        // Stop the program if reached end of video
        if (frame.empty()) {
            System.out.println( "can not read!!!");
            System.out.println( "Output file is stored as " + outputFile);
            HighGui.waitKey(3000);
            return;
        }
               
        Mat blob;
	
        // Create a 4D blob from a frame.
        blob = Dnn.blobFromImage(frame, 1.0, BLOB_MASK_RCNN_SIZE, Scalar.all(0), true, false);
        //Sets the input to the network
        dnnNet.setInput(blob);

        for(Iterator<String> iter = outputLayers.keySet().iterator(); iter.hasNext();)
        {
            String key = iter.next();
            Mat dtectMat = dnnNet.forward(key);
            outputLayers.put(key, dtectMat);
        }
       
        // Extract the bounding box and mask for each of the detected objects
        organizeDetectionTestchatGPT(frame, outputLayers, threshold, isMaskShading);

        // Put efficiency information. The function getPerfProfile returns the overall time for inference(t) 
        // and the timings for each of the layers(in layersTimes)
        MatOfDouble layersTimes = new MatOfDouble();
        double freq = Core.getTickFrequency() / 1000;
        double t = dnnNet.getPerfProfile(layersTimes) / freq;
        String label = String.format("Mask-RCNN, Inference time for a frame : %f ms", t);
        Imgproc.putText(frame, label, new Point(0, 15), 
                        Imgproc.FONT_HERSHEY_SIMPLEX, 0.5, new Scalar(0, 0, 0));
        
        HighGui.imshow(winName, frame);
        
        if(isSaveOuput)
        {
            // Write the frame with the detection boxes
            Mat detectedFrame = new Mat();
            frame.convertTo(detectedFrame, CvType.CV_8U);

            System.out.println("-->" + outputFile);
            Imgcodecs.imwrite(outputFile, detectedFrame);
            
            detectedFrame.release();
        }
        
        blob.release();
        
        layersTimes.release();
        //outNames.clear();
        
        HighGui.waitKey(0);
        
        HighGui.destroyAllWindows();
    }
    
    public static void detectObjectOnVideoTest(
       String winName, Net dnnNet, Map<String, Mat> outputLayers, String mediaFile,
       double threshold, boolean isSaveOuput, boolean isMaskShading
    ) {
        HighGui.namedWindow(winName, HighGui.WINDOW_NORMAL);
        
        Mat frame = new Mat();
        VideoCapture capture = new VideoCapture(mediaFile);
        
        if (capture.open(mediaFile)) {
            capture.read(frame);
            HighGui.imshow(winName, frame);
        } else {
            System.out.println("Capture is not opened!");
            return;
        }
        
        VideoWriter vwriter = new VideoWriter();
        Mat detectedFrame = new Mat();
        String outputFile = mediaFile.substring(0, mediaFile.length() - 4) + "_yolov4_out.avi";
        
         //List<String> outNames = new ArrayList<>(2);
        for(Iterator<String> iter = outputLayers.keySet().iterator(); iter.hasNext();)
        {
            String key = iter.next();
            //outNames.add(iter.next());
            Mat dtectMat = dnnNet.forward(key);
            outputLayers.put(key, dtectMat);
        }
        
        while (capture.isOpened())
        {
            if(HighGui.waitKey(100) == 27) //ESC key to exit
            {
                System.out.println( "[ESC]Done processing !!!");
                //System.out.println( "[ESC]Output file is stored as " + outputFile);
                HighGui.waitKey(3000);
                break;
            }
            
            // get frame from the video
            capture.read(frame);
           
            // Stop the program if reached end of video
            if (frame.empty()) {
                System.out.println( "Done processing !!!");
                System.out.println( "Output file is stored as " + outputFile);
                HighGui.waitKey(3000);
                 if(vwriter.isOpened()) {
                    vwriter.release();
                } 
                break;
            }
            // Create a 4D blob from a frame.
            Mat blob_from_image = Dnn.blobFromImage(frame, 1.0, new Size(frame.cols(), frame.rows()), Scalar.all(0), true, false);

            //Sets the input to the network
            dnnNet.setInput(blob_from_image);

            organizeDetectionTestchatGPT(frame, outputLayers, threshold, isMaskShading);

            // Put efficiency information. The function getPerfProfile returns the overall time for inference(t) and the timings for each of the layers(in layersTimes)
            MatOfDouble layersTimes = new MatOfDouble();
            double freq = Core.getTickFrequency() / 1000;
            double t = dnnNet.getPerfProfile(layersTimes) / freq;
            String label = String.format("Mask-RCNN, Inference time for a frame : %f ms", t);
            Imgproc.putText(frame, label, new Point(0, 15), 
                            Imgproc.FONT_HERSHEY_SIMPLEX, 0.5, new Scalar(0, 0, 0));
            
            HighGui.imshow(winName, frame);
            
            if(isSaveOuput)
            {
                double fps = capture.get(Videoio.CAP_PROP_FPS);
                System.out.println("fps: " + fps);
                if(fps == 0D)
                {
                    //fps = 20;
                    continue;
                }
                
                Size size 
                    = new Size(capture.get(Videoio.CAP_PROP_FRAME_WIDTH), capture.get(Videoio.CAP_PROP_FRAME_HEIGHT));

    //            int fourcc_xvid = VideoWriter.fourcc('X', 'V', 'I', 'D');   //# XVID (AVI format)
    //            int fourcc_mp4v = VideoWriter.fourcc('M', 'P', '4', 'V');  // # MP4 format
                int fourcc_mjpg = VideoWriter.fourcc('M', 'J', 'P', 'G');  //# Motion JPEG format
    //            int fourcc_h264 = VideoWriter.fourcc('H', '2', '6', '4'); // # H264 (more advanced compression)

                isSaveOuput = vwriter.open(outputFile, fourcc_mjpg, fps, size, true);

                if(!isSaveOuput)
                {
                    JOptionPane.showMessageDialog(null, 
                        String.format("Can not open the video file[sz:%s, fps:%f]", size.toString(), fps), 
                        "error", JOptionPane.INFORMATION_MESSAGE);
                    break;
                }
                
                frame.convertTo(detectedFrame, CvType.CV_8U);
                if(vwriter.isOpened()) 
                {
                    vwriter.write(detectedFrame);
                }
                else
                {
                    System.out.println("========vwriter is not opened=======");
                }
            }
            
//            if(saveOuput)
//            {
//                // Write the frame with the detection boxes
//                Mat detectedFrame = new Mat();
//                frame.convertTo(detectedFrame, CvType.CV_8U);
//
//                vwriter.write(detectedFrame);
//                
//                detectedFrame.release();
//                detectedFrame = null;
//            }
            layersTimes.release();
//            outNames.clear();
        }
        HighGui.destroyAllWindows();
        
        //CvUtils.showWindow(winName, CvUtils.matToBufferedImage(CvUtils.reSizeCommon(Imgcodecs.imread(outputFile), 2)));       
    }
    
    private static Mat organizeDetectionTest(
        Mat frame, final Map<String, Mat> outMap, double threshold, boolean isMaskShading) 
    {
        Map<Integer,String> cocoLabels = CocoLabel.getCocoLabel();
        List<Scalar> colors = CocoLabel.getColorsByLabel();
        
        Mat outDetections = outMap.get("detection_out_final");
        Mat outMasks = outMap.get("detection_masks");
        
        final int numDetections = outDetections.size(2);
        outDetections = outDetections.reshape(1, (int) outDetections.total() / 7);

        for (int i = 0; i < numDetections; ++i) {
            float score = outDetections.at(Float.class, i, 2).getV();
            //score = (float)getMatVal(outDetections, i, 2);
            
            if (score > threshold) {
                // Extract the bounding box
                int classId = outDetections.at(Float.class, i, 1).getV().intValue();
                int left = (int)(frame.cols() * outDetections.at(Float.class, i, 3).getV());
                int top = (int)(frame.rows() * outDetections.at(Float.class, i, 4).getV());
                int right = (int)(frame.cols() * outDetections.at(Float.class, i, 5).getV());
                int bottom = (int)(frame.rows() * outDetections.at(Float.class, i, 6).getV());

                left = Math.max(0, Math.min(left, frame.cols() - 1));
                top = Math.max(0, Math.min(top, frame.rows() - 1));
                right = Math.max(0, Math.min(right, frame.cols() - 1));
                bottom = Math.max(0, Math.min(bottom, frame.rows() - 1));
                Rect box = new Rect(left, top, right - left + 1, bottom - top + 1);

                // Extract the mask for the object
                // https://stackoverflow.com/questions/50325025/opencv-ptr-in-java
                // Mat objectMask(outMasks.size[2], outMasks.size[3], CV_32F, outMasks.ptr<float>(i, classId));
                //int aa[] = new int[outDetections.channels()];
                //outDetections.get(i, 1, aa);

                //float[] values = new float[outDetections.channels()];
                //outDetections.get(i, 1, values);

                float fVal = (float)CvUtils.getMatVal(outDetections, i, 1);
                
                 // Output size of masks is NxCxHxW where
                // N - number of detected boxes
                // C - number of classes (excluding background)
                // HxW - segmentation shape
                //final int numClasses = outMasks.size(1);
                Mat objectMask = new Mat(outMasks.size(1), outMasks.size(1), CvType.CV_32F, new Scalar(fVal));
                
                if(isMaskShading)
                {
                     // Draw bounding box, colorize and show the mask on the image
                    drawBoxWithMask(frame, classId, score, box, objectMask, cocoLabels, colors);
                }
               else
                {
                    drawBoxSimple(frame, classId, box, cocoLabels, colors);
                }
                
                objectMask.release(); 
            }
        }
		
		
        outDetections.release(); 
        outMasks.release(); 

        return frame;
    }
     public static void detectObjectOnVideo(
       String winName, Net dnnNet, Map<Integer, String> cocoLabels, List<Scalar> colors, 
        String mediaFile, double threshold, boolean isSaveOuput, boolean isMaskShading
    ) {
        HighGui.namedWindow(winName, HighGui.WINDOW_NORMAL);
        
        Mat frame = new Mat();
        VideoCapture capture = new VideoCapture(mediaFile);
        
        if (capture.open(mediaFile)) {
            capture.read(frame);
            HighGui.imshow(winName, frame);
        } else {
            System.out.println("Capture is not opened!");
            return;
        }
        
        VideoWriter vwriter = new VideoWriter();
        Mat detectedFrame = new Mat();
        String outputFile = mediaFile.substring(0, mediaFile.length() - 4) + "_yolov4_out.avi";
        
        while (capture.isOpened())
        {
            if(HighGui.waitKey(100) == 27) //ESC key to exit
            {
                System.out.println( "[ESC]Done processing !!!");
                //System.out.println( "[ESC]Output file is stored as " + outputFile);
                HighGui.waitKey(3000);
                break;
            }
            
            // get frame from the video
            capture.read(frame);
           
            // Stop the program if reached end of video
            if (frame.empty()) {
                System.out.println( "Done processing !!!");
                System.out.println( "Output file is stored as " + outputFile);
                HighGui.waitKey(3000);
                 if(vwriter.isOpened()) {
                    vwriter.release();
                } 
                break;
            }
            // Create a 4D blob from a frame.
            Mat blob_from_image = Dnn.blobFromImage(frame, 1.0, new Size(frame.cols(), frame.rows()), Scalar.all(0), true, false);

            //Sets the input to the network
            dnnNet.setInput(blob_from_image);

            // Runs the forward pass to get output from the output layers
            List<String> outNames = new ArrayList<>(2);
            outNames.add(0, "detection_out_final");
            outNames.add(1, "detection_masks");

            List<Mat> outs = new ArrayList<>();
            dnnNet.forward(outs, outNames);

            // Extract the bounding box and mask for each of the detected objects
            postprocess(frame, outs, threshold, cocoLabels, colors, isMaskShading);

            // Put efficiency information. The function getPerfProfile returns the overall time for inference(t) and the timings for each of the layers(in layersTimes)
            MatOfDouble layersTimes = new MatOfDouble();
            double freq = Core.getTickFrequency() / 1000;
            double t = dnnNet.getPerfProfile(layersTimes) / freq;
            String label = String.format("Mask-RCNN, Inference time for a frame : %f ms", t);
            Imgproc.putText(frame, label, new Point(0, 15), 
                            Imgproc.FONT_HERSHEY_SIMPLEX, 0.5, new Scalar(0, 0, 0));
            
            HighGui.imshow(winName, frame);
            
            if(isSaveOuput)
            {
                double fps = capture.get(Videoio.CAP_PROP_FPS);
                System.out.println("fps: " + fps);
                if(fps == 0D)
                {
                    //fps = 20;
                    continue;
                }
                
                Size size 
                    = new Size(capture.get(Videoio.CAP_PROP_FRAME_WIDTH), capture.get(Videoio.CAP_PROP_FRAME_HEIGHT));

    //            int fourcc_xvid = VideoWriter.fourcc('X', 'V', 'I', 'D');   //# XVID (AVI format)
    //            int fourcc_mp4v = VideoWriter.fourcc('M', 'P', '4', 'V');  // # MP4 format
                int fourcc_mjpg = VideoWriter.fourcc('M', 'J', 'P', 'G');  //# Motion JPEG format
    //            int fourcc_h264 = VideoWriter.fourcc('H', '2', '6', '4'); // # H264 (more advanced compression)

                isSaveOuput = vwriter.open(outputFile, fourcc_mjpg, fps, size, true);

                if(!isSaveOuput)
                {
                    JOptionPane.showMessageDialog(null, 
                        String.format("Can not open the video file[sz:%s, fps:%f]", size.toString(), fps), 
                        "error", JOptionPane.INFORMATION_MESSAGE);
                    break;
                }
                
                frame.convertTo(detectedFrame, CvType.CV_8U);
                if(vwriter.isOpened()) 
                {
                    vwriter.write(detectedFrame);
                }
                else
                {
                    System.out.println("========vwriter is not opened=======");
                }
            }
            
//            if(saveOuput)
//            {
//                // Write the frame with the detection boxes
//                Mat detectedFrame = new Mat();
//                frame.convertTo(detectedFrame, CvType.CV_8U);
//
//                vwriter.write(detectedFrame);
//                
//                detectedFrame.release();
//                detectedFrame = null;
//            }
            layersTimes.release();
            outNames.clear();
        }
        HighGui.destroyAllWindows();
        
        //CvUtils.showWindow(winName, CvUtils.matToBufferedImage(CvUtils.reSizeCommon(Imgcodecs.imread(outputFile), 2)));       
    }
    
    private static Mat organizeDetectionTestchatGPT(
        Mat frame, final Map<String, Mat> outMap, double threshold, boolean isMaskShading) 
    {
        Map<Integer,String> cocoLabels = CocoLabel.getCocoLabel();
        List<Scalar> colors = CocoLabel.getColorsByLabel();
        
        Mat outDetections = outMap.get("detection_out_final");
        Mat outMasks = outMap.get("detection_masks");
        
        final int numDetections = outDetections.size(2);
        outDetections = outDetections.reshape(1, (int) outDetections.total() / 7);

        for (int i = 0; i < numDetections; ++i) {
            float score = outDetections.at(Float.class, i, 2).getV();
            //score = (float)getMatVal(outDetections, i, 2);
            
            if (score > threshold) {
                // Extract the bounding box
                int classId = outDetections.at(Float.class, i, 1).getV().intValue();
                int left = (int)(frame.cols() * outDetections.at(Float.class, i, 3).getV());
                int top = (int)(frame.rows() * outDetections.at(Float.class, i, 4).getV());
                int right = (int)(frame.cols() * outDetections.at(Float.class, i, 5).getV());
                int bottom = (int)(frame.rows() * outDetections.at(Float.class, i, 6).getV());

                left = Math.max(0, Math.min(left, frame.cols() - 1));
                top = Math.max(0, Math.min(top, frame.rows() - 1));
                right = Math.max(0, Math.min(right, frame.cols() - 1));
                bottom = Math.max(0, Math.min(bottom, frame.rows() - 1));
                Rect box = new Rect(left, top, right - left + 1, bottom - top + 1);

                //float fVal = (float)CvUtils.getMatVal(outDetections, i, 1);
                
                 // Output size of masks is NxCxHxW where
                // N - number of detected boxes
                // C - number of classes (excluding background)
                // HxW - segmentation shape
                //(N, C, H, W)
                //N - 검출된 객체 수
                //C - 클래스 개수 (배경 제외)
                //H × W - 마스크 해상도 (보통 28×28)
                int maskHeight = outMasks.size(2);  // 마스크의 높이 (보통 28)
                int maskWidth = outMasks.size(3);   // 마스크의 너비 (보통 28)

                Mat objectMask = new Mat(maskHeight, maskWidth, CvType.CV_32F);
                outMasks.row(i).col(classId).copyTo(objectMask);
                //2️⃣ Bounding Box 크기에 맞게 마스크 리사이징
                //Mask R-CNN의 기본 마스크 크기는 28×28이므로, Bounding Box 크기에 맞게 조정해야 합니다.
                Imgproc.resize(objectMask, objectMask, new Size(box.width, box.height));
                // 마스크를 Binary 형태로 변환
                //Mask R-CNN의 출력은 0~1 사이의 Float 값입니다.
                //이를 **Binary 형태(0 또는 255)**로 변환해야 시각적으로 선명하게 표현됩니다.
                Imgproc.threshold(objectMask, objectMask, 0.5, 255, Imgproc.THRESH_BINARY);
				
                if(isMaskShading)
                {
                     // Draw bounding box, colorize and show the mask on the image
                    drawBoxWithMask(frame, classId, score, box, objectMask, cocoLabels, colors);
                }
               else
                {
                    drawBoxSimple(frame, classId, box, cocoLabels, colors);
                }
                
                objectMask.release(); 
            }
        }
		
		
        outDetections.release(); 
        outMasks.release(); 

        return frame;
    }
    
    // Draw the predicted bounding box, colorize and show the mask on the image
    private static void drawBoxWithMask(
        Mat frame, int classId, double conf, 
        Rect box, Mat objectMask, 
        Map<Integer, String> cocoLabels,
        List<Scalar> colors
    ) {
        Scalar color = colors.get(classId % colors.size());
        // Draw a rectangle displaying the bounding box
        Imgproc.rectangle(frame, new Point(box.x, box.y), new Point(box.x + box.width, box.y + box.height),
                        color, 1);

        // Get the label for the class name and its confidence
        String label = String.format("%.2f", conf);

        if (!cocoLabels.isEmpty()) {
            label = cocoLabels.get(classId) + ":" + label;
        }
		
        //System.out.println("label : " + label);

        // Display the label at the top of the bounding box
        // int baseLine[] = new int[1];
        int baseLine[] = { 0 };
        Size labelSize = Imgproc.getTextSize(label, Imgproc.FONT_HERSHEY_SIMPLEX, 0.5, 1, baseLine);
        box.y = Integer.max(box.y, (int) labelSize.height);

//        Imgproc.rectangle(
//            frame, new Point(box.x, box.y - Math.round(1.5 * labelSize.height)),
//            new Point(box.x + Math.round(1.5 * labelSize.width), box.y + baseLine[0]), color,
//            Imgproc.FILLED);
        Imgproc.rectangle(
            frame, new Point(box.x, box.y - Math.round(1.5 * labelSize.height)),
            new Point(box.x + Math.round(1.5 * labelSize.width), box.y + baseLine[0]), color);
        Imgproc.putText(frame, label, new Point(box.x, box.y), 
                        Imgproc.FONT_HERSHEY_SIMPLEX, 0.5, new Scalar(0, 0, 0),1);
        
        // ~~~~~~~~~~~~~~
        Mat image_roi = new Mat(frame, box); // frame(box)
        
        
        // 0.3 * color
        Scalar s2 = new Scalar(0.3);
        Scalar s3 = color.mul(s2);

        // 0.7 * image_roi
        Mat image_roi2 = new Mat();
        Scalar s = new Scalar(0.7);
        Core.multiply(image_roi, s, image_roi2);

        Mat coloredRoi = new Mat();
        Core.add(image_roi2, s3, coloredRoi);

        // Mat coloredRoi = new Mat(0.3 * color + 0.7 * image_roi);
        coloredRoi.convertTo(coloredRoi, CvType.CV_8UC3);

        // Draw the contours on the image
        List<MatOfPoint> contours = new ArrayList<>();
        // Mat mask = (objectMask > maskThreshold);
        Mat mask = new Mat();
        objectMask.setTo(new Scalar(maskThreshold), mask);
        Mat hierarchy = new Mat();
        mask.convertTo(mask, CvType.CV_8U);
        Imgproc.findContours(mask, contours, hierarchy, 
                        Imgproc.RETR_CCOMP, Imgproc.CHAIN_APPROX_SIMPLE);
        //public static void drawContours(
        //Mat image, List<MatOfPoint> contours, 
        //int contourIdx, Scalar color, int thickness, int lineType, Mat hierarchy, int maxLevel) {

        Imgproc.drawContours(coloredRoi, contours, -1, color, 2, 
                        Imgproc.LINE_8, hierarchy, 100);
        coloredRoi.copyTo(image_roi, mask);

        mask.release(); 
        image_roi2.release(); 
        hierarchy.release(); 
        coloredRoi.release(); 

        if(!contours.isEmpty()) contours.clear();
        
        for(MatOfPoint p: contours)
        {
            if(p != null) p.release();
        }
    }
    
    // Draw the predicted bounding box, colorize and show the mask on the image
    private static void drawBoxSimple(
        Mat frame, int classId, Rect box, 
        Map<Integer, String> cocoLabels,
        List<Scalar> colors
    ) {
        Scalar selectedColor = colors.get(classId);
        Point x_y = new Point(box.x, box.y);
        Point w_h = new Point(box.x + box.width, box.y + box.height);
        //Point text_point = new Point(box.x, box.y - 5);

        Imgproc.rectangle(frame, w_h, x_y, selectedColor, 1);
        String label = cocoLabels.get(classId);
        
        int baseLine[] = { 0 };
        Size labelSize = Imgproc.getTextSize(label, Imgproc.FONT_HERSHEY_SIMPLEX, 0.5, 1, baseLine);
        box.y = Integer.max(box.y, (int) labelSize.height);
        //public static void putText(Mat img, String text, Point org, int fontFace, double fontScale, Scalar color, int thickness)
        Imgproc.putText(frame, label, new Point(box.x, box.y), 
            Imgproc.FONT_HERSHEY_SIMPLEX, 0.75, selectedColor,1);
        
    }
    
     public void detectObjectOnImage(String winName, Net dnnNet, String mediaFile, double threshold, boolean isSaveOuput, boolean isMaskShading) 
    {
       
        Mat frame = Imgcodecs.imread(mediaFile);
        
        int idx = new File(mediaFile).getName().split("\\.").length - 2;
        
        String outputFile 
            = new File(mediaFile).getParent() + File.separator + "out" +  File.separator + 
                new File(mediaFile).getName().split("\\.")[idx] + "_mrcnn_out.jpg";
        
        // Stop the program if reached end of video
        if (frame.empty()) {
            System.out.println( "can not read!!!");
            System.out.println( "Output file is stored as " + outputFile);
            HighGui.waitKey(3000);
            return;
        }
         Map<Integer, String> cocoLabels =  CocoLabel.getCocoLabel();
        // generate radnom color in order to draw bounding boxes
        List<Scalar> colors =CocoLabel.getColorsByLabel();
        
        
        Mat blob;
	
        // Create a 4D blob from a frame.
        blob = Dnn.blobFromImage(frame, 1.0, new Size(frame.cols(), frame.rows()), Scalar.all(0), true, false);
        //Sets the input to the network
        dnnNet.setInput(blob);
        
        // Runs the forward pass to get output from the output layers
        List<String> outNames = new ArrayList<>(2);
        outNames.add(0, "detection_out_final");
        outNames.add(1, "detection_masks");
        
        List<Mat> outs = new ArrayList<>();
        dnnNet.forward(outs, outNames);
        
        System.out.println("aa =======>" + outNames.size() + ":::" + outs.size());
        
        // Extract the bounding box and mask for each of the detected objects
        postprocess(frame, outs, threshold, cocoLabels, colors, isMaskShading);

        // Put efficiency information. The function getPerfProfile returns the overall time for inference(t) 
        // and the timings for each of the layers(in layersTimes)
        MatOfDouble layersTimes = new MatOfDouble();
        double freq = Core.getTickFrequency() / 1000;
        double t = dnnNet.getPerfProfile(layersTimes) / freq;
        String label = String.format("Mask-RCNN, Inference time for a frame : %f ms", t);
        Imgproc.putText(frame, label, new Point(0, 15), 
                        Imgproc.FONT_HERSHEY_SIMPLEX, 0.5, new Scalar(0, 0, 0));
        
        HighGui.imshow(winName, frame);
        
        if(isSaveOuput)
        {
            // Write the frame with the detection boxes
            Mat detectedFrame = new Mat();
            frame.convertTo(detectedFrame, CvType.CV_8U);

            System.out.println("-->" + outputFile);
            Imgcodecs.imwrite(outputFile, detectedFrame);
            
            detectedFrame.release();
        }
        
        
        blob.release();
        
        layersTimes.release();
        outNames.clear();
        
        HighGui.waitKey(0);
        
        HighGui.destroyAllWindows();
    }
    
    // For each frame, extract the bounding box and mask for each detected object
    private static Mat postprocess(
        Mat frame, final List<Mat> outs, double threshold,
        Map<Integer,String> cocoLabels, List<Scalar> colors, boolean isMaskShading) 
    {
        Mat outDetections = outs.get(0);
        Mat outMasks = outs.get(1);

        // Output size of masks is NxCxHxW where
        // N - number of detected boxes
        // C - number of classes (excluding background)
        // HxW - segmentation shape
        final int numDetections = outDetections.size(2);
        //final int numClasses = outMasks.size(1);

        outDetections = outDetections.reshape(1, (int) outDetections.total() / 7);

        for (int i = 0; i < numDetections; ++i) {
            // float score = outDetections.at<float>(i, 2);
            // Atable<Float> s = outDetections.at(Float.class, i,2);
            float score = outDetections.at(Float.class, i, 2).getV();
            //score = (float)getMatVal(outDetections, i, 2);

            if (score > threshold) {
                // Extract the bounding box
                int classId = outDetections.at(Float.class, i, 1).getV().intValue();
                int left = (int)(frame.cols() * outDetections.at(Float.class, i, 3).getV());
                int top = (int)(frame.rows() * outDetections.at(Float.class, i, 4).getV());
                int right = (int)(frame.cols() * outDetections.at(Float.class, i, 5).getV());
                int bottom = (int)(frame.rows() * outDetections.at(Float.class, i, 6).getV());

                left = Math.max(0, Math.min(left, frame.cols() - 1));
                top = Math.max(0, Math.min(top, frame.rows() - 1));
                right = Math.max(0, Math.min(right, frame.cols() - 1));
                bottom = Math.max(0, Math.min(bottom, frame.rows() - 1));
                Rect box = new Rect(left, top, right - left + 1, bottom - top + 1);

                // Extract the mask for the object
                // https://stackoverflow.com/questions/50325025/opencv-ptr-in-java
                // Mat objectMask(outMasks.size[2], outMasks.size[3], CV_32F, outMasks.ptr<float>(i, classId));
                //int aa[] = new int[outDetections.channels()];
                //outDetections.get(i, 1, aa);

                //float[] values = new float[outDetections.channels()];
                //outDetections.get(i, 1, values);

                float fVal = (float)CvUtils.getMatVal(outDetections, i, 1);

                //Mat objectMask = new Mat(new int[] { outMasks.size(2), outMasks.size(3) }, CvType.CV_32F,
                //		new Scalar(values[0]));

                Mat objectMask = new Mat(outMasks.size(1), outMasks.size(1), CvType.CV_32F, new Scalar(fVal));
                
                if(isMaskShading)
                {
                    // Draw bounding box, colorize and show the mask on the image
                    drawBoxWithMask(frame, classId, score, box, objectMask, cocoLabels, colors);
                }
                else
                {
                    drawBoxSimple(frame, classId, box, cocoLabels, colors);
                }
                
                objectMask.release(); 
            }
        }
		
		
        outDetections.release(); 
        outMasks.release(); 

        return frame;
    }
}
