package com.tobee.opencv.detector.history;

import com.tobee.opencv.detector.ObjectDetector;
import com.tobee.opencv.detector.tests.CvUtils;
import com.tobee.opencv.detector.tests.input.DNNOption;
import conf.CocoLabel;
import java.io.File;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import javax.swing.JOptionPane;

import org.opencv.core.Core;
import org.opencv.core.CvType;
import org.opencv.core.Mat;
import org.opencv.core.MatOfDouble;
import org.opencv.core.MatOfPoint;
import org.opencv.core.Point;
import org.opencv.core.Rect;
import org.opencv.core.Scalar;
import org.opencv.core.Size;
import org.opencv.dnn.Dnn;
import org.opencv.dnn.Net;
import org.opencv.highgui.HighGui;
import org.opencv.imgcodecs.Imgcodecs;
import org.opencv.imgproc.Imgproc;
import org.opencv.videoio.VideoCapture;
import org.opencv.videoio.VideoWriter;
import org.opencv.videoio.Videoio;

public class MaskRCNNDetector_v4 extends ObjectDetector {
    //private static final double maskThreshold = 0.3; // Mask threshold
    
     public MaskRCNNDetector_v4(final DNNOption dnnOption)
     {
         super(dnnOption);
     }
     
    @Override
    public void detectObject()
    {
        if(!new File(mediaFile).exists()) {
            JOptionPane.showMessageDialog(null, String.format("File9[%s] Not Found", mediaFile), "File Error", JOptionPane.WARNING_MESSAGE);
            return;
        }
        if(!new File(modelPath).exists() || !new File(modelConfiguration).exists()) {
            JOptionPane.showMessageDialog(null, String.format("Can not open Model[%s] \n config[%s]", 
                    modelPath,modelConfiguration), "Model Error", JOptionPane.WARNING_MESSAGE);
            return;
        }
        
        Net dnnNet = readDNNetByModelName();
        Map<String, Mat> outputLayers = getMapOutputLayer(dnnNet);
                
        if (mediaType.equals("Image"))
        {
            detectObjectOnImage(winName, dnnNet, outputLayers, mediaFile, threshold, isSaveOuput, isMaskShading);
        }
        else if (mediaType.equals("Video"))
        {
            detectObjectOnVideo(winName, dnnNet, outputLayers, mediaFile, threshold, isSaveOuput, isMaskShading);
        }
    }
    public void detectObjectOnImage(
        String winName, Net dnnNet, Map<String, Mat> outputLayers, 
        String mediaFile, double threshold, boolean isSaveOuput, boolean isMaskShading
    ) {
       HighGui.namedWindow(winName, HighGui.WINDOW_NORMAL);
       
        Mat frame = Imgcodecs.imread(mediaFile);
        
        int idx = new File(mediaFile).getName().split("\\.").length - 2;
        
        String outputFile 
            = new File(mediaFile).getParent() + File.separator + "out" +  File.separator + 
                new File(mediaFile).getName().split("\\.")[idx] + "_mrcnn_out.jpg";
        
        // Stop the program if reached end of video
        if (frame.empty()) {
            System.out.println( "can not read!!!");
            System.out.println( "Output file is stored as " + outputFile);
            HighGui.waitKey(3000);
            return;
        }
        
        Map<String, Object> resultMap  = resizeMatImage(frame, BLOB_MASK_RCNN_SIZE);
        
        Mat resizedMat = (Mat)resultMap.get("mat");
        Rect padding = (Rect)resultMap.get("padding");
        float scale = (float)resultMap.get("scale");
        
        Mat blob = Dnn.blobFromImage(resizedMat, 1.0, BLOB_MASK_RCNN_SIZE, Scalar.all(0), true, false);
        dnnNet.setInput(blob);
         blob.release();
        
        for(Iterator<String> iter = outputLayers.keySet().iterator(); iter.hasNext();)
        {
            String key = iter.next();
            Mat dtectMat = dnnNet.forward(key);
            System.out.println(key + "detectionMasks size: " + dtectMat.size() + ", rows: " + dtectMat.rows() + ", cols: " + dtectMat.cols());
            outputLayers.put(key, dtectMat);
        }
        
        Map<Integer,String> cocoLabels = CocoLabel.getCocoLabel();
        List<Scalar> colors = CocoLabel.getColorsByLabel();
        
        Mat detection = outputLayers.get("detection_out_final");
        Mat detectionMasks = outputLayers.get("detection_masks");
        
        float[] data = new float[(int) (detection.total())];
        detection.get(0, 0, data);

        int rows = detection.size(2);
        
        boolean foundObject = false;
        detection = detection.reshape(1, (int) detection.total() / 7);
        
        // 4️⃣ isContinuous 체크 후 clone()
        if (!detection.isContinuous()) {
            detection = detection.clone();
        }

        // 5️⃣ confidence 값 추출
        int cols = detection.cols();
        data = new float[(int) detection.total()];
        detection.get(0, 0, data);  // Mat 데이터를 1D 배열로 변환

        for (int i = 0; i < rows; i++) {
            float confidence = data[i * cols + 2];
   
            if (confidence > 0.5) {  // 🔥 신뢰도 임계값을 0.1로 낮춤
                foundObject = true;
                int classId = (int)data[i * cols + 1] ;

                // ✅ Bounding Box 좌표 변환
                int left   = (int)((data[i * cols + 3] * MASK_RCNN_SIZE - padding.x) / scale);
                int top    = (int)((data[i * cols + 4] * MASK_RCNN_SIZE - padding.y) / scale);
                int right  = (int)((data[i * cols + 5] * MASK_RCNN_SIZE - padding.x) / scale);
                int bottom = (int)((data[i * cols + 6] * MASK_RCNN_SIZE - padding.y) / scale);
                
                Rect box = new Rect(left, top, right - left + 1, bottom - top + 1);
                // ✅ Bounding Box 그리기
                //Imgproc.rectangle(frame, new Point(left, top), new Point(right, bottom), colors.get(classId), 2);
                // 라벨 추가 (예: "Object: 85%")
                //String label = String.format("%s: %.2f%%", cocoLabels.get(classId), confidence * 100);
                String label = String.format("%s: %.2f", cocoLabels.get(classId), confidence * 100);
                
                // Display the label at the top of the bounding box
                int[] baseLine = { 0 };
                Size labelSize = Imgproc.getTextSize(label, Imgproc.FONT_HERSHEY_SIMPLEX, 0.5, 1, baseLine);
                box.y = Math.max(box.y, (int) labelSize.height);

                // Draw background for text
                Imgproc.rectangle(
                    frame, new Point(box.x, box.y - Math.round(1.5 * labelSize.height)),
                    new Point(box.x + Math.round(1.5 * labelSize.width), box.y + baseLine[0]), 
                    new Scalar(255, 255, 255), Imgproc.FILLED);
                
                Imgproc.putText(frame, label, new Point(left, top - 5), Imgproc.FONT_HERSHEY_SIMPLEX, 0.5, colors.get(classId), 2);
                foundObject = true;
                
                // frame이 비어있는지 확인
                if (frame.empty()) {
                    System.out.println("❌ frame이 비어있습니다! 이미지 로드에 실패한 것 같습니다.");
                    continue;
                }

                // frame의 크기가 올바른지 확인
                int width = frame.cols();
                int height = frame.rows();
                if (width <= 0 || height <= 0) {
                    System.out.println("❌ frame 크기가 유효하지 않습니다.");
                    continue;
                }
                
//                Map<Integer, String> cocoLabels = CocoLabel.getCocoLabel();
//                List<Scalar> colors = CocoLabel.getColorsByLabel();
                // 1. rowMat을 2D로 변환하여, 특정 클래스에 대한 마스크 데이터를 가져옵니다.
                Mat rowMat = detectionMasks.row(i).col(classId);
                Scalar classColor = colors.get(classId);
                
                // 2. 마스크 크기를 지정하고, 그레이스케일로 변환하여 객체 마스크를 생성합니다.
                Mat objectMask = new Mat(rowMat.size(2), rowMat.size(3), CvType.CV_32F);
                rowMat.reshape(1, (int) rowMat.total()).copyTo(objectMask.reshape(1, (int) rowMat.size(2)));

                // 3. 마스크를 이진화하여 적용할 수 있도록 임계값을 설정합니다.
                Imgproc.threshold(objectMask, objectMask, 0.5, 1.0, Imgproc.THRESH_BINARY);

                // 4. `objectMask`를 `CV_8UC1`로 변환하여 색상 채우기를 준비합니다.
                objectMask.convertTo(objectMask, CvType.CV_8UC1); // 변환

                // 5. `objectMask`를 그레이스케일로 만들고 색상 적용 (RGB 색상으로 덮을 수 있도록)
                Mat colorMask = new Mat();
                Imgproc.cvtColor(objectMask, colorMask, Imgproc.COLOR_GRAY2BGR);

                // 6. 색상 마스크 적용: green color (Scalar(0, 255, 0))
                //Scalar color = new Scalar(0, 255, 0); // 초록색
                Core.multiply(colorMask, classColor, colorMask); // 색상을 곱하여 마스크에 적용

                // 7. 경계를 부드럽게 처리 (마스크의 경계를 부드럽게 만드는 효과)
                Mat smoothMask = new Mat();
                Imgproc.GaussianBlur(colorMask, smoothMask, new Size(rowMat.size(2), rowMat.size(3)), 0);

                // 8. ROI (bounding box) 추출
                Mat roi = frame.submat(top, bottom, left, right); // Bounding box 영역에 해당하는 부분을 추출

                // 9. colorMask와 roi의 크기가 일치하는지 확인합니다. 일치하지 않으면 resize로 맞춥니다.
                if (roi.size().width != smoothMask.size().width || roi.size().height != smoothMask.size().height) {
                    Imgproc.resize(smoothMask, smoothMask, roi.size());
                }

                // 10. `smoothMask`의 데이터 타입을 `CV_8U`로 변환하여 `roi`와 일치시킵니다.
                smoothMask.convertTo(smoothMask, CvType.CV_8U);

                // 11. BBox 그리기 (원본 이미지에 경계를 추가합니다)
                //Scalar classColor = new Scalar(0, 255, 0); // 초록색
                Imgproc.rectangle(frame, new Point(left, top), new Point(right, bottom), classColor, 2); // BBox

                // 12. 마스크를 원본 이미지에 덮어씌웁니다 (Alpha Blending 방식)
                //Core.addWeighted(roi, 1.0, smoothMask, 0.5, 0, roi); // 반투명 덮어씌우기
                Core.addWeighted(roi, 0.8, smoothMask, 0.5, 0, roi); // 반투명 덮어씌우기
            }
        }
        
        // Put efficiency information. The function getPerfProfile returns the overall time for inference(t) 
        // and the timings for each of the layers(in layersTimes)
        MatOfDouble layersTimes = new MatOfDouble();
        double freq = Core.getTickFrequency() / 1000;
        double t = dnnNet.getPerfProfile(layersTimes) / freq;
        String label = String.format("Mask-RCNN, Inference time for a frame : %f ms", t);
        Imgproc.putText(frame, label, new Point(0, 15), 
                        Imgproc.FONT_HERSHEY_SIMPLEX, 0.5, new Scalar(0, 0, 0));
        
        if (!foundObject) {
            System.out.println("❌ 객체를 찾지 못했습니다.");
        }
        HighGui.imshow(winName, frame);
        if(isSaveOuput)
        {
            // Write the frame with the detection boxes
            Mat detectedFrame = new Mat();
            frame.convertTo(detectedFrame, CvType.CV_8U);

            Imgcodecs.imwrite(outputFile, detectedFrame);
            
            detectedFrame.release();
        }
        //blob.release();
        layersTimes.release();
        
        // ✅ 윈도우가 닫히면 종료 처리
        int key = HighGui.waitKey(0); // 키 입력 대기
        if (key == 27) {  // ESC 키가 눌리면 종료
            System.out.println("윈도우 닫힘 (ESC 눌림)");
        }

        // ✅ 윈도우 닫히기 전 명시적 종료
        HighGui.destroyAllWindows();  // 모든 윈도우 닫기
        System.exit(0);  // 프로그램 종료
    }
    
    public static void detectObjectOnVideo(
       String winName, Net dnnNet, Map<String, Mat> outputLayers, String mediaFile,
       double threshold, boolean isSaveOuput, boolean isMaskShading
    ) {
        HighGui.namedWindow(winName, HighGui.WINDOW_NORMAL);
        
        Mat frame = new Mat();
        VideoCapture capture = new VideoCapture(mediaFile);
        
        if (capture.open(mediaFile)) {
            capture.read(frame);
            HighGui.imshow(winName, frame);
        } else {
            System.out.println("Capture is not opened!");
            return;
        }
        
        VideoWriter vwriter = new VideoWriter();
        Mat detectedFrame = new Mat();
        String outputFile = mediaFile.substring(0, mediaFile.length() - 4) + "_yolov4_out.avi";
        
         //List<String> outNames = new ArrayList<>(2);
        for(Iterator<String> iter = outputLayers.keySet().iterator(); iter.hasNext();)
        {
            String key = iter.next();
            //outNames.add(iter.next());
            Mat dtectMat = dnnNet.forward(key);
            outputLayers.put(key, dtectMat);
        }
        
        while (capture.isOpened())
        {
            if(HighGui.waitKey(100) == 27) //ESC key to exit
            {
                System.out.println( "[ESC]Done processing !!!");
                //System.out.println( "[ESC]Output file is stored as " + outputFile);
                HighGui.waitKey(3000);
                break;
            }
            
            // get frame from the video
            capture.read(frame);
           
            // Stop the program if reached end of video
            if (frame.empty()) {
                System.out.println( "Done processing !!!");
                System.out.println( "Output file is stored as " + outputFile);
                HighGui.waitKey(3000);
                 if(vwriter.isOpened()) {
                    vwriter.release();
                } 
                break;
            }
            // Create a 4D blob from a frame.
            Mat blob_from_image = Dnn.blobFromImage(frame, 1.0, new Size(frame.cols(), frame.rows()), Scalar.all(0), true, false);

            //Sets the input to the network
            dnnNet.setInput(blob_from_image);

            organizeDetectionTestchatGPT(frame, outputLayers, threshold, isMaskShading, 0, 0);

            // Put efficiency information. The function getPerfProfile returns the overall time for inference(t) and the timings for each of the layers(in layersTimes)
            MatOfDouble layersTimes = new MatOfDouble();
            double freq = Core.getTickFrequency() / 1000;
            double t = dnnNet.getPerfProfile(layersTimes) / freq;
            String label = String.format("Mask-RCNN, Inference time for a frame : %f ms", t);
            Imgproc.putText(frame, label, new Point(0, 15), 
                            Imgproc.FONT_HERSHEY_SIMPLEX, 0.5, new Scalar(0, 0, 0));
            
            HighGui.imshow(winName, frame);
            
            if(isSaveOuput)
            {
                double fps = capture.get(Videoio.CAP_PROP_FPS);
                System.out.println("fps: " + fps);
                if(fps == 0D)
                {
                    //fps = 20;
                    continue;
                }
                
                Size size 
                    = new Size(capture.get(Videoio.CAP_PROP_FRAME_WIDTH), capture.get(Videoio.CAP_PROP_FRAME_HEIGHT));

    //            int fourcc_xvid = VideoWriter.fourcc('X', 'V', 'I', 'D');   //# XVID (AVI format)
    //            int fourcc_mp4v = VideoWriter.fourcc('M', 'P', '4', 'V');  // # MP4 format
                int fourcc_mjpg = VideoWriter.fourcc('M', 'J', 'P', 'G');  //# Motion JPEG format
    //            int fourcc_h264 = VideoWriter.fourcc('H', '2', '6', '4'); // # H264 (more advanced compression)

                isSaveOuput = vwriter.open(outputFile, fourcc_mjpg, fps, size, true);

                if(!isSaveOuput)
                {
                    JOptionPane.showMessageDialog(null, 
                        String.format("Can not open the video file[sz:%s, fps:%f]", size.toString(), fps), 
                        "error", JOptionPane.INFORMATION_MESSAGE);
                    break;
                }
                
                frame.convertTo(detectedFrame, CvType.CV_8U);
                if(vwriter.isOpened()) 
                {
                    vwriter.write(detectedFrame);
                }
                else
                {
                    System.out.println("========vwriter is not opened=======");
                }
            }
            
//            if(saveOuput)
//            {
//                // Write the frame with the detection boxes
//                Mat detectedFrame = new Mat();
//                frame.convertTo(detectedFrame, CvType.CV_8U);
//
//                vwriter.write(detectedFrame);
//                
//                detectedFrame.release();
//                detectedFrame = null;
//            }
            layersTimes.release();
//            outNames.clear();
        }
        HighGui.destroyAllWindows();
        
        //CvUtils.showWindow(winName, CvUtils.matToBufferedImage(CvUtils.reSizeCommon(Imgcodecs.imread(outputFile), 2)));       
    }
    
    private static Mat organizeDetectionTestchatGPTNew(
        Mat orgMat, Mat calcMat, final Map<String, Mat> outMap, double threshold, boolean isMaskShading,
        double scaleX, double scaleY
    ) 
    {
        Map<Integer,String> cocoLabels = CocoLabel.getCocoLabel();
        List<Scalar> colors = CocoLabel.getColorsByLabel();
        
        Mat outDetections = outMap.get("detection_out_final");
        Mat outMasks = outMap.get("detection_masks");
        
        final int numDetections = outDetections.size(2);
        outDetections = outDetections.reshape(1, (int) outDetections.total() / 7);

        for (int i = 0; i < numDetections; ++i) {
            float score = outDetections.at(Float.class, i, 2).getV();
            //score = (float)getMatVal(outDetections, i, 2);
            
            if (score > threshold) {
                // Extract the bounding box
                int classId = outDetections.at(Float.class, i, 1).getV().intValue();
                int left = (int)(calcMat.cols() * outDetections.at(Float.class, i, 3).getV());
                int top = (int)(calcMat.rows() * outDetections.at(Float.class, i, 4).getV());
                int right = (int)(calcMat.cols() * outDetections.at(Float.class, i, 5).getV());
                int bottom = (int)(calcMat.rows() * outDetections.at(Float.class, i, 6).getV());
//                System.out.println(String.format("원래: %d/%d/%d/%d", left,top, right,bottom));
//                // 리사이즈된 이미지에 맞게 바운딩 박스 좌표 조정
                left = (int)(left * scaleX);
                top = (int)(top * scaleY);
                right = (int)(right * scaleX);
                bottom = (int)(bottom * scaleY);
 
//                System.out.println(String.format("이동: %d/%d/%d/%d", left,top, right,bottom));
                left = Math.max(0, Math.min(left, calcMat.cols() - 1));
                top = Math.max(0, Math.min(top, calcMat.rows() - 1));
                right = Math.max(0, Math.min(right, calcMat.cols() - 1));
                bottom = Math.max(0, Math.min(bottom, calcMat.rows() - 1));
                //Rect box = new Rect(left, top+80, right - left + 1, bottom - top + 1);
                Rect box = new Rect(left, top, right - left + 1, bottom - top + 1);
                //System.out.println(String.format("rect: %d/%d/%s/%s", left, top, String.valueOf(right - left + 1), String.valueOf(bottom - top + 1)));

                float fVal = (float)CvUtils.getMatVal(outDetections, i, 1);
                Mat objectMask = new Mat(outMasks.size(1), outMasks.size(1), CvType.CV_32F, new Scalar(fVal));

//                // Output size of masks is NxCxHxW where
//                // N - number of detected boxes
//                // C - number of classes (excluding background)
//                // HxW - segmentation shape
//                //(N, C, H, W)
//                //N - 검출된 객체 수
//                //C - 클래스 개수 (배경 제외)
//                //H × W - 마스크 해상도 (보통 28×28)
//                int maskHeight = outMasks.size(2);  // 마스크의 높이 (보통 28)
//                int maskWidth = outMasks.size(3);   // 마스크의 너비 (보통 28)
//                 System.out.println(String.format("%d/%d", maskHeight,maskWidth));
//
//                Mat objectMask = new Mat(maskHeight, maskWidth, CvType.CV_32F);
 //               outMasks.row(i).col(classId).copyTo(objectMask);
//                //2️⃣ Bounding Box 크기에 맞게 마스크 리사이징
//                //Mask R-CNN의 기본 마스크 크기는 28×28이므로, Bounding Box 크기에 맞게 조정해야 합니다.
//                Imgproc.resize(objectMask, objectMask, new Size(box.width, box.height));
//                // 마스크를 Binary 형태로 변환
//                //Mask R-CNN의 출력은 0~1 사이의 Float 값입니다.
//                //이를 **Binary 형태(0 또는 255)**로 변환해야 시각적으로 선명하게 표현됩니다.
 //               Imgproc.threshold(objectMask, objectMask, 0.5, 255, Imgproc.THRESH_BINARY);
				
                if(isMaskShading)
                {
                     // Draw bounding box, colorize and show the mask on the image
                    drawBoxWithMask(orgMat, classId, score, box, objectMask, cocoLabels, colors);
                }
               else
                {
                    drawBoxSimple(orgMat, classId, box, cocoLabels, colors);
                }
                
                objectMask.release(); 
            }
        }
		
		
        outDetections.release(); 
        outMasks.release(); 

        return orgMat;
    }
    
    private static Mat organizeDetectionTestchatGPT(
        Mat frame, final Map<String, Mat> outMap, double threshold, boolean isMaskShading,
        double scaleX, double scaleY
    ) 
    {
        Map<Integer,String> cocoLabels = CocoLabel.getCocoLabel();
        List<Scalar> colors = CocoLabel.getColorsByLabel();
        
        Mat outDetections = outMap.get("detection_out_final");
        Mat outMasks = outMap.get("detection_masks");
        
        final int numDetections = outDetections.size(2);
        outDetections = outDetections.reshape(1, (int) outDetections.total() / 7);

        for (int i = 0; i < numDetections; ++i) {
            float score = outDetections.at(Float.class, i, 2).getV();
            //score = (float)getMatVal(outDetections, i, 2);
            
            if (score > threshold) {
                // Extract the bounding box
                int classId = outDetections.at(Float.class, i, 1).getV().intValue();
                int left = (int)(frame.cols() * outDetections.at(Float.class, i, 3).getV());
                int top = (int)(frame.rows() * outDetections.at(Float.class, i, 4).getV());
                int right = (int)(frame.cols() * outDetections.at(Float.class, i, 5).getV());
                int bottom = (int)(frame.rows() * outDetections.at(Float.class, i, 6).getV());
//                System.out.println(String.format("원래: %d/%d/%d/%d", left,top, right,bottom));
//                // 리사이즈된 이미지에 맞게 바운딩 박스 좌표 조정
                left = (int)(left * scaleX);
                top = (int)(top * scaleY);
                right = (int)(right * scaleX);
                bottom = (int)(bottom * scaleY);
 
//                System.out.println(String.format("이동: %d/%d/%d/%d", left,top, right,bottom));
                left = Math.max(0, Math.min(left, frame.cols() - 1));
                top = Math.max(0, Math.min(top, frame.rows() - 1));
                right = Math.max(0, Math.min(right, frame.cols() - 1));
                bottom = Math.max(0, Math.min(bottom, frame.rows() - 1));
                //Rect box = new Rect(left, top+80, right - left + 1, bottom - top + 1);
                Rect box = new Rect(left, top, right - left + 1, bottom - top + 1);
                //System.out.println(String.format("rect: %d/%d/%s/%s", left, top, String.valueOf(right - left + 1), String.valueOf(bottom - top + 1)));

                float fVal = (float)CvUtils.getMatVal(outDetections, i, 1);
                Mat objectMask = new Mat(outMasks.size(1), outMasks.size(1), CvType.CV_32F, new Scalar(fVal));

//                // Output size of masks is NxCxHxW where
//                // N - number of detected boxes
//                // C - number of classes (excluding background)
//                // HxW - segmentation shape
//                //(N, C, H, W)
//                //N - 검출된 객체 수
//                //C - 클래스 개수 (배경 제외)
//                //H × W - 마스크 해상도 (보통 28×28)
//                int maskHeight = outMasks.size(2);  // 마스크의 높이 (보통 28)
//                int maskWidth = outMasks.size(3);   // 마스크의 너비 (보통 28)
//                 System.out.println(String.format("%d/%d", maskHeight,maskWidth));
//
//                Mat objectMask = new Mat(maskHeight, maskWidth, CvType.CV_32F);
 //               outMasks.row(i).col(classId).copyTo(objectMask);
//                //2️⃣ Bounding Box 크기에 맞게 마스크 리사이징
//                //Mask R-CNN의 기본 마스크 크기는 28×28이므로, Bounding Box 크기에 맞게 조정해야 합니다.
//                Imgproc.resize(objectMask, objectMask, new Size(box.width, box.height));
//                // 마스크를 Binary 형태로 변환
//                //Mask R-CNN의 출력은 0~1 사이의 Float 값입니다.
//                //이를 **Binary 형태(0 또는 255)**로 변환해야 시각적으로 선명하게 표현됩니다.
 //               Imgproc.threshold(objectMask, objectMask, 0.5, 255, Imgproc.THRESH_BINARY);
				
                if(isMaskShading)
                {
                     // Draw bounding box, colorize and show the mask on the image
                    drawBoxWithMask(frame, classId, score, box, objectMask, cocoLabels, colors);
                }
               else
                {
                    drawBoxSimple(frame, classId, box, cocoLabels, colors);
                }
                
                objectMask.release(); 
            }
        }
		
		
        outDetections.release(); 
        outMasks.release(); 

        return frame;
    }
    
    // Draw the predicted bounding box, colorize and show the mask on the image
    private static void drawBoxWithMask(
        Mat frame, int classId, double conf, 
        Rect box, Mat objectMask, 
        Map<Integer, String> cocoLabels,
        List<Scalar> colors
    ) {
        Scalar color = colors.get(classId % colors.size());
        // Draw a rectangle displaying the bounding box
        Imgproc.rectangle(frame, new Point(box.x, box.y), new Point(box.x + box.width, box.y + box.height),
                        color, 1);

        // Get the label for the class name and its confidence
        String label = String.format("%.2f", conf);

        if (!cocoLabels.isEmpty()) {
            label = cocoLabels.get(classId) + ":" + label;
        }
		
        //System.out.println("label : " + label);

        // Display the label at the top of the bounding box
        // int baseLine[] = new int[1];
        int baseLine[] = { 0 };
        Size labelSize = Imgproc.getTextSize(label, Imgproc.FONT_HERSHEY_SIMPLEX, 0.5, 1, baseLine);
        box.y = Integer.max(box.y, (int) labelSize.height);

        Imgproc.rectangle(
            frame, new Point(box.x, box.y - Math.round(1.5 * labelSize.height)),
            new Point(box.x + Math.round(1.5 * labelSize.width), box.y + baseLine[0]), new Scalar(255, 255, 255),
            Imgproc.FILLED);
//        Imgproc.rectangle(
//            frame, new Point(box.x, box.y - Math.round(1.5 * labelSize.height)),
//            new Point(box.x + Math.round(1.5 * labelSize.width), box.y + baseLine[0]), color);
        Imgproc.putText(frame, label, new Point(box.x, box.y), 
                        Imgproc.FONT_HERSHEY_SIMPLEX, 0.5, new Scalar(0, 0, 0),1);
        
        // ~~~~~~~~~~~~~~
        Mat image_roi = new Mat(frame, box); // frame(box)
        
        
        // 0.3 * color
        Scalar s2 = new Scalar(0.3);
        Scalar s3 = color.mul(s2);

        // 0.7 * image_roi
        Mat image_roi2 = new Mat();
        Scalar s = new Scalar(0.7);
        Core.multiply(image_roi, s, image_roi2);

        Mat coloredRoi = new Mat();
        Core.add(image_roi2, s3, coloredRoi);

        // Mat coloredRoi = new Mat(0.3 * color + 0.7 * image_roi);
        coloredRoi.convertTo(coloredRoi, CvType.CV_8UC3);

        // Draw the contours on the image
        List<MatOfPoint> contours = new ArrayList<>();
        // Mat mask = (objectMask > maskThreshold);
        Mat mask = new Mat();
        objectMask.setTo(new Scalar(maskThreshold), mask);
        Mat hierarchy = new Mat();
        mask.convertTo(mask, CvType.CV_8U);
        Imgproc.findContours(mask, contours, hierarchy, 
                        Imgproc.RETR_CCOMP, Imgproc.CHAIN_APPROX_SIMPLE);
        //public static void drawContours(
        //Mat image, List<MatOfPoint> contours, 
        //int contourIdx, Scalar color, int thickness, int lineType, Mat hierarchy, int maxLevel) {

        Imgproc.drawContours(coloredRoi, contours, -1, color, 2, 
                        Imgproc.LINE_8, hierarchy, 100);
        coloredRoi.copyTo(image_roi, mask);

        mask.release(); 
        image_roi2.release(); 
        hierarchy.release(); 
        coloredRoi.release(); 

        if(!contours.isEmpty()) contours.clear();
        
        for(MatOfPoint p: contours)
        {
            if(p != null) p.release();
        }
    }
    
    // Draw the predicted bounding box, colorize and show the mask on the image
    private static void drawBoxSimple(
        Mat frame, int classId, Rect box, 
        Map<Integer, String> cocoLabels,
        List<Scalar> colors
    ) {
        Scalar selectedColor = colors.get(classId);
        Point x_y = new Point(box.x, box.y);
        Point w_h = new Point(box.x + box.width, box.y + box.height);
        //Point text_point = new Point(box.x, box.y - 5);

        Imgproc.rectangle(frame, w_h, x_y, selectedColor, 1);
        String label = cocoLabels.get(classId);
        
        int baseLine[] = { 0 };
        Size labelSize = Imgproc.getTextSize(label, Imgproc.FONT_HERSHEY_SIMPLEX, 0.5, 1, baseLine);
        box.y = Integer.max(box.y, (int) labelSize.height);
        //public static void putText(Mat img, String text, Point org, int fontFace, double fontScale, Scalar color, int thickness)
        Imgproc.putText(frame, label, new Point(box.x, box.y), 
            Imgproc.FONT_HERSHEY_SIMPLEX, 0.75, selectedColor,1);
        
    }
    
     
}
