/*
 * Click nbfs://nbhost/SystemFileSystem/Templates/Licenses/license-default.txt to change this license
 * Click nbfs://nbhost/SystemFileSystem/Templates/Classes/Class.java to edit this template
 */
package com.tobee.opencv.detector.history;

import com.tobee.opencv.detector.IObjectDetector;
import com.tobee.opencv.detector.tests.input.DNNOption;
import conf.CocoLabel;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import org.opencv.core.Core;
import org.opencv.core.Mat;
import org.opencv.core.MatOfInt;
import org.opencv.core.Rect;
import org.opencv.core.Scalar;
import org.opencv.core.Size;
import org.opencv.dnn.Dnn;
import org.opencv.dnn.Net;
import org.opencv.imgproc.Imgproc;

/**
 *
 * @author
 */
public abstract class ObjectDetector_v1 implements IObjectDetector {
    protected String winName;
    protected String modelName;
    protected String mediaType;
    protected String deviceType;
    protected String mediaFile;
    protected String modelPath; 
    protected String modelConfiguration;
    protected boolean isSaveOuput;
    public boolean isMaskShading;
    protected double threshold;
        
    private final DNNOption dnnOption;
    
    static{
        System.loadLibrary(Core.NATIVE_LIBRARY_NAME);
    }
    
    protected ObjectDetector_v1(final DNNOption dnnOption)
    {
        this.dnnOption = dnnOption;
        setupOption();
        setupLabelSet();
        
    }
    
    private void setupLabelSet()
    {
        if(modelName.equals(MODEL_MASK_RCNN))
        {
            CocoLabel.setupCocoLabelSet(MASK_RCNN_LABEL_FILE);
        }
        else if(modelName.equals(MODEL_YOLO))
        {
            CocoLabel.setupCocoLabelSet(YOLO_LABEL_FILE);
        }
    }
    
    private void setupOption()
    {
        winName = dnnOption.WinName.replace("${Model}",dnnOption.modelName);
        modelName = dnnOption.modelName;
        mediaType = dnnOption.mediaType;
        deviceType = dnnOption.deviceType;
        mediaFile = dnnOption.mediaFile;
        modelPath = dnnOption.modelPath; 
        modelConfiguration = dnnOption.modelConfiguration;
        isSaveOuput = dnnOption.isSaveOutput;
        isMaskShading = dnnOption.isMaskShading;
        threshold = dnnOption.threshold;
        
        modelName = MODEL_MASK_RCNN;
        deviceType=TYPE_CPU;
        mediaType = TYPE_IMAGE;
        
//        modelPath = "D:\\DEV\\WORK\\Vision\\pretrained_model\\mrcnn\\frozen_inference_graph.pb";
//        modelConfiguration = "D:\\DEV\\WORK\\Vision\\pretrained_model\\mrcnn\\mask_rcnn_inception_v2_coco_2018_01_28.pbtxt";
        modelPath = "C:\\DEV\\Experimental\\Vision\\pretrained_model\\mrcnn\\frozen_inference_graph.pb";
        modelConfiguration = "C:\\DEV\\Experimental\\Vision\\pretrained_model\\mrcnn\\mask_rcnn_inception_v2_coco_2018_01_28.pbtxt";
        
//        mediaFile = "D:\\DEV\\WORK\\Vision\\data\\images\\2516944023_d00345997d_z.jpg";
//	mediaFile = "C:\\DEV\\Experimental\\Vision\\data\\images\\2516944023_d00345997d_z.jpg";
        isSaveOuput= false;
        isMaskShading= false;
        
        modelName = MODEL_MASK_RCNN;
        deviceType=TYPE_CPU;
        mediaType = TYPE_VIDEO;
        mediaFile = "D:\\DEV\\WORK\\Vision\\data\\videos\\vtest.avi";
        mediaFile = "C:\\DEV\\Experimental\\Vision\\data\\videos\\vtest.avi";
        mediaFile = "C:\\DEV\\Experimental\\Vision\\data\\videos\\cars.mp4";
        isSaveOuput= false;
        isMaskShading= true;
       
//        modelName = MODEL_YOLO;
//        deviceType=TYPE_CPU;
//        mediaType = TYPE_IMAGE;
//        modelPath = "D:\\DEV\\WORK\\Vision\\pretrained_model\\yolo\\v4\\yolov4.weights";
//        modelConfiguration = "D:\\DEV\\WORK\\Vision\\pretrained_model\\yolo\\v4\\yolov4.cfg";
//        modelPath = "C:\\DEV\\Experimental\\Vision\\pretrained_model\\yolo\\v4\\yolov4.weights";
//        modelConfiguration = "C:\\DEV\\Experimental\\Vision\\pretrained_model\\yolo\\v4\\yolov4.cfg";
        
//        mediaType = TYPE_IMAGE;
//        mediaFile = "D:\\DEV\\WORK\\Vision\\data\\images\\2516944023_d00345997d_z.jpg";
//        mediaFile = "C:\\DEV\\Experimental\\Vision\\data\\images\\2516944023_d00345997d_z.jpg";
//        isSaveOuput= false;
//        isMaskShading= true;
//        
//        mediaType = TYPE_VIDEO;
//        deviceType=TYPE_CPU;
//        mediaFile = "C:\\DEV\\Experimental\\Vision\\data\\images\\2516944023_d00345997d_z.jpg";
//        mediaFile = "C:\\DEV\\Experimental\\Vision\\data\\videos\\vtest.avi";
        mediaFile = "C:\\DEV\\Experimental\\Vision\\data\\videos\\cars.mp4";
        isSaveOuput= false;
        isMaskShading= true;
    }
    
    protected Net readDNNetByModelName()
    {
        Net dnnNet = null;
        
        if(modelName.equals(MODEL_YOLO))
        {
            dnnNet = Dnn.readNetFromDarknet(modelConfiguration, modelPath);
        }
        else if(modelName.equals(MODEL_MASK_RCNN))
        {
            dnnNet = Dnn.readNetFromTensorflow(modelPath, modelConfiguration);
        }
        
        if(dnnNet == null) return null;
        
        switch(deviceType)
        {
        case "cpu" -> { 
            dnnNet.setPreferableTarget(org.opencv.dnn.Dnn.DNN_TARGET_CPU);
            }
        case "gpu" -> { 
            dnnNet.setPreferableBackend(Dnn.DNN_BACKEND_CUDA);
            dnnNet.setPreferableTarget(Dnn.DNN_TARGET_CUDA);
            }
        }
        
        return dnnNet;
    }
    
    protected Map<String, Mat> getMapOutputLayer(final Net net)
    {
        if(modelName.equals(MODEL_YOLO))
        {
            return getYoloOutputLayer(net);
        }
        else if(modelName.equals(MODEL_MASK_RCNN))
        {
            return getMaskRCnnOutputLayer(net);
        }
        
        return null;
    }
    
    protected Map<String, Mat> getMaskRCnnOutputLayer(final Net net) {
        Map<String, Mat> retMap = new HashMap<>();
        
        // 1. Auto detection of the ouput layer.
        List<String> layerNames = net.getLayerNames();
        MatOfInt outLayers = net.getUnconnectedOutLayers();
        int[] outLayersArray = outLayers.toArray();

        List<String> outputLayerNames = new ArrayList<>();
        for (int i : outLayersArray) {
            retMap.put(layerNames.get(i - 1), null);
            outputLayerNames.add(layerNames.get(i - 1)); // OpenCV uses 1-based index 
        }

        // Generally, Both `detection_out_final` and `detection_masks` layers are used by Mask R-CNN
        boolean foundDetectionOutFinal = false;
        for (String layer : outputLayerNames) {
            if (layer.contains("detection_out_final")) {
                foundDetectionOutFinal = true;
                break;
            }
        }

        // Add `detection_out_final` layer.
        if (!foundDetectionOutFinal) {
            outputLayerNames.add("detection_out_final");
            retMap.put("detection_out_final", null);
        }
        
        return retMap;
    }
    
    protected Map<String, Mat>  getYoloOutputLayer(final Net net) {
        Map<String, Mat> retMap = new HashMap<>();
        
        // 1. Retrieve all layer names in the Net.
        List<String> layerNames = net.getLayerNames();

        // 2. Get indexes for output layers
        MatOfInt outLayers = net.getUnconnectedOutLayers();
        int[] outLayersArray = outLayers.toArray();

        // 3. Auto detection of the ouput layer and add the index to the map
        for (int i : outLayersArray) {
            retMap.put(layerNames.get(i - 1), null);
        }

        
        return retMap;
    }
    
    /**
     * 
     * @param orgMat
     * @param targetSize
     * @return java.util.Map
     * 
     */
    protected Map<String, Object> resizeMatImage_org(Mat orgMat, Size targetSize) {
        Map<String, Object> retMap = new HashMap<>();
        Rect padding =  new Rect();
        int targetWidth = (int)targetSize.width;
        int targetHeight = (int)targetSize.height;
        
        int originalWidth = orgMat.cols();
        int originalHeight = orgMat.rows();

        // Minimum scale against the original ratio.
        float scaleX = (float) targetWidth / originalWidth;
        float scaleY = (float) targetHeight / originalHeight;
        float minScale = Math.min(scaleX, scaleY);

        int newWidth = (int) (originalWidth * minScale);
        int newHeight = (int) (originalHeight * minScale);

         // Padding value calculation.
        int left = (targetWidth - newWidth) / 2;
        int top = (targetHeight - newHeight) / 2;
        int right = targetWidth - newWidth - left;
        int bottom = targetHeight - newHeight - top;

        // Resizing and padding adjustment.
        Mat resizedMat = new Mat();
        Imgproc.resize(orgMat, resizedMat, new Size(newWidth, newHeight));

        Mat paddedMat = new Mat();
        Core.copyMakeBorder(resizedMat, paddedMat, top, bottom, left, right, Core.BORDER_CONSTANT, new Scalar(0, 0, 0));

        // The padding information.
        padding.x = left;
        padding.y = top;
        padding.width = right;
        padding.height = bottom;

        //scale[0] = minScale;
        retMap.put("mat", paddedMat);
        retMap.put("padding", padding);
        retMap.put("scale", minScale);// The scale information.
        return retMap;
    }
    
    protected Map<String, Object> resizeMatImage(Mat orgMat, Size targetSize) {
        Map<String, Object> retMap = new HashMap<>();

        // 기본 예외 처리
        if (orgMat == null || orgMat.empty() || targetSize == null) {
            throw new IllegalArgumentException("Input Mat or target size is invalid.");
        }

        // 상수 선언
        final Scalar BLACK = new Scalar(0, 0, 0);

        int targetWidth = (int) targetSize.width;
        int targetHeight = (int) targetSize.height;

        int originalWidth = orgMat.cols();
        int originalHeight = orgMat.rows();

        // 조건 체크: 짧은 변이 800보다 작고, 긴 변이 1333보다 크면 리사이즈
        float scaleX = (float) targetWidth / originalWidth;
        float scaleY = (float) targetHeight / originalHeight;
        float minScale = Math.min(scaleX, scaleY);

        // 리사이즈 이전 조건 체크
        if (Math.min(originalWidth, originalHeight) < 800 || Math.max(originalWidth, originalHeight) > 1333) {
            // 원본 비율 유지 & 최소 스케일 계산
            scaleX = (float) targetWidth / originalWidth;
            scaleY = (float) targetHeight / originalHeight;
            minScale = Math.min(scaleX, scaleY);
        }

        // 새로운 크기 계산 (비율 유지)
        int newWidth = Math.round(originalWidth * minScale);
        int newHeight = Math.round(originalHeight * minScale);

        // 패딩 값 계산 (가운데 정렬 기준)
        int left = (targetWidth - newWidth) / 2;
        int top = (targetHeight - newHeight) / 2;
        int right = targetWidth - newWidth - left;
        int bottom = targetHeight - newHeight - top;

        // 패딩이 이미지 크기를 넘어가지 않도록 확인
        if (left < 0 || top < 0 || right < 0 || bottom < 0) {
            throw new IllegalArgumentException("Padding values cannot be negative. Check target size and image dimensions.");
        }

        // 이미지 리사이즈 처리
        Mat resizedMat = new Mat();
        Imgproc.resize(orgMat, resizedMat, new Size(newWidth, newHeight));

        // 패딩 추가 (검은색 배경)
        Mat paddedMat = new Mat();
        Core.copyMakeBorder(resizedMat, paddedMat, top, bottom, left, right, Core.BORDER_CONSTANT, BLACK);

        // 패딩 정보 정리
        Rect padding = new Rect(left, top, right - left, bottom - top);

        // 최종 결과 정리
        retMap.put("mat", paddedMat);
        retMap.put("padding", padding);
        retMap.put("scale", minScale);

        return retMap;
    }
}
