package com.tobee.opencv.detector.history;

import com.tobee.opencv.detector.ObjectDetector;
import com.tobee.opencv.detector.tests.input.DNNOption;
import conf.CocoLabel;
import java.io.File;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import javax.swing.JOptionPane;
import org.opencv.core.Core;

import org.opencv.core.CvType;
import org.opencv.core.Mat;
import org.opencv.core.MatOfDouble;
import org.opencv.core.MatOfFloat;
import org.opencv.core.MatOfInt;
import org.opencv.core.MatOfRect2d;
import org.opencv.core.Point;
import org.opencv.core.Rect2d;
import org.opencv.core.Scalar;
import org.opencv.core.Size;
import org.opencv.dnn.Dnn;
import org.opencv.dnn.Net;
import org.opencv.highgui.HighGui;
import org.opencv.imgcodecs.Imgcodecs;
import org.opencv.imgproc.Imgproc;
import org.opencv.utils.Converters;
import org.opencv.videoio.VideoCapture;
import org.opencv.videoio.VideoWriter;
import org.opencv.videoio.Videoio;

public class  YoloDetector_v3 extends ObjectDetector{
    static{
        System.loadLibrary(Core.NATIVE_LIBRARY_NAME);
    }
    
    public YoloDetector_v3(final DNNOption dnnOption)
     {
         super(dnnOption);
     }
    
    @Override
    public void detectObject()
    {
        if(!new File(mediaFile).exists()) {
            JOptionPane.showMessageDialog(null, String.format("File9[%s] Not Found", mediaFile), "File Error", JOptionPane.WARNING_MESSAGE);
            return;
        }
        
        // load our YOLO object detector trained on COCO dataset
        Net dnnNet = readDNNetByModelName();
        
        if(dnnNet == null)
        {
            JOptionPane.showMessageDialog(null,"Failed to configure the network.", "Configuration Error", JOptionPane.WARNING_MESSAGE);
            return;
        }
        Map<String, Mat> outputLayers = getMapOutputLayer(dnnNet);
        
        if (mediaType.equals("Image"))
        {
            detectObjectOnImage(winName, dnnNet, outputLayers, mediaFile, threshold, isSaveOuput);
        }
        else if (mediaType.equals("Video"))
        {
            detectObjectOnVideo(winName, dnnNet, outputLayers, mediaFile, threshold, isSaveOuput);
        }
    }
    
    private MatOfInt applyNonMaximumSuppression(List<Rect2d> boxes, List<Float> confidences)
    {
        MatOfRect2d mOfRect = new MatOfRect2d();
        mOfRect.fromList(boxes);
        MatOfFloat mfConfs = new MatOfFloat(Converters.vector_float_to_Mat(confidences));
        MatOfInt indices = new MatOfInt();
        // 4️⃣ NMS 실행 (겹치는 박스 제거)
        //float nmsThreshold = 0.4f;
        float score_threshold = DEFAULT_THRESHOLD; //0.6f;
        float nms_threshold= DEFAULT_THRESHOLD; //0.5f;
        Dnn.NMSBoxes(mOfRect, mfConfs, score_threshold, nms_threshold, indices);
        
        return indices;
    }
    
    private MatOfInt drawBBox(Mat frame, List<Rect2d> boxes, List<Integer> classIDs, MatOfInt indices)
    {
        // create an object for label information.
        Map<Integer, String> cocoLabels =  CocoLabel.getCocoLabel();
        // generate radnom color in order to draw bounding boxes
        List<Scalar> colors = CocoLabel.getColorsByLabel();
        
        List indices_list = indices.toList();
        
        for (int i = 0; i < boxes.size(); i++) {
            if (indices_list.contains(i)) {
                Rect2d box = boxes.get(i);
                Point x_y = new Point(box.x, box.y);
                Point w_h = new Point(box.x + box.width, box.y + box.height);
                Point text_point = new Point(box.x, box.y - 5);
                
                Imgproc.rectangle(frame, w_h, x_y, colors.get(classIDs.get(i)), 1);
                String label = cocoLabels.get(classIDs.get(i));
                
                Imgproc.putText(frame, label, text_point, 
                    Imgproc.FONT_HERSHEY_SIMPLEX, 0.7, colors.get(classIDs.get(i)),2);
                
//                Imgproc.putText(img, label, text_point, 
//                    Imgproc.FONT_HERSHEY_SIMPLEX, 1, colors.get(classIdList.get(i)), 2);
            }
        }
        
        return indices;
    }
    
    public void detectObjectOnImage(
        final String winName, final Net dnnNet, Map<String, Mat> outputLayers, 
        final String mediaFile, double threshold, boolean isSaveOuput
    ) {
        Mat frame = Imgcodecs.imread(mediaFile);
        
        int idx = new File(mediaFile).getName().split("\\.").length - 2;
        
        String outputFile 
            = new File(mediaFile).getParent() + File.separator + "out" +  File.separator + 
                new File(mediaFile).getName().split("\\.")[idx] + "_yolov_out.jpg";
        
        Mat blob = Dnn.blobFromImage(frame, 1 / 255.0, BLOB_YOLO_V4_SIZE,
                        new Scalar(new double[] { 0.0, 0.0, 0.0 }), true, false);
        dnnNet.setInput(blob);
        blob.release();
        
        for(Iterator<String> iter = outputLayers.keySet().iterator(); iter.hasNext();)
        {
            String key = iter.next();
            Mat dtectMat = dnnNet.forward(key);
            outputLayers.put(key, dtectMat);
            //System.out.println(key);
        }
        
        // 5. 탐지된 객체 저장할 리스트
        List<Rect2d> boxes = new ArrayList<>();
        List<Integer> classIDs = new ArrayList<>();
        List<Float> confidences = new ArrayList<>();
		
        int width = (int) frame.size().width;
        int height = (int) frame.size().height;
		
        for(Iterator<String> outputMatIter = outputLayers.keySet().iterator(); outputMatIter.hasNext();) {
            String layerName = outputMatIter.next();
            Mat output = outputLayers.get(layerName);

            // 행(row) = 그리드 셀 개수
            int rows = output.rows();
            // 열(column) = 85 (BBox 좌표 + Objectness + 클래스 확률)
            int cols = output.cols();

            System.out.println("Layer " + layerName + " -> Grid Size: " + rows + ", Attributes per Cell: " + cols);

            for (int r = 0; r < rows; r++) {
                // BBox 정보 가져오기
                float[] detection = new float[cols];
                output.get(r, 0, detection);

                float x = detection[0];  // 중심 x 좌표
                float y = detection[1];  // 중심 y 좌표
                float w = detection[2];  // 너비
                float h = detection[3];  // 높이
                float confidence = detection[4];  // Objectness score

                // 클래스 확률 찾기
                float maxClassScore = -1;
                int classID = -1;
                for (int j = 5; j < cols; j++) {
                    if (detection[j] > maxClassScore) {
                        maxClassScore = detection[j];
                        classID = j - 5;
                    }
                }

                // 신뢰도가 높은 객체만 출력
                if (confidence > DEFAULT_THRESHOLD) {
                    System.out.printf("Detected Object: Class=%d, Confidence=%.2f, BBox=(%.2f, %.2f, %.2f, %.2f)%n",
                            classID, confidence, x, y, w, h);
							
                    int xLeft = (int) ((x - w / 2) * width);
                    int yTop = (int) ((y - h / 2) * height);
                    int boxWidth = (int) (w * width);
                    int boxHeight = (int) (h * height);
                    Rect2d box = new Rect2d(xLeft, yTop, boxWidth, boxHeight);
                    //boxes.add(new Rect(xLeft, yTop, boxWidth, boxHeight));
                    boxes.add(box);
                    classIDs.add(classID);
                    confidences.add(confidence);
                }
            }
        }
		
	MatOfInt indices = applyNonMaximumSuppression(boxes, confidences);
        drawBBox(frame, boxes, classIDs, indices);
  
        HighGui.imshow(winName, frame);
        
        MatOfDouble layersTimes = new MatOfDouble();
        double freq = Core.getTickFrequency() / 1000;
        double t = dnnNet.getPerfProfile(layersTimes) / freq;
        String label = String.format("%s, Inference time for a frame : %f ms", modelName, t);
        Imgproc.putText(frame, label, new Point(0, 15), 
                        Imgproc.FONT_HERSHEY_SIMPLEX, 0.5, new Scalar(0, 0, 0), 2);
        
        if(isSaveOuput)
        {
            // Write the frame with the detection boxes
            Mat detectedFrame = new Mat();
            frame.convertTo(detectedFrame, CvType.CV_8U);

            System.out.println("-->" + outputFile);
            Imgcodecs.imwrite(outputFile, detectedFrame);
            
            detectedFrame.release();
        }
        
        // ✅ 윈도우가 닫히면 종료 처리
        int key = HighGui.waitKey(0); // 키 입력 대기
        if (key == 27) {  // ESC 키가 눌리면 종료
            System.out.println("윈도우 닫힘 (ESC 눌림)");
        }

        // ✅ 윈도우 닫히기 전 명시적 종료
        HighGui.destroyAllWindows();  // 모든 윈도우 닫기
        System.exit(0);  // 프로그램 종료
    }
                
    public void detectObjectOnVideo(
        final String winName, final Net dnnNet, Map<String, Mat> outputLayers,
        final String mediaFile, double threshold, boolean isSaveOuput
    ) 
    {
        // Create a window
        HighGui.namedWindow(winName, HighGui.WINDOW_NORMAL);
        Mat frame = new Mat();

        VideoCapture capture = new VideoCapture(mediaFile);
        
        if (capture.open(mediaFile)) {
            capture.read(frame);
            HighGui.imshow(winName, frame);
        } else {
            System.out.println("Capture is not opened!");
            return;
        }
        
        // 5. 탐지된 객체 저장할 리스트
        List<Rect2d> boxes = new ArrayList<>();
        List<Integer> classIDs = new ArrayList<>();
        List<Float> confidences = new ArrayList<>();
        
        VideoWriter vwriter = new VideoWriter();
        Mat detectedFrame = new Mat();
        String outputFile = mediaFile.substring(0, mediaFile.length() - 4) + "_yolov4_out.avi";
        
        while (capture.isOpened()) {
            if(HighGui.waitKey(100) == 27) //ESC key to exit
            {
                System.out.println( "[ESC]Done processing !!!");
                //System.out.println( "[ESC]Output file is stored as " + outputFile);
                HighGui.waitKey(3000);
                break;
            }
             // get frame from the video
            capture.read(frame);
            
            // Stop the program if reached end of video
            if (frame.empty()) {
                System.out.println( "Done processing !!!");
                //System.out.println( "Output file is stored as " + outputFile);
                HighGui.waitKey(1000);
                
                if(vwriter.isOpened()) {
                    vwriter.release();
                } 
                vwriter = null;
                break;
            }
            
            Mat blob = Dnn.blobFromImage(frame, 1 / 255.0, BLOB_YOLO_V4_SIZE,
                        new Scalar(new double[] { 0.0, 0.0, 0.0 }), true, false);
            dnnNet.setInput(blob);
            blob.release();

            for(Iterator<String> iter = outputLayers.keySet().iterator(); iter.hasNext();)
            {
                String key = iter.next();
                Mat dtectMat = dnnNet.forward(key);
                outputLayers.put(key, dtectMat);
                //System.out.println(key);
            }

            int width = (int) frame.size().width;
            int height = (int) frame.size().height;

            for(Iterator<String> outputMatIter = outputLayers.keySet().iterator(); outputMatIter.hasNext();) {
                String layerName = outputMatIter.next();
                Mat output = outputLayers.get(layerName);

                // 행(row) = 그리드 셀 개수
                int rows = output.rows();
                // 열(column) = 85 (BBox 좌표 + Objectness + 클래스 확률)
                int cols = output.cols();

                for (int r = 0; r < rows; r++) {
                    // BBox 정보 가져오기
                    float[] detection = new float[cols];
                    output.get(r, 0, detection);

                    float x = detection[0];  // 중심 x 좌표
                    float y = detection[1];  // 중심 y 좌표
                    float w = detection[2];  // 너비
                    float h = detection[3];  // 높이
                    float confidence = detection[4];  // Objectness score

                    // 클래스 확률 찾기
                    float maxClassScore = -1;
                    int classID = -1;
                    for (int j = 5; j < cols; j++) {
                        if (detection[j] > maxClassScore) {
                            maxClassScore = detection[j];
                            classID = j - 5;
                        }
                    }

                    // 신뢰도가 높은 객체만 출력
                    if (confidence > DEFAULT_THRESHOLD) {
//                        System.out.printf("Detected Object: Class=%d, Confidence=%.2f, BBox=(%.2f, %.2f, %.2f, %.2f)%n",
//                                classID, confidence, x, y, w, h);
                        int xLeft = (int) ((x - w / 2) * width);
                        int yTop = (int) ((y - h / 2) * height);
                        int boxWidth = (int) (w * width);
                        int boxHeight = (int) (h * height);
                        Rect2d box = new Rect2d(xLeft, yTop, boxWidth, boxHeight);
                        //boxes.add(new Rect(xLeft, yTop, boxWidth, boxHeight));
                        boxes.add(box);
                        classIDs.add(classID);
                        confidences.add(confidence);
                    }
                }
            }

            MatOfInt indices = applyNonMaximumSuppression(boxes, confidences);
            drawBBox(frame, boxes, classIDs, indices);
            
            HighGui.imshow(winName, frame);
            
            MatOfDouble layersTimes = new MatOfDouble();
            double freq = Core.getTickFrequency() / 1000;
            double t = dnnNet.getPerfProfile(layersTimes) / freq;
            String label = String.format("YOLO, Inference time for a frame : %f ms", t);
            Imgproc.putText(frame, label, new Point(0, 15), 
                            Imgproc.FONT_HERSHEY_SIMPLEX, 0.5, new Scalar(0, 0, 0));
            
            if(!boxes.isEmpty()) {boxes.clear(); }
            if(!confidences.isEmpty()) {confidences.clear(); }
            if(!classIDs.isEmpty()) {classIDs.clear(); }
            if(!indices.empty()) {indices.release();}
            
            if(isSaveOuput)
            {
                double fps = capture.get(Videoio.CAP_PROP_FPS);
                System.out.println("fps: " + fps);
                if(fps == 0D)
                {
                    //fps = 20;
                    continue;
                }
                
                Size size 
                    = new Size(capture.get(Videoio.CAP_PROP_FRAME_WIDTH), capture.get(Videoio.CAP_PROP_FRAME_HEIGHT));

    //            int fourcc_xvid = VideoWriter.fourcc('X', 'V', 'I', 'D');   //# XVID (AVI format)
    //            int fourcc_mp4v = VideoWriter.fourcc('M', 'P', '4', 'V');  // # MP4 format
                int fourcc_mjpg = VideoWriter.fourcc('M', 'J', 'P', 'G');  //# Motion JPEG format
    //            int fourcc_h264 = VideoWriter.fourcc('H', '2', '6', '4'); // # H264 (more advanced compression)

                isSaveOuput = vwriter.open(outputFile, fourcc_mjpg, fps, size, true);

                if(!isSaveOuput)
                {
                    JOptionPane.showMessageDialog(null, 
                        String.format("Can not open the video file[sz:%s, fps:%f]", size.toString(), fps), 
                        "error", JOptionPane.INFORMATION_MESSAGE);
                    break;
                }
                
                 frame.convertTo(detectedFrame, CvType.CV_8U);
                if(vwriter.isOpened()) 
                {
                    vwriter.write(detectedFrame);
                }
                else
                {
                    System.out.println("========vwriter is not opened=======");
                }
            }
        } // of main while()
        if(capture.isOpened())
            capture.release();
        if(vwriter!=null && vwriter.isOpened()) {
            vwriter.release();
        } 

       // ✅ 윈도우가 닫히면 종료 처리
        int key = HighGui.waitKey(0); // 키 입력 대기
        if (key == 27) {  // ESC 키가 눌리면 종료
            System.out.println("윈도우 닫힘 (ESC 눌림)");
        }

        // ✅ 윈도우 닫히기 전 명시적 종료
        HighGui.destroyAllWindows();  // 모든 윈도우 닫기
        System.exit(0);  // 프로그램 종료
    }
}
