package com.tobee.opencv.detector.tests;

import java.awt.Dimension;
import java.awt.Toolkit;
import java.awt.Window;
import java.awt.image.BufferedImage;
import java.awt.image.DataBufferByte;
import java.awt.image.WritableRaster;
import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.List;

import javax.swing.ImageIcon;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.WindowConstants;

import org.opencv.core.CvType;
import org.opencv.core.Mat;
import org.opencv.core.Size;
import org.opencv.dnn.Net;
import org.opencv.imgproc.Imgproc;

public class CvUtils {
	public static void reSize(Mat image, int width, int height, int inter)
	{
	    Size dim;

	    if(image.empty()) return;

	    int h = image.rows();
	    int w = image.cols();
	    float r = 0.0f;
	    
	    if( width == 0 && height==0) return;

	    if( width == 0)
	    {
	        r = height / (float)(h);
	        Size sz = new Size((int)(w * r), height);
	        dim = sz;
	    }
	    else
	    {
	        r = width / (float)(w);
	        Size sz = new Size(width, (int)(h * r));
	        dim = sz;
	    }


	    Imgproc.resize(image, image, dim, inter);

	    return;
	}
	
	public static Mat reSizeByW(Mat image, int width, int inter)
	{
		reSize(image, width, 0, inter);

	    return image;
	}
	
	public static Mat reSizeByH(Mat image, int height, int inter)
	{
		reSize(image, 0, height, inter);

	    return image;
	}
	
	public static Mat reSizeCommon(Mat image, int scale)
	{
		int w = (int)(image.cols() / scale);
		//itoa(w,tmp,10);    // 정수형 -> 문자열 변환
        //name.append("_").append(tmp);
		reSizeByW(image, w, Imgproc.INTER_AREA);


	    return image;
	}
	
	public static List<String> getOutputNames(Net net) {
		List<String> names = new ArrayList<>();

		List<Integer> outLayers = net.getUnconnectedOutLayers().toList();
		List<String> layersNames = net.getLayerNames();

		outLayers.forEach((item) -> names.add(layersNames.get(item - 1)));// unfold and create R-CNN layers from the
																			// loaded YOLO model//
		return names;
	}
	
	public static List<String> getStringListOnFile(String filePath) {
		BufferedReader rdr = null;
		String line = null;
		List<String> stringList = new ArrayList<String>();

		try {
			rdr = new BufferedReader(new InputStreamReader(new FileInputStream(filePath)));
			while ((line = rdr.readLine()) != null) {
				stringList.add(line);
			}
		} catch (IOException ie) {
			ie.printStackTrace();
		} finally {
			if (rdr != null)
				try {
					rdr.close();
					rdr = null;
				} catch (IOException ie) {
				}
			line = null;
		}

		return stringList;
	}
	
	public static Object getMatVal(Mat inMat, int row, int col)
	{
		int channels = inMat.channels();
		Object reVal = null;
		
		switch(CvType.depth(inMat.type()))
		{
			case CvType.CV_32S:
			{
				//Integer retVal = -1;
				int[] values = new int[channels];
				inMat.get(row, col, values);
				reVal = values[0];
				break;
			}
			case CvType.CV_32F:
			{
				//Float retVal = -1f;
				float[] values = new float[channels];
				inMat.get(row, col, values);
				reVal = values[0];
				break;
			}
			case CvType.CV_16S:
			case CvType.CV_16U:
			{
				//Short retVal = -1;
				short[] values = new short[channels];
				inMat.get(row, col, values);
				reVal = values[0];
				break;
			}
			case CvType.CV_8U:
			case CvType.CV_8S:
			{
				//Byte retVal = -1;
				byte[] values = new byte[channels];
				inMat.get(row, col, values);
				reVal = values[0];
				break;
			}
		}
		
		return reVal;
	}
	
	//https://stackoverflow.com/questions/42466646/java-reading-a-frame-in-opencv
	public static BufferedImage matToBufferedImage(Mat frame) {
        int type = 0;
        if (frame.channels() == 1) {
            type = BufferedImage.TYPE_BYTE_GRAY;
        } else if (frame.channels() == 3) {
            type = BufferedImage.TYPE_3BYTE_BGR;
        }
        BufferedImage image = new BufferedImage(frame.width(), frame.height(), type);
        WritableRaster raster = image.getRaster();
        DataBufferByte dataBuffer = (DataBufferByte) raster.getDataBuffer();
        byte[] data = dataBuffer.getData();
        frame.get(0, 0, data);

        return image;
    }
	
	//https://stackoverflow.com/questions/42466646/java-reading-a-frame-in-opencv
	public static void showWindow(BufferedImage img) {
        showWindow("no title", img);
    }
	
    public static void showWindow(String title, BufferedImage img) {
        JFrame frame = new JFrame();
        frame.getContentPane().add(new JLabel(new ImageIcon(img)));
        frame.setLocationRelativeTo(null);
        frame.setDefaultCloseOperation(WindowConstants.EXIT_ON_CLOSE);
        frame.setSize(img.getWidth(), img.getHeight() + 30);
        frame.setTitle(title);
        centreWindow(frame);
        frame.setVisible(true);
    }
	
	public static void centreWindow(Window frame) {
	    Dimension dimension = Toolkit.getDefaultToolkit().getScreenSize();
	    int x = (int) ((dimension.getWidth() - frame.getWidth()) / 2);
	    int y = (int) ((dimension.getHeight() - frame.getHeight()) / 2);
	    frame.setLocation(x, y);
	}
}
