/*
 * Click nbfs://nbhost/SystemFileSystem/Templates/Licenses/license-default.txt to change this license
 * Click nbfs://nbhost/SystemFileSystem/Templates/Classes/Class.java to edit this template
 */
package com.tobee.opencv.detector.tests;

import static com.tobee.opencv.detector.IObjectDetector.BLOB_MASK_RCNN_SIZE;
import static com.tobee.opencv.detector.IObjectDetector.MASK_RCNN_SIZE;
import com.tobee.opencv.detector.MaskRCNNDetector;
import conf.CocoLabel;
import org.opencv.core.*;
import org.opencv.dnn.*;
import org.opencv.imgcodecs.Imgcodecs;
import org.opencv.imgproc.Imgproc;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import org.opencv.highgui.HighGui;
import static com.tobee.opencv.detector.IObjectDetector.DEFAULT_THRESHOLD;

/**
 *
 * @author
 */
public class MaskRCNNExample {
    static { System.loadLibrary(Core.NATIVE_LIBRARY_NAME); }

    public static void main(String[] args) {
        // 모델 및 구성 파일 경로
        String modelWeights = "mask_rcnn_inception_v2_coco_2018_01_28/frozen_inference_graph.pb";
        String modelConfig = "mask_rcnn_inception_v2_coco_2018_01_28.pbtxt";
        modelWeights = "D:\\DEV\\WORK\\Vision\\pretrained_model\\mrcnn\\frozen_inference_graph.pb";
        modelConfig = "D:\\DEV\\WORK\\Vision\\pretrained_model\\mrcnn\\mask_rcnn_inception_v2_coco_2018_01_28.pbtxt";
        String imageFile = "input.jpg";
        imageFile = "D:\\DEV\\WORK\\Vision\\data\\images\\2516944023_d00345997d_z.jpg";

        // 네트워크 로드
        Net net = Dnn.readNetFromTensorflow(modelWeights, modelConfig);

        // 이미지 로드
        Mat image = Imgcodecs.imread(imageFile);
        Map<String, Object> resultMap  = resizeMatImage(image, BLOB_MASK_RCNN_SIZE);
        
        //Mat blob = Dnn.blobFromImage(image, 1.0, new Size(800, 800), new Scalar(0, 0, 0), true, false);
        Mat resizedMat = (Mat)resultMap.get("mat");
        Rect padding = (Rect)resultMap.get("padding");
        float scale = (float)resultMap.get("scale");
        
        Mat blob = Dnn.blobFromImage(resizedMat, 1.0, BLOB_MASK_RCNN_SIZE, Scalar.all(0), true, false);
        
        // 네트워크에 입력 설정
        net.setInput(blob);
       // blob.release();

        // 순방향 패스 실행
        List<Mat> result = new ArrayList<>();
        List<String> outBlobNames = new ArrayList<>();
        outBlobNames.add("detection_out_final");
        outBlobNames.add("detection_masks");
        net.forward(result, outBlobNames);

        // 결과 추출
        Mat boxes = result.get(0);
        Mat masks = result.get(1);

        float confidenceThreshold = 0.5f;
        int maxObjCnt = boxes.size(2);
        int attrSize = boxes.size(3);
        boxes = boxes.reshape(1, (int) boxes.total() / attrSize);
        
        // 5️⃣ confidence 값 추출
        int cols = boxes.cols();

        float[] data = new float[(int) (boxes.total())];
        boxes.get(0, 0, data);  // Mat 데이터를 1D 배열로 변환
        CocoLabel.setupCocoLabelSet("mscoco_labels.names");
        Map<Integer,String> cocoLabels = CocoLabel.getCocoLabel();
        List<Scalar> colors = CocoLabel.getColorsByLabel();
        
        for (int i = 0; i < maxObjCnt; i++) {
            
            //double score = detection[idx + 2]; // 신뢰도 추출
            float score = boxes.at(Float.class, i, 2).getV();
            
            if (score > confidenceThreshold) {
                int classId = (int)data[i * cols + 1] ;
                int left   = (int)((data[i * cols + 3] * MASK_RCNN_SIZE - padding.x) / scale);
                int top    = (int)((data[i * cols + 4] * MASK_RCNN_SIZE - padding.y) / scale);
                int right  = (int)((data[i * cols + 5] * MASK_RCNN_SIZE - padding.x) / scale);
                int bottom = (int)((data[i * cols + 6] * MASK_RCNN_SIZE - padding.y) / scale);
                
                Rect box = new Rect(left, top, right - left + 1, bottom - top + 1);
                String label = String.format("%s: %.2f%%", cocoLabels.get(classId), score * 100);
                
                // Display the label at the top of the bounding box
                int[] baseLine = { 0 };
                Size labelSize = Imgproc.getTextSize(label, Imgproc.FONT_HERSHEY_SIMPLEX, 0.5, 1, baseLine);
                box.y = Math.max(box.y, (int) labelSize.height);

                // Draw background for text
                Imgproc.rectangle(
                    image, new Point(box.x, box.y - Math.round(1.5 * labelSize.height)),
                    new Point(box.x + Math.round(1.5 * labelSize.width), box.y + baseLine[0]), 
                    new Scalar(255, 255, 255), Imgproc.FILLED);
                
                Imgproc.putText(image, label, new Point(left, top - 5), Imgproc.FONT_HERSHEY_SIMPLEX, 0.5, colors.get(classId), 2);
                
                // frame이 비어있는지 확인
                if (image.empty()) {
                    System.out.println("❌ frame이 비어있습니다! 이미지 로드에 실패한 것 같습니다.");
                    continue;
                }

                // frame의 크기가 올바른지 확인
                int width = image.cols();
                int height = image.rows();
                if (width <= 0 || height <= 0) {
                    System.out.println("❌ frame 크기가 유효하지 않습니다.");
                    continue;
                }
                
//                Map<Integer, String> cocoLabels = CocoLabel.getCocoLabel();
//                List<Scalar> colors = CocoLabel.getColorsByLabel();
                // 1. rowMat을 2D로 변환하여, 특정 클래스에 대한 마스크 데이터를 가져옵니다.
                Mat rowMat = masks.row(i).col(classId);
                Scalar classColor = colors.get(classId);
                
                // 2. 마스크 크기를 지정하고, 그레이스케일로 변환하여 객체 마스크를 생성합니다.
                Mat objectMask = new Mat(rowMat.size(2), rowMat.size(3), CvType.CV_32F);
                rowMat.reshape(1, (int) rowMat.total()).copyTo(objectMask.reshape(1, (int) rowMat.size(2)));

                // 3. 마스크를 이진화하여 적용할 수 있도록 임계값을 설정합니다.
                Imgproc.threshold(objectMask, objectMask, 0.5, 1.0, Imgproc.THRESH_BINARY);

                // 4. `objectMask`를 `CV_8UC1`로 변환하여 색상 채우기를 준비합니다.
                objectMask.convertTo(objectMask, CvType.CV_8UC1); // 변환

                // 5. `objectMask`를 그레이스케일로 만들고 색상 적용 (RGB 색상으로 덮을 수 있도록)
                Mat colorMask = new Mat();
                Imgproc.cvtColor(objectMask, colorMask, Imgproc.COLOR_GRAY2BGR);

                // 6. 색상 마스크 적용: green color (Scalar(0, 255, 0))
                //Scalar color = new Scalar(0, 255, 0); // 초록색
                Core.multiply(colorMask, classColor, colorMask); // 색상을 곱하여 마스크에 적용

                // 7. 경계를 부드럽게 처리 (마스크의 경계를 부드럽게 만드는 효과)
                Mat smoothMask = new Mat();
                Imgproc.GaussianBlur(colorMask, smoothMask, new Size(rowMat.size(2), rowMat.size(3)), 0);

                // 8. ROI (bounding box) 추출
                Mat roi = image.submat(top, bottom, left, right); // Bounding box 영역에 해당하는 부분을 추출
                //System.out.println("ROI >>> " + roi);
                // 9. colorMask와 roi의 크기가 일치하는지 확인합니다. 일치하지 않으면 resize로 맞춥니다.
                if (roi.size().width != smoothMask.size().width || roi.size().height != smoothMask.size().height) {
                    Imgproc.resize(smoothMask, smoothMask, roi.size());
                }

                // 10. `smoothMask`의 데이터 타입을 `CV_8U`로 변환하여 `roi`와 일치시킵니다.
                smoothMask.convertTo(smoothMask, CvType.CV_8U);

                // 11. BBox 그리기 (원본 이미지에 경계를 추가합니다)
                //Scalar classColor = new Scalar(0, 255, 0); // 초록색
                Imgproc.rectangle(image, new Point(left, top), new Point(right, bottom), classColor, 2); // BBox

                // 12. 마스크를 원본 이미지에 덮어씌웁니다 (Alpha Blending 방식)
                //Core.addWeighted(roi, 1.0, smoothMask, 0.5, 0, roi); // 반투명 덮어씌우기
                Core.addWeighted(roi, 0.8, smoothMask, 0.5, 0, roi); // 반투명 덮어씌우기
//                // 바운딩 박스 그리기
//                Imgproc.rectangle(image, new Point(left, top), new Point(right, bottom), new Scalar(0, 255, 0), 2);
//
//                // 마스크 추출 및 크기 조정
//                Mat objectMask = masks.row(i).reshape(1, masks.size(2));
//                Imgproc.resize(objectMask, objectMask, new Size(right - left, bottom - top));
//
//                // 마스크 이진화 처리
//                Imgproc.threshold(objectMask, objectMask, 0.5, 1.0, Imgproc.THRESH_BINARY);
//                
//                // 마스크가 8비트로 변환되지 않았다면 강제로 CV_8U로 변환
//                if (objectMask.depth() != CvType.CV_8U) {
//                    objectMask.convertTo(objectMask, CvType.CV_8U);
//                }
//
//                // ROI 영역에 마스크 색 입히기
//                Mat roi = image.submat(top, bottom, left, right);
//                Mat coloredRoi = new Mat(roi.size(), CvType.CV_8UC3, new Scalar(0, 0, 255)); // 빨간색 마스크
//                coloredRoi.copyTo(roi, objectMask);
            }
        }
        
        HighGui.imshow("mrcnn", image);
        
         // ✅ 윈도우가 닫히면 종료 처리
        int key = HighGui.waitKey(0); // 키 입력 대기
        if (key == 27) {  // ESC 키가 눌리면 종료
            System.out.println("윈도우 닫힘 (ESC 눌림)");
        }

        // ✅ 윈도우 닫히기 전 명시적 종료
        HighGui.destroyAllWindows();  // 모든 윈도우 닫기
        System.exit(0);  // 프로그램 종료

        // 결과 이미지 저장
        //Imgcodecs.imwrite("D:\\DEV\\WORK\\Vision\\output.jpg", image);
    }
    
    /**
     * 
     * @param orgMat
     * @param targetSize
     * @return java.util.Map
     * 
     */
    protected static Map<String, Object> resizeMatImage(Mat orgMat, Size targetSize) {
        Map<String, Object> retMap = new HashMap<>();
        Rect padding =  new Rect();
        int targetWidth = (int)targetSize.width;
        int targetHeight = (int)targetSize.height;
        
        //float[] scale = null;
        int originalWidth = orgMat.cols();
        int originalHeight = orgMat.rows();

        // 비율 유지한 채 리사이징할 크기 계산
        float scaleX = (float) targetWidth / originalWidth;
        float scaleY = (float) targetHeight / originalHeight;
        float minScale = Math.min(scaleX, scaleY);

        int newWidth = (int) (originalWidth * minScale);
        int newHeight = (int) (originalHeight * minScale);

        // 패딩 값 계산
        int left = (targetWidth - newWidth) / 2;
        int top = (targetHeight - newHeight) / 2;
        int right = targetWidth - newWidth - left;
        int bottom = targetHeight - newHeight - top;

        // 리사이징 및 패딩 적용
        Mat resizedMat = new Mat();
        Imgproc.resize(orgMat, resizedMat, new Size(newWidth, newHeight));

        Mat paddedMat = new Mat();
        Core.copyMakeBorder(resizedMat, paddedMat, top, bottom, left, right, Core.BORDER_CONSTANT, new Scalar(0, 0, 0));

        // 패딩 정보 저장
        padding.x = left;
        padding.y = top;
        padding.width = right;
        padding.height = bottom;

        // 스케일 정보 저장
        //scale[0] = minScale;
        retMap.put("mat", paddedMat);
        retMap.put("padding", padding);
        retMap.put("scale", minScale);
        return retMap;
    }
    
}
