/*
 * Click nbfs://nbhost/SystemFileSystem/Templates/Licenses/license-default.txt to change this license
 * Click nbfs://nbhost/SystemFileSystem/Templates/Classes/Class.java to edit this template
 */
package com.tobee.opencv.detector.tests;
import org.opencv.core.*;
import org.opencv.dnn.*;
import org.opencv.imgcodecs.Imgcodecs;
import org.opencv.imgproc.Imgproc;

import java.util.*;
import org.opencv.highgui.HighGui;

public class MaskRCNNLayerCheck {
    public static void main(String[] args) {
        System.loadLibrary(Core.NATIVE_LIBRARY_NAME);
        // 모델 파일 경로 설정
       String modelPath = "C:\\DEV\\Experimental\\Vision\\pretrained_model\\mrcnn\\frozen_inference_graph.pb";
        String configPath = "C:\\DEV\\Experimental\\Vision\\pretrained_model\\mrcnn\\mask_rcnn_inception_v2_coco_2018_01_28.pbtxt";

        // 모델 파일과 config 파일 확인
        if (!new java.io.File(modelPath).exists() || !new java.io.File(configPath).exists()) {
            System.out.println("❌ 모델 파일을 찾을 수 없습니다!");
            return;
        }

        // OpenCV DNN 모델 로드
        Net net = Dnn.readNetFromTensorflow(modelPath, configPath);

        // 모델에서 레이어 이름을 확인합니다.
        List<String> layerNames = net.getLayerNames();
        System.out.println("모델에 포함된 레이어 이름:");
        for (String layerName : layerNames) {
            System.out.println(layerName);
        }

        // 모델 출력 레이어 확인
        List<String> modelOutputs = new ArrayList<>();
        modelOutputs.add("detection_out_final");  // BBox 정보
        modelOutputs.add("detection_masks");  // 마스크 정보

        // 원본 이미지 로드
         Mat orgMat = Imgcodecs.imread("C:\\DEV\\Experimental\\Vision\\data\\images\\2516944023_d00345997d_z.jpg");
        if (orgMat.empty()) {
            System.out.println("❌ 이미지를 찾을 수 없습니다!");
            return;
        }

        // 이미지 리사이즈 및 패딩 추가
        int targetWidth = 800;
        int targetHeight = 800;
        Rect padding = new Rect();
        float[] scale = new float[1];
        Mat calcMat = letterboxResize(orgMat, targetWidth, targetHeight, padding, scale);

        // Blob 생성
        Mat blob = Dnn.blobFromImage(calcMat, 1.0, new Size(targetWidth, targetHeight), new Scalar(0, 0, 0), true, false);
        net.setInput(blob);

        // 추론 실행
        List<Mat> outputs = new ArrayList<>();
        net.forward(outputs, modelOutputs);

        // detection_out_final과 detection_masks 출력 레이어 확인
        Mat detection = outputs.get(0);
        Mat detectionMasks = outputs.get(1);

        // detection_masks의 크기 확인
        System.out.println("detection_masks 크기: " + detectionMasks.size());
System.out.println("detection_masks 크기: " + detectionMasks.size(2));
System.out.println("detection_masks 크기: " + detectionMasks.size(3));
        // detection_masks 확인하기
        if (detectionMasks.empty()) {
            System.out.println("❌ detection_masks가 비어 있습니다.");
        } else {
            System.out.println("✅ detection_masks가 제대로 출력되었습니다.");
        }

        // detection_masks에서 첫 번째 마스크를 확인
        if (detectionMasks.rows() > 0) {
            Mat mask = detectionMasks.row(0);  // 첫 번째 객체 마스크 추출
            System.out.println("첫 번째 객체 마스크 크기: " + mask.size());
        }

        // 결과 출력
        HighGui.imshow("Result", orgMat);
        HighGui.waitKey(0);
    }

    // 이미지 리사이즈 및 패딩 함수
    public static Mat letterboxResize(Mat orgMat, int targetWidth, int targetHeight, Rect padding, float[] scale) {
        int originalWidth = orgMat.cols();
        int originalHeight = orgMat.rows();

        // 비율 유지한 채 리사이징할 크기 계산
        float scaleX = (float) targetWidth / originalWidth;
        float scaleY = (float) targetHeight / originalHeight;
        float minScale = Math.min(scaleX, scaleY);

        int newWidth = (int) (originalWidth * minScale);
        int newHeight = (int) (originalHeight * minScale);

        // 패딩 값 계산
        int left = (targetWidth - newWidth) / 2;
        int top = (targetHeight - newHeight) / 2;
        int right = targetWidth - newWidth - left;
        int bottom = targetHeight - newHeight - top;

        // 리사이징 및 패딩 적용
        Mat resizedMat = new Mat();
        Imgproc.resize(orgMat, resizedMat, new Size(newWidth, newHeight));

        Mat paddedMat = new Mat();
        Core.copyMakeBorder(resizedMat, paddedMat, top, bottom, left, right, Core.BORDER_CONSTANT, new Scalar(0, 0, 0));

        // 패딩 정보 저장
        padding.x = left;
        padding.y = top;
        padding.width = right;
        padding.height = bottom;

        // 스케일 정보 저장
        scale[0] = minScale;

        return paddedMat;
    }
}
