package com.tobee.opencv.detector.tests;

import static com.tobee.opencv.detector.IObjectDetector.BLOB_MASK_RCNN_SIZE;
import static com.tobee.opencv.detector.IObjectDetector.maskThreshold;
import com.tobee.opencv.detector.tests.CvUtils;
import conf.CocoLabel;
import java.io.File;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import org.opencv.core.Core;
import org.opencv.core.CvType;
import org.opencv.core.Mat;
import org.opencv.core.MatOfPoint;
import org.opencv.core.Point;
import org.opencv.core.Rect;
import org.opencv.core.Scalar;
import org.opencv.core.Size;
import org.opencv.dnn.Dnn;
import org.opencv.dnn.Net;
import org.opencv.highgui.HighGui;
import org.opencv.imgcodecs.Imgcodecs;
import org.opencv.imgproc.Imgproc;


public class MaskRCNNTest2 {

    static { System.loadLibrary(Core.NATIVE_LIBRARY_NAME); }
    public static void main(String[] args) {
        // ✅ 1️⃣ 모델 파일 확인
        String modelPath = "C:\\DEV\\Experimental\\Vision\\pretrained_model\\mrcnn\\frozen_inference_graph.pb";
        String configPath = "C:\\DEV\\Experimental\\Vision\\pretrained_model\\mrcnn\\mask_rcnn_inception_v2_coco_2018_01_28.pbtxt";

        if (!new File(modelPath).exists() || !new File(configPath).exists()) {
            System.out.println("❌ 모델 파일을 찾을 수 없습니다!");
            return;
        }

        // ✅ 2️⃣ OpenCV 버전 체크
        System.out.println("✅ OpenCV Version: " + Core.VERSION);

        // ✅ 3️⃣ 원본 이미지 로드
        Mat orgMat = Imgcodecs.imread("C:\\DEV\\Experimental\\Vision\\data\\images\\2516944023_d00345997d_z.jpg");
        if (orgMat.empty()) {
            System.out.println("❌ 이미지를 찾을 수 없습니다!");
            return;
        }
        // ✅ 4️⃣ OpenCV DNN 모델 로드
        Net net = Dnn.readNetFromTensorflow(modelPath, configPath);
//        net.setPreferableBackend(Dnn.DNN_BACKEND_CUDA);
//        net.setPreferableTarget(Dnn.DNN_TARGET_CUDA);

//List<String> layerNames = net.getLayerNames();
//for (String name : layerNames) {
//    System.out.println(name);
//}
        // ✅ 7️⃣ 원본 크기 유지하면서 800x800으로 리사이즈
        int targetWidth = 800, targetHeight = 800;
        Rect padding = new Rect();
        float[] scale = new float[1];
        Mat calcMat  = letterboxResize(orgMat, targetWidth, targetHeight, padding, scale);
        // ✅ 8️⃣ Blob 생성 (Mean Subtraction 적용, RGB -> BGR 변환)
        //Mat blob = Dnn.blobFromImage(calcMat, 1.0, new Size(targetWidth, targetHeight),
        //                             new Scalar(102.9801, 115.9465, 122.7717), true, false);
        Mat blob = Dnn.blobFromImage(calcMat, 1.0, BLOB_MASK_RCNN_SIZE, Scalar.all(0), true, false);
        net.setInput(blob);

        // ✅ 9️⃣ 추론 실행 (Bounding Box는 detection_out_final에서 가져옴)
        List<String> modelOutputs = new ArrayList<>();
        modelOutputs.add("detection_out_final");  // BBox 정보 포함
        modelOutputs.add("detection_masks");  // Segmentation Mask 정보 포함

        List<Mat> outputs = new ArrayList<>();
        net.forward(outputs, modelOutputs);
        
        // ✅ 1️⃣0️⃣ Bounding Box 처리
        Mat detection = outputs.get(0);
        float[] data = new float[(int) (detection.total())];
        detection.get(0, 0, data);

        int rows = detection.size(2);
//        System.out.println( ">>>>" + detection);
//        System.out.println( "1>>>>" + detection.size(0));
//        System.out.println( "2>>>>" + detection.size(1));
//        System.out.println( "3>>>>" + detection.size(2));
//        System.out.println( "4>>>>" + detection.size(3));
        
        boolean foundObject = false;
        detection = detection.reshape(1, (int) detection.total() / 7);
        
        // 4️⃣ isContinuous 체크 후 clone()
        if (!detection.isContinuous()) {
            detection = detection.clone();
        }

        // 5️⃣ confidence 값 추출
        int cols = detection.cols();
        data = new float[(int) detection.total()];
        detection.get(0, 0, data);  // Mat 데이터를 1D 배열로 변환
        
        Mat detectionMasks = outputs.get(1);
//        System.out.println( ">>>>" +(detectionMasks));
//        
//        System.out.println( "1>>>>" + detectionMasks.size(0));
//        System.out.println( "2>>>>" + detectionMasks.size(1));
//        System.out.println( "3>>>>" + detectionMasks.size(2));
//        System.out.println( "4>>>>" + detectionMasks.size(3));
        //System.out.println( ">>>>" + detectionMasks.size(4));
        
//        System.out.println("detection size: " + outputs.get(0).size());
//        System.out.println("detectionMasks size: " + outputs.get(1).size());
//        
//        
//        if (detectionMasks.rows() == 90 && detectionMasks.cols() == 100) {
//            detectionMasks = detectionMasks.reshape(1, new int[]{100, 90, 15, 15}); // (N, H, W)로 변환
//        }
//        System.out.println("Fixed detectionMasks size: " + detectionMasks.size());
//        System.out.println("detectionMasks : " + detectionMasks.rows());      
//
//        if (detectionMasks.empty()) {
//            System.out.println("detectionMasks is empty!");
//        } else {
//            System.out.println("detectionMasks size: " + detectionMasks.size());
//        }

//        System.out.println("detectionMasks total: " + detectionMasks.total());
//        // reshape 사용하여 4D 텐서로 변환 (90, 100, 15, 15) 형태로 변환
//        int[] newDims = {90, 100, 15, 15};  // (클래스 수, 객체 수, 마스크 높이, 마스크 너비)
//        detectionMasks = detectionMasks.reshape(1, newDims);  // 새로운 크기로 재구성
//
//        System.out.println("Reshaped detectionMasks size: " + detectionMasks.size());

        // 첫 번째 클래스, 첫 번째 객체의 마스크 데이터 가져오기
        //Mat mask = detectionMasks.row(0).col(0);  // 첫 번째 클래스, 첫 번째 객체의 마스크

        // 원하는 마스크 데이터를 확인
        //System.out.println("Mask size: " + mask.size());
        //
        //for (int classIdx = 0; classIdx < 90; classIdx++) {
        //    for (int objIdx = 0; objIdx < 100; objIdx++) {
        //        // 각 객체에 대한 마스크 가져오기
        //        Mat maska = detectionMasks.row(classIdx).col(objIdx);
        //        
        //        // 처리 작업 (예: 마스크 크기 출력)
        //        System.out.println("Mask size for class " + classIdx + ", object " + objIdx + ": " + maska.size());
        //    }
        //}
        Map<Integer, String> cocoLabels = CocoLabel.getCocoLabel();
        List<Scalar> colors = CocoLabel.getColorsByLabel();
        
        for (int i = 0; i < rows; i++) {
            float confidence = data[i * cols + 2];
   
            if (confidence > 0.5) {  // 🔥 신뢰도 임계값을 0.1로 낮춤
                foundObject = true;
                //System.out.println("🎯 객체 발견! Confidence: " + confidence);
                int classId = (int)data[i * cols + 1] ;
//                // ✅ 원본 크기에 맞게 Bounding Box 변환
//                int left   = (int)(data[i * cols + 3] * orgMat.width());
//                int top    = (int)(data[i * cols + 4] * orgMat.height());
//                int right  = (int)(data[i * cols + 5] * orgMat.width());
//                int bottom = (int)(data[i * cols + 6] * orgMat.height());
                // ✅ Bounding Box 좌표 변환
                int left   = (int)((data[i * cols + 3] * targetWidth - padding.x) / scale[0]);
                int top    = (int)((data[i * cols + 4] * targetHeight - padding.y) / scale[0]);
                int right  = (int)((data[i * cols + 5] * targetWidth - padding.x) / scale[0]);
                int bottom = (int)((data[i * cols + 6] * targetHeight - padding.y) / scale[0]);
                
                Mat rowMat = detectionMasks.row(i).col(classId);
                Rect box = new Rect(left, top, right - left + 1, bottom - top + 1);
                
//                 float fVal = (float)CvUtils.getMatVal(detection, i, 1);
//                Mat objectMask = new Mat(rowMat.size(1), rowMat.size(2), CvType.CV_32F, new Scalar(fVal));
                
//                Mat objectMask = new Mat(rowMat.size(2), rowMat.size(3), CvType.CV_32F);
                //Mat objectMask = new Mat(rowMat.size(2), rowMat.size(3), CvType.CV_32F, new Scalar(0, 255, 0));
//                System.out.println(objectMask);
//                 drawBoxWithMask(
//        orgMat, classId, confidence, 
//        box, objectMask, CocoLabel.getCocoLabel(), CocoLabel.getColorsByLabel()); 
                //Scalar color = colors.get(classId % colors.size());
                Scalar classColor = CocoLabel.getColorsByLabel().get(classId);
                // Draw a rectangle displaying the bounding box
                Imgproc.rectangle(orgMat, new Point(box.x, box.y), 
                                  new Point(box.x + box.width, box.y + box.height),
                                  classColor, 1);

                // Get the label for the class name and its confidence
                String label = String.format("%.2f", confidence);
                if (!cocoLabels.isEmpty()) {
                    label = cocoLabels.get(classId) + ":" + label;
                }

                // Display the label at the top of the bounding box
                int[] baseLine = { 0 };
                Size labelSize = Imgproc.getTextSize(label, Imgproc.FONT_HERSHEY_SIMPLEX, 0.5, 1, baseLine);
                box.y = Math.max(box.y, (int) labelSize.height);

                // Draw background for text
                Imgproc.rectangle(
                    orgMat, new Point(box.x, box.y - Math.round(1.5 * labelSize.height)),
                    new Point(box.x + Math.round(1.5 * labelSize.width), box.y + baseLine[0]), 
                    new Scalar(255, 255, 255), Imgproc.FILLED);

                // Put text on the image
                Imgproc.putText(orgMat, label, new Point(box.x, box.y), 
                                Imgproc.FONT_HERSHEY_SIMPLEX, 0.5, new Scalar(0, 0, 0), 1);

//case 0
//                // Get image ROI
//                Mat image_roi = new Mat(orgMat, box);
//
//                // Apply color to ROI
//                Scalar s2 = new Scalar(0.3);
//                Scalar s3 = color.mul(s2);
//
//                // 0.7 * image_roi and 0.3 * color blending
//                Mat image_roi2 = new Mat();
//                Scalar s = new Scalar(0.7);
//                Core.multiply(image_roi, s, image_roi2);
//
//                Mat coloredRoi = new Mat();
//                Core.add(image_roi2, s3, coloredRoi);
//
//                coloredRoi.convertTo(coloredRoi, CvType.CV_8UC3);
//
//                // Draw contours on the image
//                List<MatOfPoint> contours = new ArrayList<>();
//                Mat masked = new Mat();
//                objectMask.setTo(new Scalar(maskThreshold), masked);
//                masked.convertTo(masked, CvType.CV_8U);
//
//                Mat hierarchy = new Mat();
//                Imgproc.findContours(masked, contours, hierarchy, 
//                                     Imgproc.RETR_CCOMP, Imgproc.CHAIN_APPROX_SIMPLE);
//               
//                // Draw contours if they exist
//                if (!contours.isEmpty()) {
//                    Imgproc.drawContours(coloredRoi, contours, -1, color, 2, 
//                                         Imgproc.LINE_8, hierarchy, 100);
//                    coloredRoi.copyTo(image_roi, masked);
//                }
//
//                // Clean up
//                masked.release();
//                image_roi2.release();
//                hierarchy.release();
//                coloredRoi.release();
//
//                // Clear contours
//                contours.clear();

                
                //case 1
                
                // 1. rowMat을 2D로 변환하여, 특정 클래스에 대한 마스크 데이터를 가져옵니다.
//                Mat rowMat = detectionMasks.row(i).col(classId);

                // 2. 마스크 크기를 지정하고, 그레이스케일로 변환하여 객체 마스크를 생성합니다.
                Mat objectMask = new Mat(rowMat.size(2), rowMat.size(3), CvType.CV_32F);
                rowMat.reshape(1, (int) rowMat.total()).copyTo(objectMask.reshape(1, (int) rowMat.size(2)));

                // 3. 마스크를 이진화하여 적용할 수 있도록 임계값을 설정합니다.
                Imgproc.threshold(objectMask, objectMask, 0.5, 1.0, Imgproc.THRESH_BINARY);

                // 4. `objectMask`를 `CV_8UC1`로 변환하여 색상 채우기를 준비합니다.
                objectMask.convertTo(objectMask, CvType.CV_8UC1); // 변환

                // 5. `objectMask`를 그레이스케일로 만들고 색상 적용 (RGB 색상으로 덮을 수 있도록)
                Mat colorMask = new Mat();
                Imgproc.cvtColor(objectMask, colorMask, Imgproc.COLOR_GRAY2BGR);

                // 6. 색상 마스크 적용: green color (Scalar(0, 255, 0))
                //Scalar color = new Scalar(0, 255, 0); // 초록색
                Core.multiply(colorMask, classColor, colorMask); // 색상을 곱하여 마스크에 적용

                // 7. 경계를 부드럽게 처리 (마스크의 경계를 부드럽게 만드는 효과)
                Mat smoothMask = new Mat();
                Imgproc.GaussianBlur(colorMask, smoothMask, new Size(rowMat.size(2), rowMat.size(3)), 0);

                // 8. ROI (bounding box) 추출
                Mat roi = orgMat.submat(top, bottom, left, right); // Bounding box 영역에 해당하는 부분을 추출

                // 9. colorMask와 roi의 크기가 일치하는지 확인합니다. 일치하지 않으면 resize로 맞춥니다.
                if (roi.size().width != smoothMask.size().width || roi.size().height != smoothMask.size().height) {
                    Imgproc.resize(smoothMask, smoothMask, roi.size());
                }

                // 10. `smoothMask`의 데이터 타입을 `CV_8U`로 변환하여 `roi`와 일치시킵니다.
                smoothMask.convertTo(smoothMask, CvType.CV_8U);

                // 11. BBox 그리기 (원본 이미지에 경계를 추가합니다)
                //Scalar classColor = new Scalar(0, 255, 0); // 초록색
                Imgproc.rectangle(orgMat, new Point(left, top), new Point(right, bottom), classColor, 2); // BBox

                // 12. 마스크를 원본 이미지에 덮어씌웁니다 (Alpha Blending 방식)
                //Core.addWeighted(roi, 1.0, smoothMask, 0.5, 0, roi); // 반투명 덮어씌우기
                Core.addWeighted(roi, 0.8, smoothMask, 0.5, 0, roi); // 반투명 덮어씌우기

                // 13. 최종적으로 `orgMat`을 업데이트하여 마스크를 덮은 결과를 저장합니다.
                System.out.println("Applied smooth green mask to orgMat.");

      //case 2          
//                Scalar classColor = CocoLabel.getColorsByLabel().get(classId); //new Scalar(0, 255, 0)
//                System.out.println("classColor" + classColor);
//                // ✅ Bounding Box 그리기
//                Imgproc.rectangle(orgMat, new Point(left, top), new Point(right, bottom), classColor, 2);
//                // 라벨 추가 (예: "Object: 85%")
//                String label = String.format("%s: %.2f%%", CocoLabel.getCocoLabel().get(classId), confidence * 100);
//                Imgproc.putText(orgMat, label, new Point(left, top - 10), Imgproc.FONT_HERSHEY_SIMPLEX, 0.5, classColor, 2);
//                
//               // 1. rowMat을 2D로 변환하여, 특정 클래스에 대한 마스크 데이터를 가져옵니다.
//                Mat rowMat = detectionMasks.row(i).col(classId);
//
//                // 2. 마스크 크기를 지정하고, 그레이스케일로 변환하여 객체 마스크를 생성합니다.
//                Mat objectMask = new Mat(rowMat.size(2), rowMat.size(3), CvType.CV_32F);
//                rowMat.reshape(1, (int) rowMat.total()).copyTo(objectMask.reshape(1, (int) rowMat.size(2)));
//
//                // 3. 마스크를 이진화하여 적용할 수 있도록 임계값을 설정합니다.
//                Imgproc.threshold(objectMask, objectMask, 0.5, 1.0, Imgproc.THRESH_BINARY);
//
//                // 4. `objectMask`를 `CV_8UC1`로 변환하여 색상 채우기를 준비합니다.
//                objectMask.convertTo(objectMask, CvType.CV_8UC1); // 변환
//
//                // 5. `objectMask`를 그레이스케일로 만들고 색상 적용 (RGB 색상으로 덮을 수 있도록)
//                Mat colorMask = new Mat();
//                Imgproc.cvtColor(objectMask, colorMask, Imgproc.COLOR_GRAY2BGR);
//
//                // 6. 색상 마스크 적용: green color (Scalar(0, 255, 0))
//                Scalar color = new Scalar(0, 255, 0); // 초록색
//                Core.multiply(colorMask, color, colorMask); // 색상을 곱하여 마스크에 적용
//
//                // 7. ROI (bounding box) 추출
//                Mat roi = orgMat.submat(top, bottom, left, right); // Bounding box 영역에 해당하는 부분을 추출
//
//                // 8. colorMask와 roi의 크기가 일치하는지 확인합니다. 일치하지 않으면 resize로 맞춥니다.
//                if (roi.size().width != colorMask.size().width || roi.size().height != colorMask.size().height) {
//                    Imgproc.resize(colorMask, colorMask, roi.size());
//                }
//
//                // 9. `colorMask`의 데이터 타입을 `CV_8U`로 변환하여 `roi`와 일치시킵니다.
//                colorMask.convertTo(colorMask, CvType.CV_8U);
//
//                // 10. `addWeighted` 함수를 사용하여 `objectMask`를 원본 이미지에 덮어 씁니다.
//                // 마스크를 0.5 비율로 덮고, 원본 이미지를 반투명하게 결합하여 강조 효과를 줍니다.
//                Core.addWeighted(roi, 1.0, colorMask, 0.5, 0, roi);
//
//                // 11. 최종적으로 `orgMat`을 업데이트하여 마스크를 덮은 결과를 저장합니다.
//                System.out.println("Applied green mask to orgMat.");




            }
        }
        //orgMat.release();
        
        //orgMat = restoreOriginalSize(calcMat, padding, orgMat);
        
        if (!foundObject) {
            System.out.println("❌ 객체를 찾지 못했습니다.");
        }
        HighGui.imshow("Result", orgMat);
        // ✅ 윈도우가 닫히면 종료 처리
        int key = HighGui.waitKey(0); // 키 입력 대기
        if (key == 27) {  // ESC 키가 눌리면 종료
            System.out.println("윈도우 닫힘 (ESC 눌림)");
        }

        // ✅ 윈도우 닫히기 전 명시적 종료
        HighGui.destroyAllWindows();  // 모든 윈도우 닫기
        System.exit(0);  // 프로그램 종료
        
    }
    
     // Draw the predicted bounding box, colorize and show the mask on the image
    private static void drawBoxWithMask(
        Mat frame, int classId, double conf, 
        Rect box, Mat objectMask, 
        Map<Integer, String> cocoLabels,
        List<Scalar> colors) 
    {
        
        Scalar color = colors.get(classId % colors.size());

        // Draw a rectangle displaying the bounding box
        Imgproc.rectangle(frame, new Point(box.x, box.y), 
                          new Point(box.x + box.width, box.y + box.height),
                          color, 1);

        // Get the label for the class name and its confidence
        String label = String.format("%.2f", conf);
        if (!cocoLabels.isEmpty()) {
            label = cocoLabels.get(classId) + ":" + label;
        }

        // Display the label at the top of the bounding box
        int[] baseLine = { 0 };
        Size labelSize = Imgproc.getTextSize(label, Imgproc.FONT_HERSHEY_SIMPLEX, 0.5, 1, baseLine);
        box.y = Math.max(box.y, (int) labelSize.height);

        // Draw background for text
        Imgproc.rectangle(
            frame, new Point(box.x, box.y - Math.round(1.5 * labelSize.height)),
            new Point(box.x + Math.round(1.5 * labelSize.width), box.y + baseLine[0]), 
            new Scalar(255, 255, 255), Imgproc.FILLED);

        // Put text on the image
        Imgproc.putText(frame, label, new Point(box.x, box.y), 
                        Imgproc.FONT_HERSHEY_SIMPLEX, 0.5, new Scalar(0, 0, 0), 1);

        // Get image ROI
        Mat image_roi = new Mat(frame, box);

        // Apply color to ROI
        Scalar s2 = new Scalar(0.3);
        Scalar s3 = color.mul(s2);

        // 0.7 * image_roi and 0.3 * color blending
        Mat image_roi2 = new Mat();
        Scalar s = new Scalar(0.7);
        Core.multiply(image_roi, s, image_roi2);

        Mat coloredRoi = new Mat();
        Core.add(image_roi2, s3, coloredRoi);

        coloredRoi.convertTo(coloredRoi, CvType.CV_8UC3);

        // Draw contours on the image
        List<MatOfPoint> contours = new ArrayList<>();
        Mat mask = new Mat();
        objectMask.setTo(new Scalar(maskThreshold), mask);
        mask.convertTo(mask, CvType.CV_8U);

        Mat hierarchy = new Mat();
        Imgproc.findContours(mask, contours, hierarchy, 
                             Imgproc.RETR_CCOMP, Imgproc.CHAIN_APPROX_SIMPLE);

        // Draw contours if they exist
        if (!contours.isEmpty()) {
            Imgproc.drawContours(coloredRoi, contours, -1, color, 2, 
                                 Imgproc.LINE_8, hierarchy, 100);
            coloredRoi.copyTo(image_roi, mask);
        }

        // Clean up
        mask.release();
        image_roi2.release();
        hierarchy.release();
        coloredRoi.release();

        // Clear contours
        contours.clear();
    }

    
    public static Mat restoreOriginalSize(Mat paddedMat, Rect padding, Mat orgMat) {
        // 원본 이미지의 크기
        int originalWidth = orgMat.cols();
        int originalHeight = orgMat.rows();

        // 패딩 정보를 사용하여 이미지 크기 복원
        int left = padding.x;
        int top = padding.y;
        int right = padding.width;
        int bottom = padding.height;

        // 패딩 제거하여 원본 크기대로 잘라내기
        Rect cropRegion = new Rect(left, top, originalWidth, originalHeight);
        Mat restoredMat = paddedMat.submat(cropRegion);

        // 원본 크기로 맞추기 위해 리사이즈
        Mat finalMat = new Mat();
        Imgproc.resize(restoredMat, finalMat, new Size(originalWidth, originalHeight));

        return finalMat;
    }

    
    public static Mat letterboxResize(Mat orgMat, int targetWidth, int targetHeight, Rect padding, float[] scale) {
        int originalWidth = orgMat.cols();
        int originalHeight = orgMat.rows();

        // 비율 유지한 채 리사이징할 크기 계산
        float scaleX = (float) targetWidth / originalWidth;
        float scaleY = (float) targetHeight / originalHeight;
        float minScale = Math.min(scaleX, scaleY);

        int newWidth = (int) (originalWidth * minScale);
        int newHeight = (int) (originalHeight * minScale);

        // 패딩 값 계산
        int left = (targetWidth - newWidth) / 2;
        int top = (targetHeight - newHeight) / 2;
        int right = targetWidth - newWidth - left;
        int bottom = targetHeight - newHeight - top;

        // 리사이징 및 패딩 적용
        Mat resizedMat = new Mat();
        Imgproc.resize(orgMat, resizedMat, new Size(newWidth, newHeight));

        Mat paddedMat = new Mat();
        Core.copyMakeBorder(resizedMat, paddedMat, top, bottom, left, right, Core.BORDER_CONSTANT, new Scalar(0, 0, 0));

        // 패딩 정보 저장
        padding.x = left;
        padding.y = top;
        padding.width = right;
        padding.height = bottom;

        // 스케일 정보 저장
        scale[0] = minScale;

        return paddedMat;
    }

    public static void main6(String[] args) {
        // ✅ 1️⃣ 모델 파일 확인
        String modelPath = "C:\\DEV\\Experimental\\Vision\\pretrained_model\\mrcnn\\frozen_inference_graph.pb";
        String configPath = "C:\\DEV\\Experimental\\Vision\\pretrained_model\\mrcnn\\mask_rcnn_inception_v2_coco_2018_01_28.pbtxt";

        if (!new File(modelPath).exists() || !new File(configPath).exists()) {
            System.out.println("❌ 모델 파일을 찾을 수 없습니다!");
            return;
        }

        // ✅ 2️⃣ OpenCV 버전 체크
        System.out.println("✅ OpenCV Version: " + Core.VERSION);

        // ✅ 3️⃣ 원본 이미지 로드
        Mat orgMat = Imgcodecs.imread("C:\\DEV\\Experimental\\Vision\\data\\images\\2516944023_d00345997d_z.jpg");
        if (orgMat.empty()) {
            System.out.println("❌ 이미지를 찾을 수 없습니다!");
            return;
        }

        // ✅ 4️⃣ OpenCV DNN 모델 로드
        Net net = Dnn.readNetFromTensorflow(modelPath, configPath);

        // ✅ 5️⃣ 출력 레이어 확인
        List<String> outputNames = net.getUnconnectedOutLayersNames();
        System.out.println("✅ Output Layers: " + outputNames);

        // ✅ 6️⃣ "detection_out_final"이 있는지 확인 (Bounding Box 레이어)
        boolean hasDetectionLayer = outputNames.contains("detection_out_final");
        if (!hasDetectionLayer) {
            System.out.println("❌ detection_out_final 레이어가 없습니다! 모델이 잘못되었을 가능성이 높습니다.");
            //return;
            //outputNames.add("detection_out_final");
        }

        // ✅ 7️⃣ 원본 크기 유지하면서 800x800으로 리사이즈
        int targetWidth = 800, targetHeight = 800;
        Mat calcMat = new Mat();
        //Imgproc.resize(orgMat, calcMat, new Size(targetWidth, targetHeight));
        //calcMat = resizeImage(orgMat, targetWidth);
        calcMat = letterboxResize(orgMat, targetWidth, targetHeight);
        // ✅ 8️⃣ Blob 생성 (Mean Subtraction 적용, RGB -> BGR 변환)
        //Mat blob = Dnn.blobFromImage(calcMat, 1.0, new Size(targetWidth, targetHeight),
        //                             new Scalar(102.9801, 115.9465, 122.7717), true, false);
        Mat blob = Dnn.blobFromImage(calcMat, 1.0, BLOB_MASK_RCNN_SIZE, Scalar.all(0), true, false);
        net.setInput(blob);

        // ✅ 9️⃣ 추론 실행 (Bounding Box는 detection_out_final에서 가져옴)
        List<String> modelOutputs = new ArrayList<>();
        modelOutputs.add("detection_out_final");  // BBox 정보 포함
        modelOutputs.add("detection_masks");  // Segmentation Mask 정보 포함

        List<Mat> outputs = new ArrayList<>();
        net.forward(outputs, modelOutputs);
        //Mat dtectMat = net.forward("detection_out_final");
        
        // ✅ 1️⃣0️⃣ Bounding Box 처리
        Mat detection = outputs.get(0);
        float[] data = new float[(int) (detection.total())];
        detection.get(0, 0, data);
//System.out.println(outputs.size());
//System.out.println((int) (detection.total()));
        int rows = detection.size(2);
        int cols = detection.size(3);
        boolean foundObject = false;
        detection = detection.reshape(1, (int) detection.total() / 7);
        
        // 4️⃣ isContinuous 체크 후 clone()
        if (!detection.isContinuous()) {
            detection = detection.clone();
        }

        // 5️⃣ confidence 값 추출
        cols = detection.cols();
        data = new float[(int) detection.total()];
        detection.get(0, 0, data);  // Mat 데이터를 1D 배열로 변환

//        for (int i = 0; i < detection.rows(); i++) {
//            float confidence = data[i * cols + 2];  // 신뢰도 값
//            System.out.println("🎯 Confidence[" + i + "]: " + confidence);
//        }
        
        for (int i = 0; i < rows; i++) {
            float confidence = data[i * cols + 2];
            //float score = detection.at(Float.class, i, 2).getV();
//            System.out.println(confidence);
//            System.out.println(score);
            if (confidence > 0.4) {  // 🔥 신뢰도 임계값을 0.1로 낮춤
                foundObject = true;
                //System.out.println("🎯 객체 발견! Confidence: " + confidence);
                int classId = (int)data[i * cols + 1] ;
                // ✅ 원본 크기에 맞게 Bounding Box 변환
                int left   = (int)(data[i * cols + 3] * orgMat.width());
                int top    = (int)(data[i * cols + 4] * orgMat.height());
                int right  = (int)(data[i * cols + 5] * orgMat.width());
                int bottom = (int)(data[i * cols + 6] * orgMat.height());

                // ✅ Bounding Box 그리기
                Imgproc.rectangle(orgMat, new Point(left, top), new Point(right, bottom), new Scalar(0, 255, 0), 2);
                // 라벨 추가 (예: "Object: 85%")
                String label = String.format("%s: %.2f%%", CocoLabel.getCocoLabel().get(classId), confidence * 100);
                Imgproc.putText(orgMat, label, new Point(left, top - 10), Imgproc.FONT_HERSHEY_SIMPLEX, 0.5, new Scalar(0, 255, 0), 2);
            }
        }

        if (!foundObject) {
            System.out.println("❌ 객체를 찾지 못했습니다.");
        }

//        // ✅ 1️⃣1️⃣ 결과 저장
//        Imgcodecs.imwrite("output.jpg", orgMat);
//        System.out.println("✅ Bounding Box 결과 저장 완료! (output.jpg)");

        // ✅ 1️⃣2️⃣ 화면에 출력 (테스트 용도)
        HighGui.imshow("Result", orgMat);
        HighGui.waitKey(0);
    }
    
    static Mat letterboxResize(Mat orgMat, int targetWidth, int targetHeight) {
        int originalWidth = orgMat.cols();
        int originalHeight = orgMat.rows();

        float scale = Math.min((float) targetWidth / originalWidth, (float) targetHeight / originalHeight);
        int newWidth = (int) (originalWidth * scale);
        int newHeight = (int) (originalHeight * scale);

        // 이미지 리사이징 (비율 유지)
        Mat resizedMat = new Mat();
        Imgproc.resize(orgMat, resizedMat, new Size(newWidth, newHeight));

        // 패딩 추가
        int top = (targetHeight - newHeight) / 2;
        int bottom = targetHeight - newHeight - top;
        int left = (targetWidth - newWidth) / 2;
        int right = targetWidth - newWidth - left;

        Mat paddedMat = new Mat();
        Core.copyMakeBorder(resizedMat, paddedMat, top, bottom, left, right, Core.BORDER_CONSTANT, new Scalar(0, 0, 0));

        return paddedMat;  // 정비율 + 패딩된 이미지 반환
    }
    
 // 1. 이미지 크기 리사이즈 (800x800으로 리사이즈)
    static Mat resizeImage(Mat frame, int targetSize) {
        int width = frame.cols();
        int height = frame.rows();

        // 비율 유지하여 리사이즈할 새로운 크기 계산
        double scale = Math.min((double) targetSize / width, (double) targetSize / height);
        int newWidth = (int) (width * scale);
        int newHeight = (int) (height * scale);

        // 리사이즈
        Mat resized = new Mat();
        Imgproc.resize(frame, resized, new Size(newWidth, newHeight));

        // 리사이즈된 이미지가 targetSize에 맞지 않으면 패딩 추가
        Mat result = new Mat(targetSize, targetSize, frame.type(), new Scalar(0));
        resized.copyTo(result.rowRange(0, resized.rows()).colRange(0, resized.cols()));
        
        return result;
    }

    public static void main4(String[] args) {
        // OpenCV 라이브러리 초기화
        System.loadLibrary(Core.NATIVE_LIBRARY_NAME);
        // ✅ 1️⃣ 모델 파일 확인
        String modelPath = "C:\\DEV\\Experimental\\Vision\\pretrained_model\\mrcnn\\frozen_inference_graph.pb";
        String configPath = "C:\\DEV\\Experimental\\Vision\\pretrained_model\\mrcnn\\mask_rcnn_inception_v2_coco_2018_01_28.pbtxt";

        if (!new File(modelPath).exists() || !new File(configPath).exists()) {
            System.out.println("❌ 모델 파일을 찾을 수 없습니다!");
            return;
        }

        // ✅ 2️⃣ OpenCV 버전 체크
        System.out.println("✅ OpenCV Version: " + Core.VERSION);

        // ✅ 3️⃣ 원본 이미지 로드
        Mat orgMat = Imgcodecs.imread("C:\\DEV\\Experimental\\Vision\\data\\images\\2516944023_d00345997d_z.jpg");
        if (orgMat.empty()) {
            System.out.println("❌ 이미지를 찾을 수 없습니다!");
            return;
        }

        // ✅ 4️⃣ 원본 비율 유지하면서 800x800 리사이즈
        int targetWidth = 800, targetHeight = 800;
        Mat calcMat = new Mat();
        Imgproc.resize(orgMat, calcMat, new Size(targetWidth, targetHeight));

        // ✅ 5️⃣ DNN 모델 로드
        Net net = Dnn.readNetFromTensorflow(modelPath, configPath);

        // ✅ 6️⃣ Blob 생성 (Mean Subtraction 적용, RGB -> BGR 변환)
        Mat blob = Dnn.blobFromImage(calcMat, 1.0, new Size(targetWidth, targetHeight),
                                     new Scalar(102.9801, 115.9465, 122.7717), true, false);
        net.setInput(blob);

        // ✅ 7️⃣ 출력 레이어 확인
        List<String> outputNames = net.getUnconnectedOutLayersNames();
        System.out.println("✅ Output Layers: " + outputNames);

        // ✅ 8️⃣ 추론 실행
        List<Mat> outputs = new ArrayList<>();
        net.forward(outputs, outputNames);

        // ✅ 9️⃣ Detection 결과 분석
        Mat detection = outputs.get(0);
        float[] data = new float[(int) (detection.total())];
        detection.get(0, 0, data);

        int rows = detection.size(2);
        int cols = detection.size(3);
        boolean foundObject = false;

        for (int i = 0; i < rows; i++) {
            float confidence = data[i * cols + 2];

            if (confidence > 0.3) { // 최소 신뢰도 30% 이상
                foundObject = true;
                System.out.println("🎯 객체 발견! Confidence: " + confidence);

                // ✅ 10️⃣ 원본 이미지 크기로 Bounding Box 변환
                int left   = (int)(data[i * cols + 3] * orgMat.width());
                int top    = (int)(data[i * cols + 4] * orgMat.height());
                int right  = (int)(data[i * cols + 5] * orgMat.width());
                int bottom = (int)(data[i * cols + 6] * orgMat.height());

                // ✅ 11️⃣ Bounding Box 그리기
                Imgproc.rectangle(orgMat, new Point(left, top), new Point(right, bottom), new Scalar(0, 255, 0), 2);
            }
        }

        if (!foundObject) {
            System.out.println("❌ 객체를 찾지 못했습니다.");
        }

//        // ✅ 12️⃣ 결과 저장
//        Imgcodecs.imwrite("output.jpg", orgMat);
//        System.out.println("✅ Bounding Box 결과 저장 완료! (output.jpg)");

        // ✅ 13️⃣ 화면에 출력 (테스트 용도)
        HighGui.imshow("Result", orgMat);
        HighGui.waitKey(0);
    }
    
   public static void main3(String[] args) {
        // OpenCV 라이브러리 초기화
        System.loadLibrary(Core.NATIVE_LIBRARY_NAME);
        // 1️⃣ 원본 이미지 로드
        //Mat orgMat = Imgcodecs.imread("input.jpg");
        // 이미지 로드
        Mat orgMat = Imgcodecs.imread("C:\\DEV\\Experimental\\Vision\\data\\images\\2516944023_d00345997d_z.jpg");  // 다운로드한 이미지 파일 경로
	
        if (orgMat.empty()) {
            System.out.println("이미지를 찾을 수 없습니다!");
            return;
        }

        // 2️⃣ 리사이즈할 크기 (800x800)
        Size targetSize = new Size(800, 800);
        Mat calcMat = new Mat();
        Imgproc.resize(orgMat, calcMat, targetSize);

        // 3️⃣ DNN 모델 로드 (Mask R-CNN)
//        String modelPath = "mask_rcnn.pb";
//        String configPath = "mask_rcnn.pbtxt";
        String modelPath = "C:\\DEV\\Experimental\\Vision\\pretrained_model\\mrcnn\\frozen_inference_graph.pb";
        File modelFile = new File(modelPath);
        if (!modelFile.exists()) {
            System.out.println("Error: Model file not found!");
            return;
        }
        
        String configPath = "C:\\DEV\\Experimental\\Vision\\pretrained_model\\mrcnn\\mask_rcnn_inception_v2_coco_2018_01_28.pbtxt";
        Net net = Dnn.readNetFromTensorflow(modelPath, configPath);

        // 4️⃣ Blob 생성
        Mat blob = Dnn.blobFromImage(calcMat, 1.0, targetSize, new Scalar(0, 0, 0), true, false);
        net.setInput(blob);

        // 5️⃣ 추론 실행
        List<Mat> outputs = new ArrayList<>();
        List<String> outputNames = new ArrayList<>();
        outputNames.add("detection_out_final");
        outputNames.add("detection_masks");
        net.forward(outputs, outputNames);

        // 6️⃣ 결과에서 Bounding Box 추출
        Mat detection = outputs.get(0);
        float[] data = new float[(int) (detection.total())];
        detection.get(0, 0, data);

        int rows = detection.size(2);
        int cols = detection.size(3);

        // 7️⃣ 원본 이미지와 리사이즈된 이미지 크기 비율 계산
        float scaleX = (float) orgMat.width() / (float) targetSize.width;
        float scaleY = (float) orgMat.height() / (float) targetSize.height;

        for (int i = 0; i < rows; i++) {
            float confidence = data[i * cols + 2];
            //System.out.println(confidence);
            if (confidence > 0.3) { // 신뢰도 30% 이상이면 표시
                
                // 7️⃣ Bounding Box 좌표 변환 (원본 이미지 크기에 맞춤)
                int left   = (int)(data[i * cols + 3] * orgMat.width());
                int top    = (int)(data[i * cols + 4] * orgMat.height());
                int right  = (int)(data[i * cols + 5] * orgMat.width());
                int bottom = (int)(data[i * cols + 6] * orgMat.height());

                // 8️⃣ 좌표가 음수이거나 범위를 벗어나지 않도록 보정
                left = Math.max(0, Math.min(left, orgMat.width() - 1));
                top = Math.max(0, Math.min(top, orgMat.height() - 1));
                right = Math.max(0, Math.min(right, orgMat.width() - 1));
                bottom = Math.max(0, Math.min(bottom, orgMat.height() - 1));

                // 9️⃣ Bounding Box 그리기
                Imgproc.rectangle(orgMat, new Point(left, top), new Point(right, bottom), new Scalar(0, 255, 0), 2);
            }
        }

        // 🔟 결과 출력
        Imgcodecs.imwrite("C:\\Temp\\output_image.jpg", orgMat);
        System.out.println("Bounding Box가 그려진 이미지를 저장했습니다: output.jpg");
        
        // 결과 이미지 화면 출력
        Imgproc.resize(orgMat, orgMat, new Size(800, 600));
        HighGui.imshow("Result", orgMat);
        HighGui.waitKey(0);
    }
    public static void main2(String[] args) {
        // OpenCV 라이브러리 초기화
        System.loadLibrary(Core.NATIVE_LIBRARY_NAME);

        // 이미지 로드
        Mat frame = Imgcodecs.imread("C:\\DEV\\Experimental\\Vision\\data\\images\\2516944023_d00345997d_z.jpg");  // 다운로드한 이미지 파일 경로
		
		// 이미지가 로드되지 않았으면 종료
        if (frame.empty()) {
            System.out.println("Failed to load image.");
            return;
        }
		
		 // 이미지 크기 확인
        System.out.println("Original Image size: " + frame.size());
		
        // 모델 파일과 프로토콜 파일 경로 설정
        String modelWeights = "C:\\DEV\\Experimental\\Vision\\pretrained_model\\mrcnn\\frozen_inference_graph.pb";
        String modelConfig = "C:\\DEV\\Experimental\\Vision\\pretrained_model\\mrcnn\\mask_rcnn_inception_v2_coco_2018_01_28.pbtxt";
		
        // DNN 모델 로드
        Net net = Dnn.readNetFromTensorflow(modelWeights, modelConfig);
		// 이미지 크기 변경
        Mat resizedFrame = new Mat();
        Imgproc.resize(frame, resizedFrame, new Size(800, 800)); // 원하는 크기로 리사이즈
        System.out.println("Resized Image size: " + resizedFrame.size());
		
        // Mask R-CNN에 입력으로 사용할 blob 생성
        Mat blob = Dnn.blobFromImage(resizedFrame, 1.0, new Size(800, 800), new Scalar(0), true, false);
        
        // Blob이 잘 만들어졌는지 확인
        System.out.println("Blob created successfully.");
        System.out.println("Blob size: " + blob.size());
        net.setInput(blob);

         List<Mat> outs = new ArrayList<>();
        List<String> outBlobNames = List.of("detection_out_final", "detection_masks");
        net.forward(outs, outBlobNames);

        // 탐지 결과 처리
        Mat detectionOut = outs.get(0);
        Mat detectionMasks = outs.get(1);
        // 탐지된 물체의 개수
        int numDetections = detectionOut.size(2);
		System.out.println(numDetections);

		detectionOut = detectionOut.reshape(1, (int) detectionOut.total() / 7);
        // 각 객체에 대해 bounding box 및 마스크 적용
        for (int i = 0; i < numDetections; ++i) {
            float score = detectionOut.at(Float.class, i, 2).getV();
            //System.out.println(score);
            if (score > 0.5) {
                // 바운딩 박스 좌표 추출
                int classId = detectionOut.at(Float.class, i, 1).getV().intValue();
                int left = (int)(frame.cols() * detectionOut.at(Float.class, i, 3).getV());
                int top = (int)(frame.rows() * detectionOut.at(Float.class, i, 4).getV());
                int right = (int)(frame.cols() * detectionOut.at(Float.class, i, 5).getV());
                int bottom = (int)(frame.rows() * detectionOut.at(Float.class, i, 6).getV());

                // 바운딩 박스 적용
                Rect box = new Rect(left, top, right - left, bottom - top);
                
                Imgproc.rectangle(frame, box.tl(), box.br(), new Scalar(0, 255, 0), 2);
System.out.println(String.format("원래: %d/%d/%d/%d/%f/%s", left,top, right,bottom, score, CocoLabel.getCocoLabel().get(classId)));

float fVal = (float)CvUtils.getMatVal(detectionOut, i, 1);
Mat mask = new Mat(detectionMasks.size(2), detectionMasks.size(3), CvType.CV_32F, new Scalar(fVal));
                
                // 마스크 적용
                //Mat mask = detectionMasks.row(i+1).rowRange(0, classId);
                Imgproc.resize(mask, mask, new Size(right - left, bottom - top));
                //Imgproc.applyColorMap(mask, mask, Imgproc.COLORMAP_JET);
                mask.copyTo(frame.rowRange(top, bottom).colRange(left, right));
            }
        }

        // 결과 이미지 출력
        Imgcodecs.imwrite("C:\\Temp\\output_image.jpg", frame);
        System.out.println("Result saved as output.jpg");
    }
}
   