package com.tobee.opencv.detector.tests;

import java.util.ArrayList;
import java.util.List;

import org.opencv.core.Core;
import org.opencv.core.CvType;
import org.opencv.core.Mat;
import org.opencv.core.MatOfPoint;
import org.opencv.core.Point;
import org.opencv.core.Rect;
import org.opencv.core.Scalar;
import org.opencv.core.Size;
import org.opencv.dnn.Net;
import org.opencv.highgui.HighGui;
import org.opencv.imgcodecs.Imgcodecs;
import org.opencv.imgproc.Imgproc;
import org.opencv.videoio.VideoCapture;
import org.opencv.videoio.VideoWriter;
import org.opencv.videoio.Videoio;

public class TestMaskRCNNNewStyleCode {
	private static double confThreshold = 0.5; // Confidence threshold
	private static double maskThreshold = 0.3; // Mask threshold
	
	private static List<Scalar> colors;
	private static List<String> classes;
	
	public static void mainMaskRCNNStyleNew(String[] args) {
		org.apache.commons.cli.CommandLineParser parser = new org.apache.commons.cli.DefaultParser();
		org.apache.commons.cli.HelpFormatter formatter = new org.apache.commons.cli.HelpFormatter();
		org.apache.commons.cli.CommandLine cmd = null;// not a good practice, it serves it purpose
		org.apache.commons.cli.Options options = TestUtils.setupOptions4MskRCNN();
		
		try {
			cmd = parser.parse(options, args);
		} catch (org.apache.commons.cli.ParseException e) {
			System.out.println(e.getMessage());
			formatter.printHelp("MaskRCNN2", options);

			System.exit(1);
		}

		if (cmd.hasOption("help"))
		{
			String header = "Use this script to run object detection using mask-rcnn in OpenCV\n\n";
			String footer = "\nPlease report issues at http://example.com/issues";

			// HelpFormatter formatter = new HelpFormatter();
			formatter.printHelp("MaskRCNN2", header, options, footer, true);
			return;
		}
		
		List<String> colorStringList = CvUtils.getStringListOnFile(TestUtils.COLOR_FILE);
		classes = CvUtils.getStringListOnFile(TestUtils.CLASSES_FILE);
		String[] COLORS = null;
		colors = new ArrayList<Scalar>();
		String modelFile = TestUtils.MODEL_WEIGHTS_MASK_RCNN; 
		String graphFile = TestUtils.TEXT_GRAPH;
		
		if (cmd.hasOption("color"))
		{
			String colorFile = cmd.getOptionValue("color");
			
			colorStringList = CvUtils.getStringListOnFile(colorFile);
			COLORS = colorStringList.toArray(new String[colorStringList.size()]);
			
			for(String color: COLORS)
			{
				String[] rgb = color.split(" ");
				
				Double r = Double.parseDouble(rgb[0]);
				Double g = Double.parseDouble(rgb[1]);
				Double b = Double.parseDouble(rgb[2]);
				
				colors.add(new Scalar(r,g,b,255));
				
				r = g = b = null;
				rgb = null;
			}
			
			return;
		}
		
		if (cmd.hasOption("label"))
		{
			String labelFile = cmd.getOptionValue("label");
			
			classes = CvUtils.getStringListOnFile(labelFile);
			
			return;
		}
		
		if (cmd.hasOption("model"))
		{
			modelFile = cmd.getOptionValue("model");
			
			return;
		}
		
		if (cmd.hasOption("graph"))
		{
			graphFile = cmd.getOptionValue("graph");
			
			return;
		}
		
		int cameraDevice = args.length > 2 ? Integer.parseInt(args[2]) : 0;
		
		
		// Load the network
		Net net 
			= org.opencv.dnn.Dnn.readNetFromTensorflow(modelFile, graphFile);
		
		if (cmd.hasOption("deivce"))
		{
			String device = cmd.getOptionValue("deivce");
			
			switch(device)
			{
			case "cpu": 
				System.out.println( "Using CPU device");
				net.setPreferableTarget(org.opencv.dnn.Dnn.DNN_TARGET_CPU);
				break;
			case "gpu": 
				System.out.println( "Using GPU device");
				net.setPreferableBackend(org.opencv.dnn.Dnn.DNN_BACKEND_CUDA);
                                net.setPreferableTarget(org.opencv.dnn.Dnn.DNN_TARGET_CUDA);
				break;
			}
		}
		
		VideoCapture cap = null;
		VideoWriter video = null;
		String outputFile = "mask_rcnn_out_cpp.avi";
		String kWinName = "Deep learning object detection in OpenCV";
		Mat frame = null;
		
		
		if (cmd.hasOption("image"))
		{
			outputFile  = cmd.getOptionValue("image");
			
			//isOpen = cap.open(outputFile);
			//System.out.println( "Using image source. open ? " + isOpen + ":" + outputFile);
			//frame = Imgcodecs.imread(outputFile, Imgcodecs.IMREAD_COLOR);
			//frame = Imgcodecs.imread(outputFile, Imgcodecs.IMREAD_GRAYSCALE);
			//frame = new Mat();
			//frame =new Mat(480,640,CvType.CV_8UC3);
			frame = Imgcodecs.imread(outputFile);
			//Imgproc.resize(frame, frame, new Size(480,640));
			//frame = CvUtils.reSizeCommon(frame, 2);
//			cap.open(outputFile);
//			
//			if (!cap.isOpened()) {
//	            System.err.println(">>>> Error! Camera can't be opened!");
//	            return;
//	        }
			
			HighGui.namedWindow(kWinName, HighGui.WINDOW_NORMAL);
			//frame.convertTo(frame, CvType.CV_8U, 255);
			//frame.convertTo(frame, CvType.CV_8UC3, 255);
			outputFile = outputFile.substring(0, outputFile.length() - 4) + "_mask_rcnn_out.jpg";
			System.out.println( "Using video source. open ? " + outputFile);
			HighGui.imshow(kWinName, frame);
		}
		
		else if (cmd.hasOption("video"))
		{
			cap = new VideoCapture();
			video = new VideoWriter();
			
			outputFile = cmd.getOptionValue("video");
			// Create a window
                        HighGui.namedWindow(kWinName, HighGui.WINDOW_NORMAL);
			frame = new Mat();
			//isOpen = cap.open(outputFile);
			
			outputFile = outputFile.substring(0, outputFile.length() - 4) + "_mask_rcnn_out.avi";
			System.out.println( "Using video source. open ? " + outputFile);
		}
		else
		{
			cap.open(cameraDevice);
		}
		
		if (!cmd.hasOption("image"))
		{
                    video.open(outputFile, VideoWriter.fourcc('M', 'J', 'P', 'G'), 28, new Size(cap.get(Videoio.CAP_PROP_FRAME_WIDTH), cap.get(Videoio.CAP_PROP_FRAME_HEIGHT)));
		}
		
		
	    Mat blob = null;
	    
	    while (true)
	    {
	    	if (!cmd.hasOption("image"))
	    	{
	    		// get frame from the video
		    	cap.read(frame);
		        //cap >> frame;
	    	}
	       
	        // Stop the program if reached end of video
	        if (frame.empty()) {
	            System.out.println( "Done processing !!!");
	            System.out.println( "Output file is stored as " + outputFile);
	            HighGui.waitKey(3000);
	            break;
	        }
	        // Create a 4D blob from a frame.
	        blob = org.opencv.dnn.Dnn.blobFromImage(frame, 1.0, new Size(frame.cols(), frame.rows()), Scalar.all(0), true, false);
	        
	        //Sets the input to the network
	        net.setInput(blob);

	        // Runs the forward pass to get output from the output layers
	        List<String> outNames = new ArrayList<String>(2);
	        outNames.add(0, "detection_out_final");
	        outNames.add(1, "detection_masks");
	        
	        List<Mat> outs = new ArrayList<Mat>();
	        net.forward(outs, outNames);

	        // Extract the bounding box and mask for each of the detected objects
	        postprocess(frame, outs);

	        // Put efficiency information. The function getPerfProfile returns the overall time for inference(t) and the timings for each of the layers(in layersTimes)
	        org.opencv.core.MatOfDouble layersTimes = new org.opencv.core.MatOfDouble();
	        double freq = Core.getTickFrequency() / 1000;
	        double t = net.getPerfProfile(layersTimes) / freq;
	        String label = String.format("Mask-RCNN, Inference time for a frame : %f ms", t);
	        Imgproc.putText(frame, label, new Point(0, 15), 
	        		Imgproc.FONT_HERSHEY_SIMPLEX, 0.5, new Scalar(0, 0, 0));

	        // Write the frame with the detection boxes
	        Mat detectedFrame = new Mat();
	        frame.convertTo(detectedFrame, CvType.CV_8U);
	        
	        if (cmd.hasOption("image")) {
	        	System.out.println("-->" + outputFile);
	        	Imgcodecs.imwrite(outputFile, detectedFrame);
		    }
	        else
	        {
	        	video.write(detectedFrame);
	        }
	        
	        HighGui.imshow(kWinName, frame);
	        
	        if(HighGui.waitKey(1000) < 0) {
	        	blob.release();
	        	detectedFrame.release();
	        	layersTimes.release();
	        	layersTimes = null;
	        	detectedFrame = null;
	        	outNames.clear();
	        	outNames = null;
	        	blob = null;
	        	HighGui.destroyWindow(kWinName);
	        	break;
	        }
	    }
	    HighGui.destroyAllWindows();
	    
	     
	    
	}
	
	// For each frame, extract the bounding box and mask for each detected object
	private static Mat postprocess(Mat frame, final List<Mat> outs) {
		Mat outDetections = outs.get(0);
		Mat outMasks = outs.get(1);

		// Output size of masks is NxCxHxW where
		// N - number of detected boxes
		// C - number of classes (excluding background)
		// HxW - segmentation shape
		final int numDetections = outDetections.size(2);
		//final int numClasses = outMasks.size(1);

		outDetections = outDetections.reshape(1, (int) outDetections.total() / 7);
		
		for (int i = 0; i < numDetections; ++i) {
			// float score = outDetections.at<float>(i, 2);
			// Atable<Float> s = outDetections.at(Float.class, i,2);
			float score = outDetections.at(Float.class, i, 2).getV();
			//score = (float)getMatVal(outDetections, i, 2);
			
			if (score > confThreshold) {
				// Extract the bounding box
//	            int classId = static_cast<int>(outDetections.at<float>(i, 1));
//	            int left = static_cast<int>(frame.cols * outDetections.at<float>(i, 3));
//	            int top = static_cast<int>(frame.rows * outDetections.at<float>(i, 4));
//	            int right = static_cast<int>(frame.cols * outDetections.at<float>(i, 5));
//	            int bottom = static_cast<int>(frame.rows * outDetections.at<float>(i, 6));
				
				int classId = outDetections.at(Float.class, i, 1).getV().intValue();
				int left = (int)(frame.cols() * outDetections.at(Float.class, i, 3).getV());
				int top = (int)(frame.rows() * outDetections.at(Float.class, i, 4).getV());
				int right = (int)(frame.cols() * outDetections.at(Float.class, i, 5).getV());
				int bottom = (int)(frame.rows() * outDetections.at(Float.class, i, 6).getV());

				left = Math.max(0, Math.min(left, frame.cols() - 1));
				top = Math.max(0, Math.min(top, frame.rows() - 1));
				right = Math.max(0, Math.min(right, frame.cols() - 1));
				bottom = Math.max(0, Math.min(bottom, frame.rows() - 1));
				Rect box = new Rect(left, top, right - left + 1, bottom - top + 1);
				
				// Extract the mask for the object
				// https://stackoverflow.com/questions/50325025/opencv-ptr-in-java
				// Mat objectMask(outMasks.size[2], outMasks.size[3], CV_32F, outMasks.ptr<float>(i, classId));
				//int aa[] = new int[outDetections.channels()];
				//outDetections.get(i, 1, aa);
				
				//float[] values = new float[outDetections.channels()];
				//outDetections.get(i, 1, values);
				
				float fVal = (float)CvUtils.getMatVal(outDetections, i, 1);
				
				//Mat objectMask = new Mat(new int[] { outMasks.size(2), outMasks.size(3) }, CvType.CV_32F,
				//		new Scalar(values[0]));
				
				Mat objectMask = new Mat(outMasks.size(2), outMasks.size(3), CvType.CV_32F, new Scalar(fVal));

				// Draw bounding box, colorize and show the mask on the image
				drawBox(frame, classId, score, box, objectMask);
				
				objectMask.release(); objectMask = null;

			}
		}
		
		
		outDetections.release(); outDetections = null;
		outMasks.release(); outMasks = null;
		
		return frame;
	}

	// Draw the predicted bounding box, colorize and show the mask on the image
	private static void drawBox(Mat frame, int classId, double conf, Rect box, Mat objectMask) {
		// Draw a rectangle displaying the bounding box
		Imgproc.rectangle(frame, new Point(box.x, box.y), new Point(box.x + box.width, box.y + box.height),
				new Scalar(255, 178, 50), 3);

		// Get the label for the class name and its confidence
		String label = String.format("%.2f", conf);

		if (!classes.isEmpty()) {
//		        CV_Assert(classId < (int)classes.size());
			label = classes.get(classId) + ":" + label;
		}
		
		//System.out.println("label : " + label);
		
		// Display the label at the top of the bounding box
		// int baseLine[] = new int[1];
		int baseLine[] = { 0 };
		Size labelSize = Imgproc.getTextSize(label, Imgproc.FONT_HERSHEY_SIMPLEX, 0.5, 1, baseLine);
		box.y = Integer.max(box.y, (int) labelSize.height);

		Imgproc.rectangle(
				frame, new Point(box.x, box.y - Math.round(1.5 * labelSize.height)),
				new Point(box.x + Math.round(1.5 * labelSize.width), box.y + baseLine[0]), new Scalar(255, 255, 255),
				Imgproc.FILLED);
		Imgproc.putText(frame, label, new Point(box.x, box.y), 
				Imgproc.FONT_HERSHEY_SIMPLEX, 0.75, new Scalar(0, 0, 0),1);

		Scalar color = colors.get(classId % colors.size());
		// Comment the above line and uncomment the two lines below to generate
		// different instance colors
		//int colorInd = (int) (Math.random() % colors.size());
		// Scalar color = colors[colorInd];

		// Resize the mask, threshold, color and apply it on the image
		Imgproc.resize(objectMask, objectMask, new Size(box.width, box.height));

		// ~~~~~~~~~~~~~~
		// Mat mask = (objectMask > maskThreshold);
		Mat mask = new Mat();
		objectMask.setTo(new Scalar(maskThreshold), mask);

		// ~~~~~~~~~~~~~~
		Mat image_roi = new Mat(frame, box); // frame(box)
		//image_roi.release();
		// ~~~~~~~~~~~~~~
		// Mat coloredRoi = new Mat(0.3 * color + 0.7 * image_roi);

		// 0.3 * color
		Scalar s2 = new Scalar(0.3);
		Scalar s3 = color.mul(s2);

		// 0.7 * image_roi
		Mat image_roi2 = new Mat();
		Scalar s = new Scalar(0.7);
		Core.multiply(image_roi, s, image_roi2);

		Mat coloredRoi = new Mat();
		Core.add(image_roi2, s3, coloredRoi);

		// Mat coloredRoi = new Mat(0.3 * color + 0.7 * image_roi);
		coloredRoi.convertTo(coloredRoi, CvType.CV_8UC3);

		// Draw the contours on the image
		List<MatOfPoint> contours = new ArrayList<MatOfPoint>();
		Mat hierarchy = new Mat();
		mask.convertTo(mask, CvType.CV_8U);
		Imgproc.findContours(mask, contours, hierarchy, 
				Imgproc.RETR_CCOMP, Imgproc.CHAIN_APPROX_SIMPLE);
		Imgproc.drawContours(coloredRoi, contours, -1, color, 5, 
				Imgproc.LINE_8, hierarchy, 100);
		coloredRoi.copyTo(image_roi, mask);
		
		mask.release(); mask = null;
		image_roi2.release(); image_roi2 = null;
		hierarchy.release(); hierarchy = null;
		coloredRoi.release(); coloredRoi = null;
		
		if(contours.size() > 0)
		for(MatOfPoint p: contours)
		{
			p.release();
			p = null;
		}
		
		contours.clear();
		contours = null;
	}

	
}
