package com.tobee.opencv.detector.tests;

import java.io.File;
import java.util.ArrayList;
import java.util.List;

import org.opencv.core.Core;
import org.opencv.core.Mat;
import org.opencv.core.MatOfFloat;
import org.opencv.core.MatOfInt;
import org.opencv.core.MatOfRect2d;
import org.opencv.core.Point;
import org.opencv.core.Rect;
import org.opencv.core.Rect2d;
import org.opencv.core.Scalar;
import org.opencv.core.Size;
import org.opencv.dnn.Dnn;
import org.opencv.dnn.Net;
import org.opencv.imgcodecs.Imgcodecs;
import org.opencv.imgproc.Imgproc;
import org.opencv.utils.Converters;

public class TestYolov3 {
	private static List<String> getOutputNames(Net net) {
		List<String> names = new ArrayList<>();

		List<Integer> outLayers = net.getUnconnectedOutLayers().toList();
		List<String> layersNames = net.getLayerNames();

		outLayers.forEach((item) -> names.add(layersNames.get(item - 1)));// unfold and create R-CNN layers from the
																			// loaded YOLO model//
		return names;
	}

	public static void mainYolov3(String[] args) throws InterruptedException {
		org.apache.commons.cli.CommandLineParser parser = new org.apache.commons.cli.DefaultParser();
		org.apache.commons.cli.HelpFormatter formatter = new org.apache.commons.cli.HelpFormatter();
		org.apache.commons.cli.CommandLine cmd = null;// not a good practice, it serves it purpose
		org.apache.commons.cli.Options options = TestUtils.setupOptions4YoloV3();
		
		try {
			cmd = parser.parse(options, args);
		} catch (org.apache.commons.cli.ParseException e) {
			System.out.println(e.getMessage());
			formatter.printHelp("Yolov3", options);

			System.exit(1);
		}

		if (cmd.hasOption("help"))
		{
			String header = "Use this script to run object detection using yolo3 model in OpenCV\n\n";
			String footer = "\nPlease report issues at http://example.com/issues";

			// HelpFormatter formatter = new HelpFormatter();
			formatter.printHelp("MaskRCNN2", header, options, footer, true);
			return;
		}
		
		// Download and  load only wights for YOLO , this  is obtained from official YOLO site//
		String modelWeights = "/Users/ds/projetos/core_surveillance/master/javacvtools/yolov3.weights"; 
		// Download and  load cfg file for YOLO , can be obtained from official site//
		String modelConfiguration = "/Users/ds/projetos/core_surveillance/master/javacvtools/yolov3.cfg";
		
		String imageFile = "";
		String outputFile = "";
		
		if (cmd.hasOption("image"))
		{
			imageFile  = cmd.getOptionValue("image");
			
			outputFile = imageFile.substring(0, imageFile.length() - 4) + "_yolov3_out.jpg";
		}
		
		if (cmd.hasOption("model"))
		{
			modelWeights  = cmd.getOptionValue("model");
		}
		
		if (cmd.hasOption("config"))
		{
			modelConfiguration  = cmd.getOptionValue("config");
		}
		
		boolean isFileExists = true;
		
		if(!new File(imageFile).exists())
		{
			System.out.println(String.format("File Not Found %s", imageFile));
			isFileExists = false;
		}
		
		if(!new File(modelWeights).exists())
		{
			System.out.println(String.format("File Not Found %s", modelWeights));
			isFileExists = false;
		}
		
		if(!new File(modelConfiguration).exists())
		{
			System.out.println(String.format("File Not Found %s", modelConfiguration));
			isFileExists = false;
		}
		
		if(!isFileExists) return;
		
		// OpenCV DNN supports models trained from  various frameworks like Caffe and TensorFlow. 
		//It also supports various networks architectures based on YOLO//
		Net net = Dnn.readNetFromDarknet(modelConfiguration, modelWeights);
		// Thread.sleep(5000);

		// Mat image =
		// Imgcodecs.imread("D:\\yolo-object-detection\\yolo-object-detection\\images\\soccer.jpg");
		Size sz = new Size(416, 416);

		List<Mat> result = new ArrayList<>();
		List<String> outBlobNames = getOutputNames(net);

		// while (true) {
		// if (cap.read(frame)) {
		Mat src = Imgcodecs.imread(imageFile);
		
		// We feed one frame of video into  the network at a time, we have to convert the image to a blob. 
		// A blob is a pre-processed image that serves as the input.//
		Mat blob = Dnn.blobFromImage(src, 0.00392, sz, new Scalar(0), true, false); 
		net.setInput(blob);
		net.forward(result, outBlobNames); // Feed forward the model to get output //

		// outBlobNames.forEach(System.out::println);
		// result.forEach(System.out::println);
		float confThreshold = 0.5f; // Insert thresholding beyond which the model will detect objects//
		List<Integer> clsIds = new ArrayList<>();
		List<Float> confs = new ArrayList<>();
		List<Rect2d> rects = new ArrayList<Rect2d>();
		for (int i = 0; i < result.size(); ++i) {
			// each row is a candidate detection, the 1st 4 numbers are
			// [center_x, center_y, width, height], followed by (N-4) class probabilities
			Mat level = result.get(i);
			for (int j = 0; j < level.rows(); ++j) {
				Mat row = level.row(j);
				Mat scores = row.colRange(5, level.cols());
				Core.MinMaxLocResult mm = Core.minMaxLoc(scores);
				float confidence = (float) mm.maxVal;
				Point classIdPoint = mm.maxLoc;
				if (confidence > confThreshold) {
					int centerX = (int) (row.get(0, 0)[0] * src.cols()); // scaling for drawing the bounding boxes//
					int centerY = (int) (row.get(0, 1)[0] * src.rows());
					int width = (int) (row.get(0, 2)[0] * src.cols());
					int height = (int) (row.get(0, 3)[0] * src.rows());
					int left = centerX - width / 2;
					int top = centerY - height / 2;

					clsIds.add((int) classIdPoint.x);
					confs.add((float) confidence);
					rects.add(new Rect2d(left, top, width, height));
				}
			}
		}
		float nmsThresh = 0.5f;
		MatOfFloat confidences = new MatOfFloat(Converters.vector_float_to_Mat(confs));
		Rect2d[] boxesArray = rects.toArray(new Rect2d[0]);
		MatOfRect2d boxes = new MatOfRect2d(boxesArray);
		MatOfInt indices = new MatOfInt();
		
		// We draw the bounding boxes for objects  here//
		Dnn.NMSBoxes(boxes, confidences, confThreshold, nmsThresh, indices); 

		int[] ind = indices.toArray();
		int j = 0;
		for (int i = 0; i < ind.length; ++i) {
			int idx = ind[i];
			Rect2d box = boxesArray[idx];
			if (clsIds.get(i) == 0) {
				Imgproc.rectangle(src, box.tl(), box.br(), new Scalar(0, 0, 255), 2);
			} else if (clsIds.get(i) == 27) {
				Imgproc.rectangle(src, box.tl(), box.br(), new Scalar(0, 255, 0), 2);
			} else {
				Imgproc.rectangle(src, box.tl(), box.br(), new Scalar(255, 0, 0), 2);
			}
			// i=j;
//            System.out.println(idx);
			System.out.println(clsIds.get(i) + " | " + confs.get(i) + "%" + " | " + idx);

		}
		
		Imgcodecs.imwrite(outputFile, src);
	}

}
