package com.tobee.opencv.detector.tests;

import conf.CocoLabel;
import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.io.File;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JComboBox;
import javax.swing.JDialog;
import javax.swing.JFileChooser;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JTextField;
import javax.swing.filechooser.FileNameExtensionFilter;
import org.opencv.core.Core;

import org.opencv.core.CvType;
import org.opencv.core.Mat;
import org.opencv.core.MatOfFloat;
import org.opencv.core.MatOfInt;
import org.opencv.core.MatOfRect2d;
import org.opencv.core.Point;
import org.opencv.core.Rect2d;
import org.opencv.core.Scalar;
import org.opencv.core.Size;
import org.opencv.dnn.Dnn;
import org.opencv.dnn.Net;
import org.opencv.highgui.HighGui;
import org.opencv.imgcodecs.Imgcodecs;
import org.opencv.imgproc.Imgproc;
import org.opencv.utils.Converters;
import org.opencv.videoio.VideoCapture;
import org.opencv.videoio.VideoWriter;
import org.opencv.videoio.Videoio;

//https://www.linkedin.com/pulse/algorithm-can-learn-neural-network-20-lines-code-svetozar-radoj%C4%8Din/
//https://www.linkedin.com/pulse/opencv-java-yolo-object-detection-images-svetozar-radoj%C4%8Din
public class  TestYolov4_Tobee2{
    private static final String THIS_NAME = TestYolov4_Tobee2.class.getSimpleName();
    static{
        System.loadLibrary(Core.NATIVE_LIBRARY_NAME);
    }

    private static void detectObjectOnImage(final String winName, final Net dnnNet, List<String> outputLayers, final String mediaFile) {
        Mat frame = Imgcodecs.imread(mediaFile);
        String outputFile = mediaFile.substring(0, mediaFile.length() - 4) + "_yolov4_out.jpg";
        Map<Integer, String> cocoLabels =  CocoLabel.getCocoLabel();
        
        // generate radnom color in order to draw bounding boxes
        List<Scalar> colors =CocoLabel.getColorsByLabel();
        
        HashMap<String, List> result = forwardImageOverNetwork(frame, dnnNet, outputLayers);

        ArrayList<Rect2d> boxes = (ArrayList<Rect2d>) result.get("boxes");
        ArrayList<Float> confidences = (ArrayList<Float>) result.get("confidences");
        ArrayList<Integer> class_ids = (ArrayList<Integer>) result.get("class_ids");

        // -- Now , do so-called “non-maxima suppression”
        // Non-maximum suppression is performed on the boxes whose confidence is equal
        // to or greater than the threshold.
        // This will reduce the number of overlapping boxes:
        MatOfInt indices = getBBoxIndicesFromNonMaximumSuppression(boxes, confidences);
        // -- Finally, go over indices in order to draw bounding boxes on the image:
        frame = drawBoxesOnTheImage(frame, indices, boxes, cocoLabels, class_ids, colors);
        //vwriter.write(frame);
        //System.out.println("===============");
        HighGui.imshow(winName, frame);

        Mat detectedFrame = new Mat();
        frame.convertTo(detectedFrame, CvType.CV_8U);
        Imgcodecs.imwrite(outputFile, detectedFrame);
        
        //CvUtils.showWindow(THIS_NAME, CvUtils.matToBufferedImage(CvUtils.reSizeCommon(Imgcodecs.imread(outputFile), 2)));
        CvUtils.showWindow(THIS_NAME, CvUtils.matToBufferedImage(Imgcodecs.imread(outputFile)));
    }
    
    private static void detectObjectOnVideo(
        final String winName, final Net dnnNet, List<String> outputLayers, 
        final String mediaFile, final boolean isSaveOuput
    ) 
    {
        Mat frame;
        boolean isOpen;
        double frmCount;
        
        // load our video:
        VideoCapture capture = new VideoCapture();
        VideoWriter vwriter = new VideoWriter();
        double fps = capture.get(Videoio.CAP_PROP_FPS);
                
        if(isSaveOuput)
        {
            String outputFile = mediaFile.substring(0, mediaFile.length() - 4) + "_yolov4_out.mp4";
            
            Size size 
                = new Size(capture.get(Videoio.CAP_PROP_FRAME_WIDTH), capture.get(Videoio.CAP_PROP_FRAME_HEIGHT));
            //vwriter = new VideoWriter(outputFile, VideoWriter.fourcc('m', 'p', 'g', '4'), fps, size, true);
            
            vwriter.open(outputFile, VideoWriter.fourcc('M', 'J', 'P', 'G'), fps, size);
        }
        
        Map<Integer, String> cocoLabels =  CocoLabel.getCocoLabel();
        // generate radnom color in order to draw bounding boxes
        List<Scalar> colors =CocoLabel.getColorsByLabel();
        System.out.println(String.format("cocoLabels[%d]/coldor[%d]", cocoLabels.size(), colors.size()));
        // Create a window
        HighGui.namedWindow(winName, HighGui.WINDOW_NORMAL);
        frame = new Mat();
        isOpen = capture.open(mediaFile);

        if (isOpen) {
            frmCount = capture.get(Videoio.CAP_PROP_FRAME_COUNT);
            System.out.println("frmCount = " + frmCount);
        } else {
            System.out.println("Capture is not opened!");
            return;
        }

        capture.read(frame);
        HighGui.imshow(winName, frame);
        
        //vwriter.open(outputFile, VideoWriter.fourcc('M', 'J', 'P', 'G'), 28, size);
        //vwriter.open(outputFile, VideoWriter.fourcc('M', 'J', 'P', 'G'), fps, size);
        HashMap<String, List> result;
        List<Rect2d> bboxes;
        List<Float> confidences;
        List<Integer> classIdList;
        MatOfInt indices;
        
        while (true) {
            if(HighGui.waitKey(100) == 27) //ESC key to exit
            {
                System.out.println( "[ESC]Done processing !!!");
                //System.out.println( "[ESC]Output file is stored as " + outputFile);
                HighGui.waitKey(3000);
                break;
            }
             // get frame from the video
            capture.read(frame);
  
            // Stop the program if reached end of video
            if (frame.empty()) {
                System.out.println( "Done processing !!!");
                //System.out.println( "Output file is stored as " + outputFile);
                HighGui.waitKey(3000);
                if(!isSaveOuput) {
                } else {
                    vwriter.release();
                }
                vwriter = null;
                break;
            }

            result = forwardImageOverNetwork(frame, dnnNet, outputLayers);

            bboxes = (List<Rect2d>) result.get("boxes");
            confidences = (List<Float>) result.get("confidences");
            classIdList = (List<Integer>) result.get("class_ids");
            
            // -- Now , do so-called “non-maxima suppression”
            // Non-maximum suppression is performed on the boxes whose confidence is equal
            // to or greater than the threshold.
            // This will reduce the number of overlapping boxes:
            indices = getBBoxIndicesFromNonMaximumSuppression(bboxes, confidences);
            // -- Finally, go over indices in order to draw bounding boxes on the image:
            frame = drawBoxesOnTheImage(frame, indices, bboxes, cocoLabels, classIdList, colors);
            vwriter.write(frame);
            if(isSaveOuput) 
            {
//                Mat detectedFrame = new Mat();
//                frame.convertTo(detectedFrame, CvType.CV_8U);
//                vwriter.write(detectedFrame);
//                System.out.println("===============");
            } 
            
            //
            HighGui.imshow(winName, frame);
            
            //HighGui.imshow(winName, frame);

            if(!bboxes.isEmpty()) {bboxes.clear(); bboxes = null;}
            if(!confidences.isEmpty()) {confidences.clear(); confidences = null;}
            if(!classIdList.isEmpty()) {classIdList.clear(); classIdList = null;}
            if(!indices.empty()) {indices.release();indices = null;}
        } // of main while()
        capture.release();
        if(vwriter!= null)  vwriter.release();

        HighGui.destroyAllWindows();
    }
    
        
    private static final class YoloOption
    {
        static String mediaType;
        static String deviceType;

        static String mediaFile;
        static String modelWeights;
        static String modelConfiguration;
        static String WinName = "Object detection in OpenCV Yolov4";
        static boolean isSaveVideoOutput;
        
        public static boolean isValid()
        {
            boolean isValid;

            isValid = (mediaFile == null ? false : (mediaFile.isBlank() ? false : (!mediaFile.isEmpty())));
            System.out.println(isValid);
            //System.out.println(mediaFile != null && !mediaFile.isBlank() && mediaFile.isEmpty());
            if(!isValid) return false;
            isValid = mediaFile == null ? false : (mediaFile.isBlank() ? false : (!mediaFile.isEmpty()));
            System.out.println(isValid);
            if(!isValid) return false;
            isValid = modelWeights == null ? false : (modelWeights.isBlank() ? false : (!modelWeights.isEmpty()));
            System.out.println(isValid);
            if(!isValid) return false;
            isValid = modelConfiguration == null ? false : (modelConfiguration.isBlank() ? false : (!modelConfiguration.isEmpty()));
            System.out.println(isValid);
            if(!isValid) return false;

            return isValid;
        }
            
        public static boolean setUpOptions() {
            // Create main panel
            JPanel panel = new JPanel();
            panel.setPreferredSize(new Dimension(360,250));
            panel.setLayout(new GridBagLayout());
            GridBagConstraints gbc = new GridBagConstraints();
            gbc.insets = new Insets(5, 5, 5, 5);
            gbc.fill = GridBagConstraints.HORIZONTAL;
            gbc.gridx = 0;
            gbc.gridy = 0;

            // Create a JComboBox group
            JPanel comboBoxGroup = new JPanel(new FlowLayout(FlowLayout.LEFT, 5, 0));
            JComboBox<String> comboMediaType = new JComboBox<>(new String[]{"None", "Image", "Video"});
            JComboBox<String> comboDeviceType = new JComboBox<>(new String[]{"cpu", "gpu"});
            comboBoxGroup.add(comboMediaType);
            comboBoxGroup.add(comboDeviceType);
            comboMediaType.setPreferredSize(new Dimension(150,30));
            comboDeviceType.setPreferredSize(new Dimension(150, 30));
            panel.add(comboBoxGroup, gbc);
            comboBoxGroup.setPreferredSize(new Dimension(320,40));

            // Create  JTextField + JButton Group
            gbc.gridy++;
            JPanel group1 = new JPanel(new FlowLayout(FlowLayout.LEFT, 5, 0));
            JPanel ingroup1 = new JPanel(new BorderLayout());
            JLabel lblWeights = new JLabel("Weights: ");
            JTextField txtWeights = new JTextField(10);
            JButton btnFind1 = new JButton("Find");

            txtWeights.setPreferredSize(new Dimension(200,30));
            btnFind1.setPreferredSize(new Dimension(60,30));

            ingroup1.add(lblWeights,BorderLayout.WEST);
            ingroup1.add(txtWeights,BorderLayout.CENTER);
            ingroup1.add(btnFind1,BorderLayout.EAST);
            ingroup1.setPreferredSize(new Dimension(310,30));

            group1.setPreferredSize(new Dimension(350,40));
            group1.add(ingroup1);

            panel.add(group1, gbc);

            // Create  JTextField + JButton Group
            gbc.gridy++;
            JPanel group2 = new JPanel(new FlowLayout(FlowLayout.LEFT, 5, 0));
            JPanel ingroup2 = new JPanel(new BorderLayout());
            group2.setPreferredSize(new Dimension(100,100));
            JLabel lblConfig = new JLabel("Config: ");
            JTextField txtConfig = new JTextField(10);
            JButton btnFind2 = new JButton("Find");

            txtConfig.setPreferredSize(new Dimension(200,30));
            btnFind2.setPreferredSize(new Dimension(60,30));

            ingroup2.add(lblConfig,BorderLayout.WEST);
            ingroup2.add(txtConfig,BorderLayout.CENTER);
            ingroup2.add(btnFind2,BorderLayout.EAST);
            ingroup2.setPreferredSize(new Dimension(310,30));

            group2.setPreferredSize(new Dimension(350,40));
            group2.add(ingroup2);

            panel.add(group2, gbc);

            // Create  JTextField + JButton Group 
            gbc.gridy++;
            JPanel group3 = new JPanel(new FlowLayout(FlowLayout.LEFT, 5, 0));
            JPanel ingroup3 = new JPanel(new BorderLayout());
            group3.setPreferredSize(new Dimension(100,100));
            JLabel lblFileIn = new JLabel("InFile: ");
            JTextField txtFileIn = new JTextField(10);
            JButton btnFind3 = new JButton("Find");

            txtFileIn.setPreferredSize(new Dimension(200,30));
            btnFind3.setPreferredSize(new Dimension(60,30));

            ingroup3.add(lblFileIn,BorderLayout.WEST);
            ingroup3.add(txtFileIn,BorderLayout.CENTER);
            ingroup3.add(btnFind3,BorderLayout.EAST);
            ingroup3.setPreferredSize(new Dimension(310,30));

            group3.setPreferredSize(new Dimension(350,40));
            group3.add(ingroup3);

            panel.add(group3, gbc);
            
            // Create  JLabel + JCheckBox Group 
            gbc.gridy++;
            JPanel group4 = new JPanel(new FlowLayout(FlowLayout.LEFT, 5, 0));
            JPanel ingroup4 = new JPanel(new BorderLayout());
            group4.setPreferredSize(new Dimension(100,100));
            JLabel lblChkSaveVdo = new JLabel("Save Videio: ");
            JCheckBox chkSaveVdo = new JCheckBox();
            chkSaveVdo.setEnabled(false);
            
            lblChkSaveVdo.setPreferredSize(new Dimension(100,30));
            
            ingroup4.add(lblChkSaveVdo,BorderLayout.CENTER);
            ingroup4.add(chkSaveVdo,BorderLayout.EAST);
            ingroup4.setPreferredSize(new Dimension(310,30));

            group4.setPreferredSize(new Dimension(350,40));
            group4.add(ingroup4);

            panel.add(group4, gbc);

            comboMediaType.addActionListener((ActionEvent e) -> {
                mediaType = (String) comboMediaType.getSelectedItem();
            });
            
            comboMediaType.addItemListener((ItemEvent e) -> {
                if (e.getStateChange() == ItemEvent.SELECTED) {
                    if(e.getItem() == "Video")
                    {
                        chkSaveVdo.setEnabled(true);
                    }
                    else
                    {
                        chkSaveVdo.setEnabled(false);
                    }
                }
            });
            
            comboDeviceType.addActionListener((ActionEvent e) -> {
                deviceType = (String) comboDeviceType.getSelectedItem();
            });

            btnFind1.addActionListener((ActionEvent e) -> {
                JFileChooser fileChooser = new JFileChooser();
                fileChooser.setCurrentDirectory(new File("C:\\"));
                int returnValue = fileChooser.showOpenDialog(null);
                if (returnValue == JFileChooser.APPROVE_OPTION) {
                    File selectedFile = fileChooser.getSelectedFile();
                    txtWeights.setText(selectedFile.getAbsolutePath());
                }
            });

            btnFind2.addActionListener((ActionEvent e) -> {
                JFileChooser fileChooser = new JFileChooser();
                fileChooser.setCurrentDirectory(new File("C:\\"));
                int returnValue = fileChooser.showOpenDialog(null);
                if (returnValue == JFileChooser.APPROVE_OPTION) {
                    File selectedFile = fileChooser.getSelectedFile();
                    
                    txtConfig.setText(selectedFile.getAbsolutePath());
                }
            });

            btnFind3.addActionListener((ActionEvent e) -> {
                JFileChooser fileChooser = new JFileChooser();
                fileChooser.setCurrentDirectory(new File("C:\\")); // Default directory
                
                String mType = comboMediaType.getSelectedItem().toString().trim();
                
                if("Image".equals(mType))
                {
                    chkSaveVdo.setEnabled(false);
                    
                    FileNameExtensionFilter filter = new FileNameExtensionFilter(
                            "Image Files (JPG, PNG)", "jpg", "jpeg", "png");
                    fileChooser.setFileFilter(filter);
                }
                else if("Video".equals(mType))
                {
                    chkSaveVdo.setEnabled(true);
                    FileNameExtensionFilter filter = new FileNameExtensionFilter(
                            "Video Files (MP4)", "mp4");
                    fileChooser.setFileFilter(filter);
                }
                
                int returnValue = fileChooser.showOpenDialog(null);
                if (returnValue == JFileChooser.APPROVE_OPTION) {
                    File selectedFile = fileChooser.getSelectedFile();
                    txtFileIn.setText(selectedFile.getAbsolutePath());
                }
            });

            // OK / CANCEL button dialog
            int result = JOptionPane.showConfirmDialog(null, panel, "OpenCV YOLO Options",
                    JOptionPane.OK_CANCEL_OPTION, JOptionPane.PLAIN_MESSAGE);
            
            if (result == JOptionPane.OK_OPTION) {
                System.out.println("OK clicked");

                mediaType = comboMediaType.getSelectedItem().toString().trim();
                deviceType = comboDeviceType.getSelectedItem().toString().trim();
                modelWeights = txtWeights.getText();
                modelConfiguration = txtConfig.getText();
                mediaFile = txtFileIn.getText();
                isSaveVideoOutput = chkSaveVdo.isSelected();
                return true;
            } else {
                //System.out.println("Cancelled");
                return false;
            }
        }
    }
    
    private static final boolean IS_DEBUG = true;
    
    public static void main(String[] args) throws Exception {
        String winName;
        String mediaType;
        String deviceType;
        String mediaFile;
        String modelWeights; 
        String modelConfiguration;
        if(!IS_DEBUG)
        {
            if(!YoloOption.setUpOptions())
            {
                JOptionPane.showMessageDialog(null, "Action canceled by user", "Info", JOptionPane.INFORMATION_MESSAGE);
                return;
            } 

            if(!YoloOption.isValid()) {
                JOptionPane.showMessageDialog(null, "Fill in all options", "error", JOptionPane.WARNING_MESSAGE);
                return;
            }
        }
 
        mediaType = YoloOption.mediaType;
        deviceType = YoloOption.deviceType;

        mediaFile = YoloOption.mediaFile;
        modelWeights = YoloOption.modelWeights; 
        modelConfiguration = YoloOption.modelConfiguration;
        boolean isSaveOuput = YoloOption.isSaveVideoOutput;
        if(IS_DEBUG)
        {
            mediaType = "Video";
            deviceType = "cpu";

//            mediaFile = "D:\\DEV\\WORK\\Vision\\data\\videos\\vtest.avi";
//            modelWeights = "D:\\DEV\\WORK\\Vision\\pretrained_model\\yolo\\v4\\yolov4.weights";
//            modelConfiguration = "D:\\DEV\\WORK\\Vision\\pretrained_model\\yolo\\v4\\yolov4.cfg";
            
            mediaFile = "D:\\DEV\\WORK\\Vision\\data\\videos\\vtest.avi";
            modelWeights = "D:\\DEV\\WORK\\Vision\\pretrained_model\\yolo\\v3\\yolov3.weights";
            modelConfiguration = "D:\\DEV\\WORK\\Vision\\pretrained_model\\yolo\\v3\\yolov3.cfg";
            
            isSaveOuput = true;
        }      
        
        winName = YoloOption.WinName;
        
        if(!new File(mediaFile).exists()) {
            JOptionPane.showMessageDialog(null, String.format("File9[%s] Not Found", mediaFile), "File Error", JOptionPane.WARNING_MESSAGE);
            return;
        }
        
        // load our YOLO object detector trained on COCO dataset
        Net dnnNet = org.opencv.dnn.Dnn.readNetFromDarknet(modelConfiguration, modelWeights);
        switch(deviceType)
        {
        case "cpu" -> { 
            System.out.println( "Using CPU device");
            dnnNet.setPreferableTarget(org.opencv.dnn.Dnn.DNN_TARGET_CPU);
            }
        case "gpu" -> { 
            System.out.println( "Using GPU device");
            dnnNet.setPreferableBackend(org.opencv.dnn.Dnn.DNN_BACKEND_CUDA);
            dnnNet.setPreferableTarget(org.opencv.dnn.Dnn.DNN_TARGET_CUDA);
            }
        }
        
        List<String> layerNames = dnnNet.getLayerNames();
        List<String> outputLayers = new ArrayList<>();
        for (Integer i : dnnNet.getUnconnectedOutLayers().toList()) {
            outputLayers.add(layerNames.get(i - 1));
        }
        
        if (mediaType.equals("Image"))
        {
            detectObjectOnImage(winName, dnnNet, outputLayers, mediaFile);
        }
        else if (mediaType.equals("Video"))
        {
            detectObjectOnVideo(winName, dnnNet, outputLayers, mediaFile, isSaveOuput);
        }
    }

    private static HashMap<String, List> forwardImageOverNetwork(Mat img, Net dnnNet, List<String> outputLayers) {
        // --We need to prepare some data structure in order to store the data returned
        // by the network
        // (ie, after Net.forward() call))
        // So, Initialize our lists of detected bounding boxes, confidences, and class
        // IDs, respectively
        // This is what this method will return:
        HashMap<String, List> result = new HashMap<>();
        result.put("boxes", new ArrayList<Rect2d>());
        result.put("confidences", new ArrayList<Float>());
        result.put("class_ids", new ArrayList<Integer>());

        // -- The input image to a neural network needs to be in a certain format called
        // a blob.
        // In this process, it scales the image pixel values to a target range of 0 to 1
        // using a scale factor of 1/255.
        // It also resizes the image to the given size of (416, 416) without cropping
        // Construct a blob from the input image and then perform a forward pass of the
        // YOLO object detector,
        // giving us our bounding boxes and associated probabilities:

        // Here we supply the spatial size that the Convolutional Neural Network
        // expects.
        Mat blob_from_image = Dnn.blobFromImage(img, 1 / 255.0, new Size(416, 416),
                        new Scalar(new double[] { 0.0, 0.0, 0.0 }), true, false);
        dnnNet.setInput(blob_from_image);

        // -- the output from network's forward() method will contain a List of OpenCV
        // Mat object, so lets prepare one
        List<Mat> outputs = new ArrayList<>();

        // -- Finally, let pass forward throught network. The main work is done here:
        dnnNet.forward(outputs, outputLayers);

        // --Each output of the network outs (ie, each row of the Mat from 'outputs')
        // is represented by a vector of the number
        // of classes + 5 elements. The first 4 elements represent center_x, center_y,
        // width and height.
        // The fifth element represents the confidence that the bounding box encloses
        // the object.
        // The remaining elements are the confidence levels (ie object types) associated
        // with each class.
        // The box is assigned to the category corresponding to the highest score of the
        // box:
        for (Mat output : outputs) {
            // loop over each of the detections. Each row is a candidate detection,
            System.out.println("Output.rows(): " + output.rows() + ", Output.cols(): " + output.cols());
            for (int i = 0; i < output.rows(); i++) {
                Mat row = output.row(i);
                List<Float> detect = new MatOfFloat(row).toList();
                List<Float> score = detect.subList(5, output.cols());
                int class_id = argmax(score); // index maximalnog elementa liste
//                for(Float ff: score)
//                {
//                    System.out.print(ff); System.out.print(",");
//                }
//                System.out.println();
                float conf = score.get(class_id);
                if (conf >= 0.5) {
                    int center_x = (int) (detect.get(0) * img.cols());
                    int center_y = (int) (detect.get(1) * img.rows());
                    int width = (int) (detect.get(2) * img.cols());
                    int height = (int) (detect.get(3) * img.rows());
                    int x = (center_x - width / 2);
                    int y = (center_y - height / 2);
                    Rect2d box = new Rect2d(x, y, width, height);
                    result.get("boxes").add(box);
                    result.get("confidences").add(conf);
                    result.get("class_ids").add(class_id);
                }
            }
        }
        return result;
    }

    /**
     * Returns index of maximum element in the list
     */
    private static int argmax(List<Float> array) {
        float max = array.get(0);
        int re = 0;
        for (int i = 1; i < array.size(); i++) {
            if (array.get(i) > max) {
                max = array.get(i);
                re = i;
            }
        }
        return re;
    }

    private static MatOfInt getBBoxIndicesFromNonMaximumSuppression(List<Rect2d> boxes, List<Float> confidences) {
        MatOfRect2d mOfRect = new MatOfRect2d();
        mOfRect.fromList(boxes);
        MatOfFloat mfConfs = new MatOfFloat(Converters.vector_float_to_Mat(confidences));
        MatOfInt result = new MatOfInt();
        Dnn.NMSBoxes(mOfRect, mfConfs, (float) (0.6), (float) (0.5), result);
        return result;
    }

    private static Mat drawBoxesOnTheImage(
        Mat img, 
        MatOfInt indices, 
        List<Rect2d> boxes, 
        Map<Integer, String> cocoLabels,
        List<Integer> classIdList, 
        List<Scalar> colors
    ) {
        //Scalar color = new Scalar( new double[]{255, 255, 0});
        List indices_list = indices.toList();
        for (int i = 0; i < boxes.size(); i++) {
            if (indices_list.contains(i)) {
                Rect2d box = boxes.get(i);
                Point x_y = new Point(box.x, box.y);
                Point w_h = new Point(box.x + box.width, box.y + box.height);
                Point text_point = new Point(box.x, box.y - 5);
                
                Imgproc.rectangle(img, w_h, x_y, colors.get(classIdList.get(i)), 1);
                String label = cocoLabels.get(classIdList.get(i));
                Imgproc.putText(img, label, text_point, Imgproc.FONT_HERSHEY_SIMPLEX, 1, colors.get(classIdList.get(i)), 2);
            }
        }
        return img;
    }
}
