package com.tobee.opencv.detector.tests.mrcnn;

import com.tobee.opencv.detector.tests.CvUtils;
import com.tobee.opencv.detector.tests.TestUtils;
import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.List;

import org.opencv.core.Core;
import org.opencv.core.CvType;
import org.opencv.core.Mat;
import org.opencv.core.Mat.Atable;
import org.opencv.core.MatOfFloat;
import org.opencv.core.MatOfInt;
import org.opencv.core.MatOfPoint;
import org.opencv.core.MatOfRect2d;
import org.opencv.core.Point;
import org.opencv.core.Rect;
import org.opencv.core.Rect2d;
import org.opencv.core.Scalar;
import org.opencv.core.Size;
import org.opencv.core.TickMeter;
import org.opencv.dnn.Dnn;
import org.opencv.dnn.Net;
import org.opencv.highgui.HighGui;
import org.opencv.imgcodecs.Imgcodecs;
import org.opencv.imgproc.Imgproc;
import org.opencv.utils.Converters;
import org.opencv.videoio.VideoCapture;
import org.opencv.videoio.VideoWriter;
import org.opencv.videoio.Videoio;

public class TestMaskRCNN_BackupCode {
	private static final int TYPE_IMAGE = 1;
	private static final int TYPE_VIDEO = 2;
	private static final int TYPE_WEBCAM = 3;

	private static final int TYPE_INPUT = TYPE_IMAGE;

	private static final int TYPE_CPU = 1;
	private static final int TYPE_GPU = 2;

	private static final int TARGET_INPUT = TYPE_CPU;
	
	
	
	public static void mainMaskRCNN(String[] args) {
		// List<String> classStringList = getStringListOnFile(CLASSES_FILE);
		List<String> colorStringList = CvUtils.getStringListOnFile(TestUtils.COLOR_FILE);
		classes = CvUtils.getStringListOnFile(TestUtils.CLASSES_FILE);

		String[] CLASSES = classes.toArray(new String[classes.size()]);

		String[] COLORS = colorStringList.toArray(new String[colorStringList.size()]);
		colors = new ArrayList<Scalar>();

		for (String color : COLORS) {
			String[] rgb = color.split(" ");

			Double r = Double.parseDouble(rgb[0]);
			Double g = Double.parseDouble(rgb[1]);
			Double b = Double.parseDouble(rgb[2]);

			colors.add(new Scalar(r, g, b, 255));

			r = g = b = null;
			rgb = null;
		}

		// VideoCapture capture = new VideoCapture(cameraDevice);

//        if (!capture.isOpened()) {
//            System.err.println("--(!)Error opening video capture");
//            System.exit(0);
//        }
//		Mat frame = new Mat();
//        while (capture.read(frame)) {
//            if (frame.empty()) {
//                System.err.println("--(!) No captured frame -- Break!");
//                break;
//            }
//
//            //-- 3. Apply the classifier to the frame
//            detectAndDisplay(frame, faceCascade, eyesCascade);
//
//            if (HighGui.waitKey(10) == 27) {
//                break;// escape
//            }
//        }

		Mat frame = Imgcodecs.imread(TestUtils.IMAGE_SRC_FILE);
		Size size = frame.size();
		int cols = frame.cols();
		int rows = frame.rows();
		double h = size.height;
		double w = size.width;
		int hh = (int) size.height;
		int ww = (int) size.width;

		Mat blob = org.opencv.dnn.Dnn.blobFromImage(frame, 1.0, new Size(w, h), Scalar.all(0), true, false);

		// Load the network
		Net net = org.opencv.dnn.Dnn.readNetFromTensorflow(TestUtils.MODEL_WEIGHTS_MASK_RCNN, TestUtils.TEXT_GRAPH);
		net.setPreferableBackend(org.opencv.dnn.Dnn.DNN_BACKEND_OPENCV);
		net.setPreferableTarget(org.opencv.dnn.Dnn.DNN_TARGET_CPU);

		net.setInput(blob);

		List<String> outNames = new ArrayList<String>();
		List<Mat> outs = new ArrayList<Mat>();

		outNames.add("detection_out_final");
		outNames.add("detection_masks");
		net.forward(outs, outNames);

		// Extract the bounding box and mask for each of the detected objects
		postprocess(frame, outs);

		Mat numClasses = outs.get(0);
		numClasses = numClasses.reshape(1, (int) numClasses.total() / 7);

		for (int i = 0; i < numClasses.rows(); ++i) {
			double confidence = numClasses.get(i, 2)[0];
			// System.out.println(confidence);
			if (confidence > 0.2) {
				int classId = (int) numClasses.get(i, 1)[0];
				String label = CLASSES[classId] + ": " + confidence;
				System.out.println(label);
				int left = (int) (numClasses.get(i, 3)[0] * cols);
				int top = (int) (numClasses.get(i, 4)[0] * rows);
				int right = (int) (numClasses.get(i, 5)[0] * cols);
				int bottom = (int) (numClasses.get(i, 6)[0] * rows);
				System.out.println(left + " " + top + " " + right + " " + bottom);
			}
		}
		// Mat numMasks = outputMats.get(1);
	}
	
	public static void mainMaskRCNN2(String[] args) {
		//List<String> classStringList = getStringListOnFile(CLASSES_FILE);
		List<String> colorStringList = CvUtils.getStringListOnFile(TestUtils.COLOR_FILE);
		classes = CvUtils.getStringListOnFile(TestUtils.CLASSES_FILE);
		int cameraDevice = args.length > 2 ? Integer.parseInt(args[2]) : 0;
		//String[] CLASSES = classes.toArray(new String[classes.size()]);
		
		String[] COLORS = colorStringList.toArray(new String[colorStringList.size()]);
		colors = new ArrayList<Scalar>();
		
		for(String color: COLORS)
		{
			String[] rgb = color.split(" ");
			
			Double r = Double.parseDouble(rgb[0]);
			Double g = Double.parseDouble(rgb[1]);
			Double b = Double.parseDouble(rgb[2]);
			
			colors.add(new Scalar(r,g,b,255));
			
			r = g = b = null;
			rgb = null;
		}
		
		// Load the network
		Net net = org.opencv.dnn.Dnn.readNetFromTensorflow(TestUtils.MODEL_WEIGHTS_MASK_RCNN, TestUtils.TEXT_GRAPH);
		
		switch(TARGET_INPUT)
		{
		case TYPE_CPU:
			System.out.println( "Using CPU device");
			net.setPreferableTarget(org.opencv.dnn.Dnn.DNN_TARGET_CPU);
			
			break;
		case TYPE_GPU:
			System.out.println( "Using GPU device");
			net.setPreferableBackend(org.opencv.dnn.Dnn.DNN_BACKEND_CUDA);
	        net.setPreferableTarget(org.opencv.dnn.Dnn.DNN_TARGET_CUDA);
	        
			break;
			
		}

		VideoCapture cap = new VideoCapture();
		VideoWriter video = new VideoWriter();
		String outputFile = "mask_rcnn_out_cpp.avi";
		
		boolean isOpen = false;
		
		switch(TYPE_INPUT)
		{
		case TYPE_IMAGE:
			
			outputFile = TestUtils.IMAGE_SRC_FILE;
			isOpen = cap.open(outputFile);
			System.out.println( "Using image source. open?" + isOpen);
			
			outputFile = outputFile.substring(outputFile.length() - 4, outputFile.length()) + "_mask_rcnn_out.jpg";
			
			break;
		case TYPE_VIDEO:
			
			outputFile = TestUtils.VIDEO_SRC_FILE;
			isOpen = cap.open(outputFile);
			
			System.out.println( "Using video source. open? " + isOpen);
			
			outputFile = outputFile.substring(outputFile.length() - 4, outputFile.length()) + "_mask_rcnn_out.avi";
			
			break;
			
		case TYPE_WEBCAM:
			cap.open(cameraDevice);
			break;
		}
		
		if (TYPE_INPUT != TYPE_IMAGE) {
	        video.open(outputFile, VideoWriter.fourcc('M', 'J', 'P', 'G'), 28, new Size(cap.get(Videoio.CAP_PROP_FRAME_WIDTH), cap.get(Videoio.CAP_PROP_FRAME_HEIGHT)));
	    }
		
		// Create a window
	    String kWinName = "Deep learning object detection in OpenCV";
	    HighGui.namedWindow(kWinName, HighGui.WINDOW_NORMAL);
	    Mat frame = new Mat(), blob;
	    
	    while (HighGui.waitKey(1) < 0)
	    {
	        // get frame from the video
	    	cap.read(frame);
	        //cap >> frame;
	        
	        // Stop the program if reached end of video
	        if (frame.empty()) {
	            System.out.println( "Done processing !!!");
	            System.out.println( "Output file is stored as " + outputFile);
	            HighGui.waitKey(3000);
	            break;
	        }
	        // Create a 4D blob from a frame.
	        blob = org.opencv.dnn.Dnn.blobFromImage(frame, 1.0, new Size(frame.cols(), frame.rows()), Scalar.all(0), true, false);
	        
	        //Sets the input to the network
	        net.setInput(blob);

	        // Runs the forward pass to get output from the output layers
	        List<String> outNames = new ArrayList<String>(2);
	        outNames.add(0, "detection_out_final");
	        outNames.add(1, "detection_masks");
	        
	        List<Mat> outs = new ArrayList<Mat>();
	        net.forward(outs, outNames);

	        // Extract the bounding box and mask for each of the detected objects
	        postprocess(frame, outs);

	        // Put efficiency information. The function getPerfProfile returns the overall time for inference(t) and the timings for each of the layers(in layersTimes)
	        org.opencv.core.MatOfDouble layersTimes = new org.opencv.core.MatOfDouble();
	        double freq = Core.getTickFrequency() / 1000;
	        double t = net.getPerfProfile(layersTimes) / freq;
	        String label = String.format("Mask-RCNN on 2.5 GHz Intel Core i7 CPU, Inference time for a frame : %0.0f ms", t);
	        Imgproc.putText(frame, label, new Point(0, 15), Imgproc.FONT_HERSHEY_SIMPLEX, 0.5, new Scalar(0, 0, 0));

	        // Write the frame with the detection boxes
	        Mat detectedFrame = new Mat();
	        frame.convertTo(detectedFrame, CvType.CV_8U);
	        
	        if (TYPE_INPUT == TYPE_IMAGE) {
	        	Imgcodecs.imwrite(outputFile, detectedFrame);
		    }
	        else
	        {
	        	video.write(detectedFrame);
	        }
	        
	        HighGui.imshow(kWinName, frame);
	    }
	}
	
	private static void postprocess(Mat frame, List<Mat> outs) {
		// TODO Auto-generated method stub
		
	}

	private static int getMatIVal(Mat eye, int ro, int co)
	{
		if (CvType.depth(eye.type()) != CvType.CV_32S) {
            return -1;
        }
		
		int value =  -1;
		int[] values = new int[eye.channels()];
		for (int y = 0; y < eye.rows(); y++) {
		  for (int x = 0; x < eye.cols(); x++) {
		    eye.get(x, y, values);
		    if(ro == y && co == x)
		    {
		    	value = values[0];
		    	break;
		    }
		  }
		}
		
		return value;
	}
	
	private static float getMatFVal(Mat eye, int ro, int co)
	{
		System.out.println("Mat type" + eye.type());
		System.out.println("Mat type" + CvType.depth(eye.type()));
		
		if (CvType.depth(eye.type()) != CvType.CV_32F) {
            return -1;
        }
		
		float value =  -1f;
		float[] values = new float[eye.channels()];
		
		eye.get(ro, co, values);
		
		value = values[0];
		
//		for (int y = 0; y < eye.rows(); y++) {
//		  for (int x = 0; x < eye.cols(); x++) {
//		    eye.get(x, y, values);
//		    if(ro == y && co == x)
//		    {
//		    	value = values[0];
//		    	break;
//		    }
//		  }
//		}
		
		return value;
	}
	
	private static List<Scalar> colors;
	private static List<String> classes;
	
	private static void drawBox2(Mat frame, int classId, float conf, Rect box, Mat objectMask) {
		// Draw a rectangle displaying the bounding box
		Imgproc.rectangle(frame, new Point(box.x, box.y), new Point(box.x + box.width, box.y + box.height),
				new Scalar(255, 178, 50), 3);

		// Get the label for the class name and its confidence
		String label = String.format("%.2f", conf);
		if (!classes.isEmpty()) {
			// CV_Assert(classId < (int)classes.size());
			label = classes.get(classId) + ":" + label;
		}

		// Display the label at the top of the bounding box
		int baseLine[] = new int[1];
		Size labelSize = Imgproc.getTextSize(label, Imgproc.FONT_HERSHEY_SIMPLEX, 0.5, 1, baseLine);
		box.y = (int) Math.max(box.y, labelSize.height);
		Imgproc.rectangle(frame, new Point(box.x, box.y - Math.round(1.5 * labelSize.height)),
				new Point(box.x + Math.round(1.5 * labelSize.width), box.y + baseLine[0]), new Scalar(255, 255, 255),
				Imgproc.FILLED);
		Imgproc.putText(frame, label, new Point(box.x, box.y), Imgproc.FONT_HERSHEY_SIMPLEX, 0.75, new Scalar(0, 0, 0),
				1);

		Scalar color = colors.get(classId % colors.size());

		// Resize the mask, threshold, color and apply it on the image
		Imgproc.resize(objectMask, objectMask, new Size(box.width, box.height));

//	    Mat mask = (objectMask > maskThreshold);
//	    Core.compare
//	    Mat coloredRoi = (0.3 * color + 0.7 * frame.submat(box));
//	    
//	    
//	    coloredRoi.convertTo(coloredRoi, CvType.CV_8UC3);
//
//	    // Draw the contours on the image
//	    List<MatOfPoint> contours = new ArrayList<MatOfPoint>();
//		Mat hierarchy = new Mat();
//	    mask.convertTo(mask, CvType.CV_8U);
//	    Imgproc.findContours(mask, contours, hierarchy, Imgproc.RETR_CCOMP, Imgproc.CHAIN_APPROX_SIMPLE);
//	    Imgproc.drawContours(coloredRoi, contours, -1, color, 5, Imgproc.LINE_8, hierarchy, 100);
//	    coloredRoi.copyTo(frame.submat(box), mask);

	}

	public static void main(String[] args) {
		System.out.println(Core.NATIVE_LIBRARY_NAME);
		System.loadLibrary(Core.NATIVE_LIBRARY_NAME);
		//mainMaskRCNN3(args);
	}
}
