package com.tobee.opencv.detector.tests.mrcnn;

import com.tobee.opencv.detector.tests.CvUtils;
import com.tobee.opencv.detector.tests.input.DNNOption;
import com.tobee.opencv.detector.tests.input.DNNOptionTest;
import conf.CocoLabel;
import java.io.File;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import javax.swing.JOptionPane;

import org.opencv.core.Core;
import org.opencv.core.CvType;
import org.opencv.core.Mat;
import org.opencv.core.MatOfDouble;
import org.opencv.core.MatOfPoint;
import org.opencv.core.Point;
import org.opencv.core.Rect;
import org.opencv.core.Scalar;
import org.opencv.core.Size;
import org.opencv.dnn.Dnn;
import org.opencv.dnn.Net;
import org.opencv.highgui.HighGui;
import org.opencv.imgcodecs.Imgcodecs;
import org.opencv.imgproc.Imgproc;
import org.opencv.videoio.VideoCapture;
import org.opencv.videoio.VideoWriter;
import org.opencv.videoio.Videoio;

public class TestMaskRCNN_tobee2 {
   
    static{
        System.loadLibrary(Core.NATIVE_LIBRARY_NAME);
    }
    
    private static final double maskThreshold = 0.3; // Mask threshold

    public static void main(String[] args) {
        String winName;
        String mediaType;
        String deviceType;
        String mediaFile;
        String modelKind; 
        String modelConfiguration;
        double threshold;
        
        if(!DNNOptionTest.isDebug)
        {
            if(!DNNOptionTest.setUpOptions())
            {
                JOptionPane.showMessageDialog(null, "Action canceled by user", "Info", JOptionPane.INFORMATION_MESSAGE);
                return;
            } 

            if(!DNNOptionTest.isValid()) {
                JOptionPane.showMessageDialog(null, "Fill in all options", "error", JOptionPane.WARNING_MESSAGE);
                return;
            }
        }
        
        mediaType = DNNOptionTest.mediaType;
        deviceType = DNNOptionTest.deviceType;

        mediaFile = DNNOptionTest.mediaFile;
        modelKind = DNNOptionTest.modelKind; 
        modelConfiguration = DNNOptionTest.modelConfiguration;
        boolean isSaveOuput = DNNOptionTest.isSaveOutput;
        threshold = DNNOptionTest.threshold;
        
        if(DNNOptionTest.isDebug)
        {
            mediaType = "Video";
            deviceType = "cpu";

            mediaFile = "C:\\DEV\\Experimental\\Vision\\data\\videos\\vtest.avi";
            modelKind = "C:\\DEV\\Experimental\\Vision\\pretrained_model\\mrcnn\\frozen_inference_graph.pb";
            modelConfiguration = "C:\\DEV\\Experimental\\Vision\\pretrained_model\\mrcnn\\mask_rcnn_inception_v2_coco_2018_01_28.pbtxt";
            
            isSaveOuput = true;
        }   
        winName = DNNOptionTest.WinName.replace("${Model}","MRCNN");
        
        if(!new File(mediaFile).exists()) {
            JOptionPane.showMessageDialog(null, String.format("File9[%s] Not Found", mediaFile), "File Error", JOptionPane.WARNING_MESSAGE);
            return;
        }
        if(!new File(modelKind).exists() || !new File(modelConfiguration).exists()) {
            JOptionPane.showMessageDialog(null, String.format("Can not open Model[%s] \n config[%s]", 
                    modelKind,modelConfiguration), "Model Error", JOptionPane.WARNING_MESSAGE);
            return;
        }
        
        // Load the network
        Net dnnNet = Dnn.readNetFromTensorflow(modelKind, modelConfiguration);

        switch(deviceType)
        {
        case "cpu": 
            System.out.println( "Using CPU device");
            dnnNet.setPreferableTarget(org.opencv.dnn.Dnn.DNN_TARGET_CPU);
            break;
        case "gpu": 
            System.out.println( "Using GPU device");
            dnnNet.setPreferableBackend(Dnn.DNN_BACKEND_CUDA);
            dnnNet.setPreferableTarget(Dnn.DNN_TARGET_CUDA);
            break;
        }
        
        if (mediaType.equals("Image"))
        {
            System.out.println( "Using Image");
            detectObjectOnImage(winName, dnnNet, mediaFile, threshold);
        }
        else if (mediaType.equals("Video"))
        {
            System.out.println( "Using Video");
            detectObjectOnVideo(winName, dnnNet, mediaFile, isSaveOuput, threshold);
        }
    }
    
    
    private static void detectObjectOnImage(String winName, Net dnnNet, String mediaFile, double threshold) 
    {
        Mat frame = Imgcodecs.imread(mediaFile);
        String outputFile = mediaFile.substring(0, mediaFile.length() - 4) + "_mask_rcnn_out.jpg";
        System.out.println( "Using video source. open ? " + outputFile);
        
        // Stop the program if reached end of video
        if (frame.empty()) {
            System.out.println( "can not read!!!");
            System.out.println( "Output file is stored as " + outputFile);
            HighGui.waitKey(3000);
            return;
        }
        Mat blob;
	
        // Create a 4D blob from a frame.
        blob = Dnn.blobFromImage(frame, 1.0, new Size(frame.cols(), frame.rows()), Scalar.all(0), true, false);
        //Sets the input to the network
        dnnNet.setInput(blob);
        
        // Runs the forward pass to get output from the output layers
        List<String> outNames = new ArrayList<>(2);
        outNames.add(0, "detection_out_final");
        outNames.add(1, "detection_masks");
        
        List<Mat> outs = new ArrayList<>();
        dnnNet.forward(outs, outNames);
        
         // generate radnom color in order to draw bounding boxes
        List<Scalar> colors =CocoLabel.getColorsByLabel();
        Map<Integer, String> cocoLabels =  CocoLabel.getCocoLabel();
        
        // Extract the bounding box and mask for each of the detected objects
        postprocess(frame, outs, threshold, cocoLabels, colors);

        // Put efficiency information. The function getPerfProfile returns the overall time for inference(t) 
        // and the timings for each of the layers(in layersTimes)
        MatOfDouble layersTimes = new MatOfDouble();
        double freq = Core.getTickFrequency() / 1000;
        double t = dnnNet.getPerfProfile(layersTimes) / freq;
        String label = String.format("Mask-RCNN, Inference time for a frame : %f ms", t);
        Imgproc.putText(frame, label, new Point(0, 15), 
                        Imgproc.FONT_HERSHEY_SIMPLEX, 0.5, new Scalar(0, 0, 0));

        // Write the frame with the detection boxes
        Mat detectedFrame = new Mat();
        frame.convertTo(detectedFrame, CvType.CV_8U);
       
        System.out.println("-->" + outputFile);
        Imgcodecs.imwrite(outputFile, detectedFrame);
        
        blob.release();
        detectedFrame.release();
        layersTimes.release();
        layersTimes = null;
        detectedFrame = null;
        outNames.clear();
        outNames = null;
        blob = null;
        
        CvUtils.showWindow(winName, CvUtils.matToBufferedImage(CvUtils.reSizeCommon(Imgcodecs.imread(outputFile), 2)));
    }

    private static void detectObjectOnVideo(
       String winName, Net dnnNet, String mediaFile, boolean isSaveOuput,
       double threshold
    ) {
        Map<Integer, String> cocoLabels =  CocoLabel.getCocoLabel();
        // generate radnom color in order to draw bounding boxes
        List<Scalar> colors =CocoLabel.getColorsByLabel();
        HighGui.namedWindow(winName, HighGui.WINDOW_NORMAL);
        
        Mat frame = new Mat();
        VideoCapture capture = new VideoCapture(mediaFile);
        
        if (capture.open(mediaFile)) {
            capture.read(frame);
            HighGui.imshow(winName, frame);
        } else {
            System.out.println("Capture is not opened!");
            return;
        }
        
        VideoWriter vwriter = new VideoWriter();
        Mat detectedFrame = new Mat();
        String outputFile = mediaFile.substring(0, mediaFile.length() - 4) + "_yolov4_out.avi";
        
        while (capture.isOpened())
        {
            if(HighGui.waitKey(100) == 27) //ESC key to exit
            {
                System.out.println( "[ESC]Done processing !!!");
                //System.out.println( "[ESC]Output file is stored as " + outputFile);
                HighGui.waitKey(3000);
                break;
            }
            
            // get frame from the video
            capture.read(frame);
           
            // Stop the program if reached end of video
            if (frame.empty()) {
                System.out.println( "Done processing !!!");
                System.out.println( "Output file is stored as " + outputFile);
                HighGui.waitKey(3000);
                 if(vwriter.isOpened()) {
                    vwriter.release();
                } 
                vwriter = null;
                break;
            }
            // Create a 4D blob from a frame.
            Mat blob_from_image = Dnn.blobFromImage(frame, 1.0, new Size(frame.cols(), frame.rows()), Scalar.all(0), true, false);

            //Sets the input to the network
            dnnNet.setInput(blob_from_image);

            // Runs the forward pass to get output from the output layers
            List<String> outNames = new ArrayList<>(2);
            outNames.add(0, "detection_out_final");
            outNames.add(1, "detection_masks");

            List<Mat> outs = new ArrayList<>();
            dnnNet.forward(outs, outNames);

            // Extract the bounding box and mask for each of the detected objects
            postprocess(frame, outs, threshold, cocoLabels, colors);

            // Put efficiency information. The function getPerfProfile returns the overall time for inference(t) and the timings for each of the layers(in layersTimes)
            MatOfDouble layersTimes = new MatOfDouble();
            double freq = Core.getTickFrequency() / 1000;
            double t = dnnNet.getPerfProfile(layersTimes) / freq;
            String label = String.format("Mask-RCNN, Inference time for a frame : %f ms", t);
            Imgproc.putText(frame, label, new Point(0, 15), 
                            Imgproc.FONT_HERSHEY_SIMPLEX, 0.5, new Scalar(0, 0, 0));
            
            HighGui.imshow(winName, frame);
            
            if(isSaveOuput)
            {
                double fps = capture.get(Videoio.CAP_PROP_FPS);
                System.out.println("fps: " + fps);
                if(fps == 0D)
                {
                    //fps = 20;
                    continue;
                }
                
                Size size 
                    = new Size(capture.get(Videoio.CAP_PROP_FRAME_WIDTH), capture.get(Videoio.CAP_PROP_FRAME_HEIGHT));

    //            int fourcc_xvid = VideoWriter.fourcc('X', 'V', 'I', 'D');   //# XVID (AVI format)
    //            int fourcc_mp4v = VideoWriter.fourcc('M', 'P', '4', 'V');  // # MP4 format
                int fourcc_mjpg = VideoWriter.fourcc('M', 'J', 'P', 'G');  //# Motion JPEG format
    //            int fourcc_h264 = VideoWriter.fourcc('H', '2', '6', '4'); // # H264 (more advanced compression)

                isSaveOuput = vwriter.open(outputFile, fourcc_mjpg, fps, size, true);

                if(!isSaveOuput)
                {
                    JOptionPane.showMessageDialog(null, 
                        String.format("Can not open the video file[sz:%s, fps:%f]", size.toString(), fps), 
                        "error", JOptionPane.INFORMATION_MESSAGE);
                    break;
                }
                
                frame.convertTo(detectedFrame, CvType.CV_8U);
                if(vwriter.isOpened()) 
                {
                    vwriter.write(detectedFrame);
                }
                else
                {
                    System.out.println("========vwriter is not opened=======");
                }
            }
            
//            if(saveOuput)
//            {
//                // Write the frame with the detection boxes
//                Mat detectedFrame = new Mat();
//                frame.convertTo(detectedFrame, CvType.CV_8U);
//
//                vwriter.write(detectedFrame);
//                
//                detectedFrame.release();
//                detectedFrame = null;
//            }
            layersTimes.release();
            layersTimes = null;
            
            outNames.clear();
            outNames = null;
            //HighGui.destroyWindow(kWinName);
            //break;
        }
        HighGui.destroyAllWindows();
        
        //CvUtils.showWindow(winName, CvUtils.matToBufferedImage(CvUtils.reSizeCommon(Imgcodecs.imread(outputFile), 2)));       
    }
    
     // For each frame, extract the bounding box and mask for each detected object
    private static Mat postprocess(
        Mat frame, final List<Mat> outs, double threshold,
        Map<Integer,String> cocoLabels, List<Scalar> colors) 
    {
        Mat outDetections = outs.get(0);
        Mat outMasks = outs.get(1);

        // Output size of masks is NxCxHxW where
        // N - number of detected boxes
        // C - number of classes (excluding background)
        // HxW - segmentation shape
        final int numDetections = outDetections.size(2);
        //final int numClasses = outMasks.size(1);

        outDetections = outDetections.reshape(1, (int) outDetections.total() / 7);

        for (int i = 0; i < numDetections; ++i) {
            // float score = outDetections.at<float>(i, 2);
            // Atable<Float> s = outDetections.at(Float.class, i,2);
            float score = outDetections.at(Float.class, i, 2).getV();
            //score = (float)getMatVal(outDetections, i, 2);

            if (score > threshold) {
                // Extract the bounding box
                int classId = outDetections.at(Float.class, i, 1).getV().intValue();
                int left = (int)(frame.cols() * outDetections.at(Float.class, i, 3).getV());
                int top = (int)(frame.rows() * outDetections.at(Float.class, i, 4).getV());
                int right = (int)(frame.cols() * outDetections.at(Float.class, i, 5).getV());
                int bottom = (int)(frame.rows() * outDetections.at(Float.class, i, 6).getV());

                left = Math.max(0, Math.min(left, frame.cols() - 1));
                top = Math.max(0, Math.min(top, frame.rows() - 1));
                right = Math.max(0, Math.min(right, frame.cols() - 1));
                bottom = Math.max(0, Math.min(bottom, frame.rows() - 1));
                Rect box = new Rect(left, top, right - left + 1, bottom - top + 1);

                // Extract the mask for the object
                // https://stackoverflow.com/questions/50325025/opencv-ptr-in-java
                // Mat objectMask(outMasks.size[2], outMasks.size[3], CV_32F, outMasks.ptr<float>(i, classId));
                //int aa[] = new int[outDetections.channels()];
                //outDetections.get(i, 1, aa);

                //float[] values = new float[outDetections.channels()];
                //outDetections.get(i, 1, values);

                float fVal = (float)CvUtils.getMatVal(outDetections, i, 1);

                //Mat objectMask = new Mat(new int[] { outMasks.size(2), outMasks.size(3) }, CvType.CV_32F,
                //		new Scalar(values[0]));

                Mat objectMask = new Mat(outMasks.size(1), outMasks.size(1), CvType.CV_32F, new Scalar(fVal));

                // Draw bounding box, colorize and show the mask on the image
                //drawBoxWithMask(frame, classId, score, box, objectMask, cocoLabels, colors);
                drawBoxSimple(frame, classId, box, cocoLabels, colors);
                objectMask.release(); 
                objectMask = null;

            }
        }
		
		
        outDetections.release(); outDetections = null;
        outMasks.release(); outMasks = null;

        return frame;
    }

    // Draw the predicted bounding box, colorize and show the mask on the image
    private static void drawBoxWithMask(
        Mat frame, int classId, double conf, 
        Rect box, Mat objectMask, 
        Map<Integer, String> cocoLabels,
        List<Scalar> colors
    ) {
        // Draw a rectangle displaying the bounding box
        Imgproc.rectangle(frame, new Point(box.x, box.y), new Point(box.x + box.width, box.y + box.height),
                        new Scalar(255, 178, 50), 3);

        // Get the label for the class name and its confidence
        String label = String.format("%.2f", conf);

        if (!cocoLabels.isEmpty()) {
            label = cocoLabels.get(classId) + ":" + label;
        }
		
        //System.out.println("label : " + label);

        // Display the label at the top of the bounding box
        // int baseLine[] = new int[1];
        int baseLine[] = { 0 };
        Size labelSize = Imgproc.getTextSize(label, Imgproc.FONT_HERSHEY_SIMPLEX, 0.5, 1, baseLine);
        box.y = Integer.max(box.y, (int) labelSize.height);

        Imgproc.rectangle(
            frame, new Point(box.x, box.y - Math.round(1.5 * labelSize.height)),
            new Point(box.x + Math.round(1.5 * labelSize.width), box.y + baseLine[0]), new Scalar(255, 255, 255),
            Imgproc.FILLED);
        Imgproc.putText(frame, label, new Point(box.x, box.y), 
                        Imgproc.FONT_HERSHEY_SIMPLEX, 0.75, new Scalar(0, 0, 0),1);

        Scalar color = colors.get(classId % colors.size());
        // Comment the above line and uncomment the two lines below to generate
        // different instance colors
        //int colorInd = (int) (Math.random() % colors.size());
        // Scalar color = colors[colorInd];

        // Resize the mask, threshold, color and apply it on the image
        Imgproc.resize(objectMask, objectMask, new Size(box.width, box.height));

        // ~~~~~~~~~~~~~~
        

        // ~~~~~~~~~~~~~~
        Mat image_roi = new Mat(frame, box); // frame(box)
        //image_roi.release();
        // ~~~~~~~~~~~~~~
        // Mat coloredRoi = new Mat(0.3 * color + 0.7 * image_roi);

        // 0.3 * color
        Scalar s2 = new Scalar(0.3);
        Scalar s3 = color.mul(s2);

        // 0.7 * image_roi
        Mat image_roi2 = new Mat();
        Scalar s = new Scalar(0.7);
        Core.multiply(image_roi, s, image_roi2);

        Mat coloredRoi = new Mat();
        Core.add(image_roi2, s3, coloredRoi);

        // Mat coloredRoi = new Mat(0.3 * color + 0.7 * image_roi);
        coloredRoi.convertTo(coloredRoi, CvType.CV_8UC3);

        // Draw the contours on the image
        List<MatOfPoint> contours = new ArrayList<>();
        // Mat mask = (objectMask > maskThreshold);
        Mat mask = new Mat();
        objectMask.setTo(new Scalar(maskThreshold), mask);
        Mat hierarchy = new Mat();
        mask.convertTo(mask, CvType.CV_8U);
        Imgproc.findContours(mask, contours, hierarchy, 
                        Imgproc.RETR_CCOMP, Imgproc.CHAIN_APPROX_SIMPLE);
        //public static void drawContours(
        //Mat image, List<MatOfPoint> contours, 
        //int contourIdx, Scalar color, int thickness, int lineType, Mat hierarchy, int maxLevel) {

        Imgproc.drawContours(coloredRoi, contours, -1, color, 2, 
                        Imgproc.LINE_8, hierarchy, 100);
        coloredRoi.copyTo(image_roi, mask);

        mask.release(); mask = null;
        image_roi2.release(); image_roi2 = null;
        hierarchy.release(); hierarchy = null;
        coloredRoi.release(); coloredRoi = null;

        if(!contours.isEmpty())
        for(MatOfPoint p: contours)
        {
            p.release();
            p = null;
        }

        contours.clear();
        contours = null;
    }
    
    // Draw the predicted bounding box, colorize and show the mask on the image
    private static void drawBoxSimple(
        Mat frame, int classId, Rect box, 
        Map<Integer, String> cocoLabels,
        List<Scalar> colors
    ) {
        Scalar selectedColor = colors.get(classId);
        Point x_y = new Point(box.x, box.y);
        Point w_h = new Point(box.x + box.width, box.y + box.height);
        //Point text_point = new Point(box.x, box.y - 5);

        Imgproc.rectangle(frame, w_h, x_y, selectedColor, 1);
        String label = cocoLabels.get(classId);
        
        int baseLine[] = { 0 };
        Size labelSize = Imgproc.getTextSize(label, Imgproc.FONT_HERSHEY_SIMPLEX, 0.5, 1, baseLine);
        box.y = Integer.max(box.y, (int) labelSize.height);
        //public static void putText(Mat img, String text, Point org, int fontFace, double fontScale, Scalar color, int thickness)
        Imgproc.putText(frame, label, new Point(box.x, box.y), 
            Imgproc.FONT_HERSHEY_SIMPLEX, 0.75, selectedColor,1);
        
        
        //Imgproc.putText(frame, label, text_point, Imgproc.FONT_HERSHEY_SIMPLEX, 1, colors.get(classId), 2);        
       
    }
    
    

}
