package com.tobee.opencv.detector.tests.yolo;

import com.tobee.opencv.detector.tests.CvUtils;
import com.tobee.opencv.detector.tests.TestUtils;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Random;

import org.opencv.core.CvType;
import org.opencv.core.Mat;
import org.opencv.core.MatOfFloat;
import org.opencv.core.MatOfInt;
import org.opencv.core.MatOfRect2d;
import org.opencv.core.Point;
import org.opencv.core.Rect2d;
import org.opencv.core.Scalar;
import org.opencv.core.Size;
import org.opencv.dnn.Dnn;
import org.opencv.dnn.Net;
import org.opencv.highgui.HighGui;
import org.opencv.imgcodecs.Imgcodecs;
import org.opencv.imgproc.Imgproc;
import org.opencv.utils.Converters;
import org.opencv.videoio.VideoCapture;
import org.opencv.videoio.VideoWriter;
import org.opencv.videoio.Videoio;

//https://www.linkedin.com/pulse/algorithm-can-learn-neural-network-20-lines-code-svetozar-radoj%C4%8Din/
//https://www.linkedin.com/pulse/opencv-java-yolo-object-detection-images-svetozar-radoj%C4%8Din
public class  TestYolov3_linkedin{
	private static final String THIS_NAME = TestYolov3_linkedin.class.getSimpleName();
        
	public static void mainYolov3(String[] args) throws Exception {
		org.apache.commons.cli.CommandLineParser parser = new org.apache.commons.cli.DefaultParser();
		org.apache.commons.cli.HelpFormatter formatter = new org.apache.commons.cli.HelpFormatter();
		org.apache.commons.cli.CommandLine cmd = null;// not a good practice, it serves it purpose
		org.apache.commons.cli.Options options = TestUtils.setupOptions4YoloV3();
		
		try {
			cmd = parser.parse(options, args);
		} catch (org.apache.commons.cli.ParseException e) {
			System.out.println(e.getMessage());
			formatter.printHelp("Yolov3", options);

			System.exit(1);
		}

		if (cmd.hasOption("help"))
		{
			String header = "Use this script to run object detection using yolo3 model in OpenCV\n\n";
			String footer = "\nPlease report issues at http://example.com/issues";

			// HelpFormatter formatter = new HelpFormatter();
			formatter.printHelp("MaskRCNN2", header, options, footer, true);
			return;
		}
		
		// load the COCO class labels our YOLO model was trained on
		List<String> cocoLabels = null;
		
		if (cmd.hasOption("label"))
		{
                    String labelFile = cmd.getOptionValue("label");
                    cocoLabels = CvUtils.getStringListOnFile(labelFile);
		}
		
		String imageFile = "d:\\Eclipse_2019_06\\myWork\\OpenCV_Demo\\images\\dining_table.jpg";
		String videoFile = "d:\\Eclipse_2019_06\\myWork\\OpenCV_Demo\\videos\\outputs\\video_out.mp4";
		String outputFile = "";
		// Download and  load only wights for YOLO , this  is obtained from official YOLO site//
		String modelWeights = "d:\\Eclipse_2019_06\\myWork\\OpenCV_Demo\\yolo-coco\\yolov3.weights"; 
		// Download and  load cfg file for YOLO , can be obtained from official site//
		String modelConfiguration = "d:\\Eclipse_2019_06\\myWork\\OpenCV_Demo\\yolo-coco\\yolov3.cfg";
		String kWinName = "Deep learning object detection in OpenCV Yolov3";
		
		// load our video:
		VideoCapture capture = new VideoCapture();
		VideoWriter vwriter = new VideoWriter();//new VideoWriter(videoFile, VideoWriter.fourcc('m', 'p', 'g', '4'), fps, size, true);
		int cameraDevice = 0;//args.length > 2 ? Integer.parseInt(args[2]) : 0;
		double fps = capture.get(Videoio.CAP_PROP_FPS);
		boolean isOpen = false;
		
		double frmCount = 0;
		Mat frame = null;
		
		if (cmd.hasOption("image"))
		{
			imageFile  = cmd.getOptionValue("image");
			frame = Imgcodecs.imread(imageFile);
			outputFile = imageFile.substring(0, imageFile.length() - 4) + "_yolov3_out.jpg";
		}
		
		else if (cmd.hasOption("video"))
		{
			videoFile = cmd.getOptionValue("video");
			
			// Create a window
		    HighGui.namedWindow(kWinName, HighGui.WINDOW_NORMAL);
			frame = new Mat();
			isOpen = capture.open(videoFile);
			
			outputFile = videoFile.substring(0, videoFile.length() - 4) + "_yolov3_rcnn_out.avi";
			System.out.println( "Using video source. open ? " + outputFile);
			
			if (isOpen) {
				frmCount = capture.get(Videoio.CAP_PROP_FRAME_COUNT);
				System.out.println("frmCount = " + frmCount);
				// CAP_PROP_FRAME_COUNT
			} else {
				System.out.println("Capture is not opened!");
				return;
			}
			
			capture.read(frame);
			HighGui.imshow(kWinName, frame);
		}
		else
		{
			capture.open(cameraDevice);
		}
		
		if (cmd.hasOption("model"))
		{
			modelWeights  = cmd.getOptionValue("model");
		}
		
		if (cmd.hasOption("config"))
		{
			modelConfiguration  = cmd.getOptionValue("config");
		}
		
		// load our YOLO object detector trained on COCO dataset
		Net dnnNet = org.opencv.dnn.Dnn.readNetFromDarknet(modelConfiguration, modelWeights);
		
		if (cmd.hasOption("deivce"))
		{
			String device = cmd.getOptionValue("deivce");
			
			switch(device)
			{
			case "cpu": 
				System.out.println( "Using CPU device");
				dnnNet.setPreferableTarget(org.opencv.dnn.Dnn.DNN_TARGET_CPU);
				break;
			case "gpu": 
				System.out.println( "Using GPU device");
				dnnNet.setPreferableBackend(org.opencv.dnn.Dnn.DNN_BACKEND_CUDA);
				dnnNet.setPreferableTarget(org.opencv.dnn.Dnn.DNN_TARGET_CUDA);
				break;
			}
		}
		
		if (!cmd.hasOption("image"))
		{
			// # loop over frames from the video file stream
			//Mat img = new Mat();
			
			Size size 
				= new Size(capture.get(Videoio.CAP_PROP_FRAME_WIDTH), capture.get(Videoio.CAP_PROP_FRAME_HEIGHT));
			//vwriter.open(outputFile, VideoWriter.fourcc('M', 'J', 'P', 'G'), 28, size);
			vwriter.open(outputFile, VideoWriter.fourcc('M', 'J', 'P', 'G'), fps, size);
		}

		// generate radnom color in order to draw bounding boxes
		Random random = new Random();
		ArrayList<Scalar> colors = new ArrayList<Scalar>();
		for (int i = 0; i < cocoLabels.size(); i++) {
			colors.add(
				new Scalar(
					new double[] { random.nextInt(255), random.nextInt(255), random.nextInt(255) }));
		}

		List<String> layerNames = dnnNet.getLayerNames();
		List<String> outputLayers = new ArrayList<String>();
		for (Integer i : dnnNet.getUnconnectedOutLayers().toList()) {
			outputLayers.add(layerNames.get(i - 1));
		}

		while (true) {
			
                    if (!cmd.hasOption("image")) 
                    if(HighGui.waitKey(100) ==27)
                    {
                            System.out.println( "[ESC]Done processing !!!");
                        System.out.println( "[ESC]Output file is stored as " + outputFile);
                        HighGui.waitKey(3000);
                        break;
                    }

                    if (!cmd.hasOption("image"))
                    {
                            // get frame from the video
                            capture.read(frame);
                            //cap >> frame;
                    }

                    // Stop the program if reached end of video
                    if (frame.empty()) {
                        System.out.println( "Done processing !!!");
                        System.out.println( "Output file is stored as " + outputFile);
                        HighGui.waitKey(3000);
                        break;
                    }
			//
                    HashMap<String, List> result = forwardImageOverNetwork(frame, dnnNet, outputLayers);

                    ArrayList<Rect2d> boxes = (ArrayList<Rect2d>) result.get("boxes");
                    ArrayList<Float> confidences = (ArrayList<Float>) result.get("confidences");
                    ArrayList<Integer> class_ids = (ArrayList<Integer>) result.get("class_ids");

                    // -- Now , do so-called “non-maxima suppression”
                    // Non-maximum suppression is performed on the boxes whose confidence is equal
                    // to or greater than the threshold.
                    // This will reduce the number of overlapping boxes:
                    MatOfInt indices = getBBoxIndicesFromNonMaximumSuppression(boxes, confidences);
                    // -- Finally, go over indices in order to draw bounding boxes on the image:
                    frame = drawBoxesOnTheImage(frame, indices, boxes, cocoLabels, class_ids, colors);
                    //vwriter.write(frame);
                    //System.out.println("===============");
                    HighGui.imshow(kWinName, frame);
			
                    Mat detectedFrame = new Mat();
                    frame.convertTo(detectedFrame, CvType.CV_8U);

                    if (cmd.hasOption("image")) {
                        Imgcodecs.imwrite(outputFile, detectedFrame);

                        break;
                    }
                    else if (cmd.hasOption("video")) 
                    {
                            vwriter.write(detectedFrame);

                            HighGui.imshow(kWinName, frame);


                            boxes.clear(); boxes = null;
                            confidences.clear(); confidences = null;
                            class_ids.clear(); class_ids = null;
                            indices.release();indices = null;
                    }
			
			
		} // of main while()
		capture.release();
		vwriter.release();
		
		HighGui.destroyAllWindows();
		
		
		if (cmd.hasOption("image")) {
	    	CvUtils.showWindow(THIS_NAME, CvUtils.matToBufferedImage(CvUtils.reSizeCommon(Imgcodecs.imread(outputFile), 2)));
	    }   
	}
	
	public static void mainYolov3_2(String[] args) throws Exception {
		org.apache.commons.cli.CommandLineParser parser = new org.apache.commons.cli.DefaultParser();
		org.apache.commons.cli.HelpFormatter formatter = new org.apache.commons.cli.HelpFormatter();
		org.apache.commons.cli.CommandLine cmd = null;// not a good practice, it serves it purpose
		org.apache.commons.cli.Options options = TestUtils.setupOptions4YoloV3();
		
		try {
			cmd = parser.parse(options, args);
		} catch (org.apache.commons.cli.ParseException e) {
			System.out.println(e.getMessage());
			formatter.printHelp("Yolov3", options);

			System.exit(1);
		}

		if (cmd.hasOption("help"))
		{
			String header = "Use this script to run object detection using yolo3 model in OpenCV\n\n";
			String footer = "\nPlease report issues at http://example.com/issues";

			// HelpFormatter formatter = new HelpFormatter();
			formatter.printHelp("MaskRCNN2", header, options, footer, true);
			return;
		}
		
		// load the COCO class labels our YOLO model was trained on
		List<String> cocoLabels = null;
		
		if (cmd.hasOption("label"))
		{
			String labelFile = cmd.getOptionValue("label");
			cocoLabels = CvUtils.getStringListOnFile(labelFile);
		}
		
		String imageFile = "d:\\Eclipse_2019_06\\myWork\\OpenCV_Demo\\images\\dining_table.jpg";
		String outputFile = "";
		// Download and  load only wights for YOLO , this  is obtained from official YOLO site//
		String modelWeights = "d:\\Eclipse_2019_06\\myWork\\OpenCV_Demo\\yolo-coco\\yolov3.weights"; 
		// Download and  load cfg file for YOLO , can be obtained from official site//
		String modelConfiguration = "d:\\Eclipse_2019_06\\myWork\\OpenCV_Demo\\yolo-coco\\yolov3.cfg";
		
		if (cmd.hasOption("image"))
		{
			imageFile  = cmd.getOptionValue("image");
			
			outputFile = imageFile.substring(0, imageFile.length() - 4) + "_yolov3_out.jpg";
		}
		
		if (cmd.hasOption("model"))
		{
			modelWeights  = cmd.getOptionValue("model");
		}
		
		if (cmd.hasOption("config"))
		{
			modelConfiguration  = cmd.getOptionValue("config");
		}
		
		// load our YOLO object detector trained on COCO dataset
		Net dnnNet = Dnn.readNetFromDarknet(modelConfiguration, modelWeights);
		
		if (cmd.hasOption("deivce"))
		{
			String device = cmd.getOptionValue("deivce");
			
			switch(device)
			{
			case "cpu": 
				System.out.println( "Using CPU device");
				dnnNet.setPreferableTarget(org.opencv.dnn.Dnn.DNN_TARGET_CPU);
				break;
			case "gpu": 
				System.out.println( "Using GPU device");
				dnnNet.setPreferableBackend(org.opencv.dnn.Dnn.DNN_BACKEND_CUDA);
				dnnNet.setPreferableTarget(org.opencv.dnn.Dnn.DNN_TARGET_CUDA);
				break;
			}
		}
		
		
		// generate radnom color in order to draw bounding boxes
		Random random = new Random();
		ArrayList<Scalar> colors = new ArrayList<Scalar>();
		for (int i = 0; i < cocoLabels.size(); i++) {
			colors.add(new Scalar(new double[] { random.nextInt(255), random.nextInt(255), random.nextInt(255) }));
		}

		// load our input image
		Mat img = Imgcodecs.imread(imageFile,Imgcodecs.IMREAD_COLOR); 
		
		// dining_table.jpg soccer.jpg baggage_claim.jpg
		// -- determine the output layer names that we need from YOLO
		// The forward() function in OpenCV’s Net class needs the ending layer till
		// which it should run in the network.
		// getUnconnectedOutLayers() vraca indexe za: yolo_82, yolo_94, yolo_106,
		// (indexi su 82, 94 i 106) i to su poslednji layeri
		// u networku:
		List<String> layerNames = dnnNet.getLayerNames();
		List<String> outputLayers = new ArrayList<String>();
		for (Integer i : dnnNet.getUnconnectedOutLayers().toList()) {
			outputLayers.add(layerNames.get(i - 1));
		}
		HashMap<String, List> result = forwardImageOverNetwork(img, dnnNet, outputLayers);

		ArrayList<Rect2d> boxes = (ArrayList<Rect2d>) result.get("boxes");
		ArrayList<Float> confidences = (ArrayList<Float>) result.get("confidences");
		ArrayList<Integer> class_ids = (ArrayList<Integer>) result.get("class_ids");

		// -- Now , do so-called “non-maxima suppression”
		// Non-maximum suppression is performed on the boxes whose confidence is equal
		// to or greater than the threshold.
		// This will reduce the number of overlapping boxes:
		MatOfInt indices = getBBoxIndicesFromNonMaximumSuppression(boxes, confidences);
		// -- Finally, go over indices in order to draw bounding boxes on the image:
		img = drawBoxesOnTheImage(img, indices, boxes, cocoLabels, class_ids, colors);
		HighGui.imshow("Test", img);
		HighGui.waitKey(10000);
	}

	private static HashMap<String, List> forwardImageOverNetwork(Mat img, Net dnnNet, List<String> outputLayers) {
		// --We need to prepare some data structure in order to store the data returned
		// by the network
		// (ie, after Net.forward() call))
		// So, Initialize our lists of detected bounding boxes, confidences, and class
		// IDs, respectively
		// This is what this method will return:
		HashMap<String, List> result = new HashMap<String, List>();
		result.put("boxes", new ArrayList<Rect2d>());
		result.put("confidences", new ArrayList<Float>());
		result.put("class_ids", new ArrayList<Integer>());

		// -- The input image to a neural network needs to be in a certain format called
		// a blob.
		// In this process, it scales the image pixel values to a target range of 0 to 1
		// using a scale factor of 1/255.
		// It also resizes the image to the given size of (416, 416) without cropping
		// Construct a blob from the input image and then perform a forward pass of the
		// YOLO object detector,
		// giving us our bounding boxes and associated probabilities:

		// Here we supply the spatial size that the Convolutional Neural Network
		// expects.
		Mat blob_from_image = Dnn.blobFromImage(img, 1 / 255.0, new Size(416, 416),
				new Scalar(new double[] { 0.0, 0.0, 0.0 }), true, false);
		dnnNet.setInput(blob_from_image);

		// -- the output from network's forward() method will contain a List of OpenCV
		// Mat object, so lets prepare one
		List<Mat> outputs = new ArrayList<Mat>();

		// -- Finally, let pass forward throught network. The main work is done here:
		dnnNet.forward(outputs, outputLayers);

		// --Each output of the network outs (ie, each row of the Mat from 'outputs')
		// is represented by a vector of the number
		// of classes + 5 elements. The first 4 elements represent center_x, center_y,
		// width and height.
		// The fifth element represents the confidence that the bounding box encloses
		// the object.
		// The remaining elements are the confidence levels (ie object types) associated
		// with each class.
		// The box is assigned to the category corresponding to the highest score of the
		// box:

		for (Mat output : outputs) {
			// loop over each of the detections. Each row is a candidate detection,
			System.out.println("Output.rows(): " + output.rows() + ", Output.cols(): " + output.cols());
			for (int i = 0; i < output.rows(); i++) {
				Mat row = output.row(i);
				List<Float> detect = new MatOfFloat(row).toList();
				List<Float> score = detect.subList(5, output.cols());
				int class_id = argmax(score); // index maximalnog elementa liste
				float conf = score.get(class_id);
				if (conf >= 0.5) {
					int center_x = (int) (detect.get(0) * img.cols());
					int center_y = (int) (detect.get(1) * img.rows());
					int width = (int) (detect.get(2) * img.cols());
					int height = (int) (detect.get(3) * img.rows());
					int x = (center_x - width / 2);
					int y = (center_y - height / 2);
					Rect2d box = new Rect2d(x, y, width, height);
					result.get("boxes").add(box);
					result.get("confidences").add(conf);
					result.get("class_ids").add(class_id);
				}
			}
		}
		return result;
	}

	/**
	 * Returns index of maximum element in the list
	 */
	private static int argmax(List<Float> array) {
		float max = array.get(0);
		int re = 0;
		for (int i = 1; i < array.size(); i++) {
			if (array.get(i) > max) {
				max = array.get(i);
				re = i;
			}
		}
		return re;
	}

	private static MatOfInt getBBoxIndicesFromNonMaximumSuppression(ArrayList<Rect2d> boxes, ArrayList<Float> confidences) {
		MatOfRect2d mOfRect = new MatOfRect2d();
		mOfRect.fromList(boxes);
		MatOfFloat mfConfs = new MatOfFloat(Converters.vector_float_to_Mat(confidences));
		MatOfInt result = new MatOfInt();
		Dnn.NMSBoxes(mOfRect, mfConfs, (float) (0.6), (float) (0.5), result);
		return result;
	}

	private static Mat drawBoxesOnTheImage(
		Mat img, 
		MatOfInt indices, 
		ArrayList<Rect2d> boxes, 
		List<String> cocoLabels,
		ArrayList<Integer> class_ids, ArrayList<Scalar> colors
	) {
//Scalar color = new Scalar( new double[]{255, 255, 0});
		List indices_list = indices.toList();
		for (int i = 0; i < boxes.size(); i++) {
			if (indices_list.contains(i)) {
				Rect2d box = boxes.get(i);
				Point x_y = new Point(box.x, box.y);
				Point w_h = new Point(box.x + box.width, box.y + box.height);
				Point text_point = new Point(box.x, box.y - 5);
				Imgproc.rectangle(img, w_h, x_y, colors.get(class_ids.get(i)), 1);
				String label = cocoLabels.get(class_ids.get(i));
				Imgproc.putText(img, label, text_point, Imgproc.FONT_HERSHEY_SIMPLEX, 1, colors.get(class_ids.get(i)),
						2);
			}
		}
		return img;
	}
}
