package tobee.test;

import com.tobee.opencv.detector.tests.input.DNNOptionTest;
import conf.CocoLabel;
import java.io.File;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import javax.swing.JOptionPane;

import org.opencv.core.Core;
import org.opencv.core.CvType;
import org.opencv.core.Mat;
import org.opencv.core.MatOfDouble;
import org.opencv.core.MatOfFloat;
import org.opencv.core.MatOfInt;
import org.opencv.core.MatOfRect2d;
import org.opencv.core.Point;
import org.opencv.core.Rect2d;
import org.opencv.core.Scalar;
import org.opencv.core.Size;
import org.opencv.dnn.Dnn;
import org.opencv.dnn.Net;
import org.opencv.highgui.HighGui;
import org.opencv.imgcodecs.Imgcodecs;
import org.opencv.imgproc.Imgproc;
import org.opencv.utils.Converters;
import org.opencv.videoio.VideoCapture;
import org.opencv.videoio.VideoWriter;
import org.opencv.videoio.Videoio;

public class MaskRCNNTest {
   
    static{
        System.loadLibrary(Core.NATIVE_LIBRARY_NAME);
    }
    
    private static final double maskThreshold = 0.3; // Mask threshold

    public static void main(String[] args) {
        String winName;
        String mediaType;
        String deviceType;
        String mediaFile;
        String modelKind; 
        String modelConfiguration;
        double threshold;
        
        if(!DNNOptionTest.isDebug)
        {
            if(!DNNOptionTest.setUpOptions())
            {
                JOptionPane.showMessageDialog(null, "Action canceled by user", "Info", JOptionPane.INFORMATION_MESSAGE);
                return;
            } 

            if(!DNNOptionTest.isValid()) {
                JOptionPane.showMessageDialog(null, "Fill in all options", "error", JOptionPane.WARNING_MESSAGE);
                return;
            }
        }
        
        mediaType = DNNOptionTest.mediaType;
        deviceType = DNNOptionTest.deviceType;

        mediaFile = DNNOptionTest.mediaFile;
        modelKind = DNNOptionTest.modelKind; 
        modelConfiguration = DNNOptionTest.modelConfiguration;
        boolean isSaveOuput = DNNOptionTest.isSaveOutput;
        threshold = DNNOptionTest.threshold;
        
        if(DNNOptionTest.isDebug)
        {
            mediaType = "Video";
            deviceType = "cpu";
            mediaFile = "C:\\DEV\\Experimental\\Vision\\data\\videos\\vtest.avi";
            
            mediaType = "Image";
            deviceType = "cpu";
            mediaFile = "C:\\DEV\\Experimental\\Vision\\data\\images\\2516944023_d00345997d_z.jpg";
            
            modelKind = "C:\\DEV\\Experimental\\Vision\\pretrained_model\\mrcnn\\frozen_inference_graph.pb";
            modelConfiguration = "C:\\DEV\\Experimental\\Vision\\pretrained_model\\mrcnn\\mask_rcnn_inception_v2_coco_2018_01_28.pbtxt";
            
            isSaveOuput = true;
        }   
        winName = DNNOptionTest.WinName.replace("${Model}","MRCNN");
        
        if(!new File(mediaFile).exists()) {
            JOptionPane.showMessageDialog(null, String.format("File9[%s] Not Found", mediaFile), "File Error", JOptionPane.WARNING_MESSAGE);
            return;
        }
        if(!new File(modelKind).exists() || !new File(modelConfiguration).exists()) {
            JOptionPane.showMessageDialog(null, String.format("Can not open Model[%s] \n config[%s]", 
                    modelKind,modelConfiguration), "Model Error", JOptionPane.WARNING_MESSAGE);
            return;
        }
        
        // Load the network
        Net dnnNet = Dnn.readNetFromTensorflow(modelKind, modelConfiguration);

        switch(deviceType)
        {
        case "cpu": 
            System.out.println( "Using CPU device");
            dnnNet.setPreferableTarget(org.opencv.dnn.Dnn.DNN_TARGET_CPU);
            break;
        case "gpu": 
            System.out.println( "Using GPU device");
            dnnNet.setPreferableBackend(Dnn.DNN_BACKEND_CUDA);
            dnnNet.setPreferableTarget(Dnn.DNN_TARGET_CUDA);
            break;
        }
        
        List<String> layerNames = dnnNet.getLayerNames();
        
        // 출력 레이어 인덱스 가져오기
        MatOfInt outLayers = dnnNet.getUnconnectedOutLayers();
        int[] outLayersArray = outLayers.toArray();

        // 출력 레이어의 이름 가져오기
        List<String> outputLayerNames = new ArrayList<>();
        for (int i : outLayersArray) {
            outputLayerNames.add(layerNames.get(i - 1));  // 인덱스 조정 (1-based index)
        }

        // 출력 레이어 이름 출력
        System.out.println("Mask R-CNN 출력 레이어: " + outputLayerNames);
		
//        //List<String> layerNames = net.getLayerNames();
//        System.out.println("🔍 Mask R-CNN 모델의 모든 레이어 목록:");
//        for (String layer : layerNames) {
//            System.out.println(layer);
//        }
//
//        // 출력 레이어 인덱스 확인
//        //MatOfInt outLayers = net.getUnconnectedOutLayers();
//        //int[] outLayersArray = outLayers.toArray();
//        System.out.println("\n📌 감지된 출력 레이어 인덱스:");
//        for (int i : outLayersArray) {
//            System.out.println("레이어 인덱스: " + i + " | 이름: " + layerNames.get(i - 1));
//        }
        
        if (mediaType.equals("Image"))
        {
            System.out.println( "Using Image");
            detectObjectOnImage(winName, dnnNet, outputLayerNames, mediaFile, isSaveOuput, threshold);
        }
        else if (mediaType.equals("Video"))
        {
            System.out.println( "Using Video");
//            detectObjectOnVideo(winName, dnnNet, mediaFile, isSaveOuput, threshold);
        }
    }
    
    public static void detectObjectOnImage(
         final String winName, final Net dnnNet, List<String> outputLayers, final String mediaFile, 
         boolean isSaveOuput, double threshold
    ) {
        Mat frame = Imgcodecs.imread(mediaFile);
        
        int idx = new File(mediaFile).getName().split("\\.").length - 2;
        
        String outputFile 
            = new File(mediaFile).getParent() + File.separator + "out" +  File.separator + 
                new File(mediaFile).getName().split("\\.")[idx] + "_yolov_out.jpg";
        
        System.out.println(idx + ":" + outputFile);
        
        Map<Integer, String> cocoLabels =  CocoLabel.getCocoLabel();
        
        // generate radnom color in order to draw bounding boxes
        List<Scalar> colors =CocoLabel.getColorsByLabel();
        
        HashMap<String, List> result = forwardImageOverNetwork(frame, dnnNet, outputLayers, threshold);

        ArrayList<Rect2d> boxes = (ArrayList<Rect2d>) result.get("boxes");
        ArrayList<Float> confidences = (ArrayList<Float>) result.get("confidences");
        ArrayList<Integer> class_ids = (ArrayList<Integer>) result.get("class_ids");

        // -- Now , do so-called “non-maxima suppression”
        // Non-maximum suppression is performed on the boxes whose confidence is equal
        // to or greater than the threshold.
        // This will reduce the number of overlapping boxes:
        MatOfInt indices = getBBoxIndicesFromNonMaximumSuppression(boxes, confidences);
        // -- Finally, go over indices in order to draw bounding boxes on the image:
        frame = drawBoxesOnTheImage(frame, indices, boxes, cocoLabels, class_ids, colors);
        //vwriter.write(frame);
        //System.out.println("===============");
        HighGui.imshow(winName, frame);
        
        MatOfDouble layersTimes = new MatOfDouble();
        double freq = Core.getTickFrequency() / 1000;
        double t = dnnNet.getPerfProfile(layersTimes) / freq;
        String label = String.format("Mask-RCNN, Inference time for a frame : %f ms", t);
        Imgproc.putText(frame, label, new Point(0, 15), 
                        Imgproc.FONT_HERSHEY_SIMPLEX, 0.5, new Scalar(0, 0, 0));
        
        if(isSaveOuput)
        {
            // Write the frame with the detection boxes
            Mat detectedFrame = new Mat();
            frame.convertTo(detectedFrame, CvType.CV_8U);

            System.out.println("-->" + outputFile);
            Imgcodecs.imwrite(outputFile, detectedFrame);
            
            detectedFrame.release();
        }
        
        HighGui.waitKey(0);
        
        HighGui.destroyAllWindows();
    }
    
    private static HashMap<String, List> forwardImageOverNetwork(
        Mat img, Net dnnNet, List<String> outputLayers,
        double threshold
    ) {
        // --We need to prepare some data structure in order to store the data returned
        // by the network
        // (ie, after Net.forward() call))
        // So, Initialize our lists of detected bounding boxes, confidences, and class
        // IDs, respectively
        // This is what this method will return:
        HashMap<String, List> result = new HashMap<>();
        result.put("boxes", new ArrayList<Rect2d>());
        result.put("confidences", new ArrayList<Float>());
        result.put("class_ids", new ArrayList<Integer>());

        // -- The input image to a neural network needs to be in a certain format called
        // a blob.
        // In this process, it scales the image pixel values to a target range of 0 to 1
        // using a scale factor of 1/255.
        // It also resizes the image to the given size of (416, 416) without cropping
        // Construct a blob from the input image and then perform a forward pass of the
        // YOLO object detector,
        // giving us our bounding boxes and associated probabilities:

        // Here we supply the spatial size that the Convolutional Neural Network
        // expects.
        Mat blob_from_image = Dnn.blobFromImage(img, 1 / 255.0, new Size(416, 416),
                        new Scalar(new double[] { 0.0, 0.0, 0.0 }), true, false);
        dnnNet.setInput(blob_from_image);

        // -- the output from network's forward() method will contain a List of OpenCV
        // Mat object, so lets prepare one
        List<Mat> outputs = new ArrayList<>();

        // -- Finally, let pass forward throught network. The main work is done here:
        dnnNet.forward(outputs, outputLayers);

        // --Each output of the network outs (ie, each row of the Mat from 'outputs')
        // is represented by a vector of the number
        // of classes + 5 elements. The first 4 elements represent center_x, center_y,
        // width and height.
        // The fifth element represents the confidence that the bounding box encloses
        // the object.
        // The remaining elements are the confidence levels (ie object types) associated
        // with each class.
        // The box is assigned to the category corresponding to the highest score of the
        // box:
        for (Mat output : outputs) {
            // loop over each of the detections. Each row is a candidate detection,
            System.out.println("Output.rows(): " + output.rows() + ", Output.cols(): " + output.cols());
            for (int i = 0; i < output.rows(); i++) {
                Mat row = output.row(i);
                List<Float> detect = new MatOfFloat(row).toList();
                List<Float> score = detect.subList(5, output.cols());
                int class_id = argmax(score); // index maximalnog elementa liste
//                for(Float ff: score)
//                {
//                    System.out.print(ff); System.out.print(",");
//                }
//                System.out.println();
                float conf = score.get(class_id);
                if (conf >= threshold) {
                    int center_x = (int) (detect.get(0) * img.cols());
                    int center_y = (int) (detect.get(1) * img.rows());
                    int width = (int) (detect.get(2) * img.cols());
                    int height = (int) (detect.get(3) * img.rows());
                    int x = (center_x - width / 2);
                    int y = (center_y - height / 2);
                    Rect2d box = new Rect2d(x, y, width, height);
                    result.get("boxes").add(box);
                    result.get("confidences").add(conf);
                    result.get("class_ids").add(class_id);
                }
            }
        }
        return result;
    }
    
     private static MatOfInt getBBoxIndicesFromNonMaximumSuppression(List<Rect2d> boxes, List<Float> confidences) {
        MatOfRect2d mOfRect = new MatOfRect2d();
        mOfRect.fromList(boxes);
        MatOfFloat mfConfs = new MatOfFloat(Converters.vector_float_to_Mat(confidences));
        MatOfInt result = new MatOfInt();
        Dnn.NMSBoxes(mOfRect, mfConfs, (float) (0.6), (float) (0.5), result);
        return result;
    }
     
     private static Mat drawBoxesOnTheImage(
        Mat img, 
        MatOfInt indices, 
        List<Rect2d> boxes, 
        Map<Integer, String> cocoLabels,
        List<Integer> classIdList, 
        List<Scalar> colors
    ) {
        //Scalar color = new Scalar( new double[]{255, 255, 0});
        List indices_list = indices.toList();
        for (int i = 0; i < boxes.size(); i++) {
            if (indices_list.contains(i)) {
                Rect2d box = boxes.get(i);
                Point x_y = new Point(box.x, box.y);
                Point w_h = new Point(box.x + box.width, box.y + box.height);
                Point text_point = new Point(box.x, box.y - 5);
                
                Imgproc.rectangle(img, w_h, x_y, colors.get(classIdList.get(i)), 1);
                String label = cocoLabels.get(classIdList.get(i));
                Imgproc.putText(img, label, text_point, Imgproc.FONT_HERSHEY_SIMPLEX, 1, colors.get(classIdList.get(i)), 2);
            }
        }
        return img;
    }
     
     /**
     * Returns index of maximum element in the list
     */
    private static int argmax(List<Float> array) {
        float max = array.get(0);
        int re = 0;
        for (int i = 1; i < array.size(); i++) {
            if (array.get(i) > max) {
                max = array.get(i);
                re = i;
            }
        }
        return re;
    }
}
