package tobee.test;

import com.tobee.opencv.detector.tests.input.DNNOptionTest;
import java.io.File;
import java.util.ArrayList;
import java.util.List;
import javax.swing.JOptionPane;
import org.opencv.core.Core;

import org.opencv.core.MatOfInt;
import org.opencv.dnn.Dnn;
import org.opencv.dnn.Net;

//https://www.linkedin.com/pulse/algorithm-can-learn-neural-network-20-lines-code-svetozar-radoj%C4%8Din/
//https://www.linkedin.com/pulse/opencv-java-yolo-object-detection-images-svetozar-radoj%C4%8Din
public class  YoloTest{
    private static final String THIS_NAME = YoloTest.class.getSimpleName();
    static{
        System.loadLibrary(Core.NATIVE_LIBRARY_NAME);
    }
    
    public static void main(String[] args) throws Exception {
        String winName;
        String mediaType;
        String deviceType;
        String mediaFile;
        String modelKind; 
        String modelConfiguration;
        double threshold;
        if(!DNNOptionTest.isDebug)
        {
            if(!DNNOptionTest.setUpOptions())
            {
                JOptionPane.showMessageDialog(null, "Action canceled by user", "Info", JOptionPane.INFORMATION_MESSAGE);
                return;
            } 

            if(!DNNOptionTest.isValid()) {
                JOptionPane.showMessageDialog(null, "Fill in all options", "error", JOptionPane.WARNING_MESSAGE);
                return;
            }
        }
 
        mediaType = DNNOptionTest.mediaType;
        deviceType = DNNOptionTest.deviceType;

        mediaFile = DNNOptionTest.mediaFile;
        modelKind = DNNOptionTest.modelKind; 
        modelConfiguration = DNNOptionTest.modelConfiguration;
        boolean isSaveOuput = DNNOptionTest.isSaveOutput;
        threshold = DNNOptionTest.threshold;
        
        if(DNNOptionTest.isDebug)
        {
            mediaType = "Video";
            deviceType = "cpu";
            mediaFile = "C:\\DEV\\Experimental\\Vision\\data\\videos\\vtest.avi";
            
            mediaType = "Image";
            deviceType = "cpu";
            mediaFile = "C:\\DEV\\Experimental\\Vision\\data\\images\\2516944023_d00345997d_z.jpg";
//            mediaFile = "D:\\DEV\\WORK\\Vision\\data\\videos\\vtest.avi";
//            modelWeights = "D:\\DEV\\WORK\\Vision\\pretrained_model\\yolo\\v4\\yolov4.weights";
//            modelConfiguration = "D:\\DEV\\WORK\\Vision\\pretrained_model\\yolo\\v4\\yolov4.cfg";
            
//            mediaFile = "D:\\DEV\\WORK\\Vision\\data\\videos\\vtest.avi";
//            modelWeights = "D:\\DEV\\WORK\\Vision\\pretrained_model\\yolo\\v3\\yolov3.weights";
//            modelConfiguration = "D:\\DEV\\WORK\\Vision\\pretrained_model\\yolo\\v3\\yolov3.cfg";
            
            
            modelKind = "C:\\DEV\\Experimental\\Vision\\pretrained_model\\yolo\\v4\\yolov4.weights";
            modelConfiguration = "C:\\DEV\\Experimental\\Vision\\pretrained_model\\yolo\\v4\\yolov4.cfg";
            
            isSaveOuput = true;
        }      
        
        winName = DNNOptionTest.WinName.replace("${Model}","YOLO");
        
        if(!new File(mediaFile).exists()) {
            JOptionPane.showMessageDialog(null, String.format("File9[%s] Not Found", mediaFile), "File Error", JOptionPane.WARNING_MESSAGE);
            return;
        }
        
        // load our YOLO object detector trained on COCO dataset
        Net dnnNet = Dnn.readNetFromDarknet(modelConfiguration, modelKind);
        switch(deviceType)
        {
        case "cpu" -> { 
            System.out.println( "Using CPU device");
            dnnNet.setPreferableTarget(Dnn.DNN_TARGET_CPU);
            }
        case "gpu" -> { 
            System.out.println( "Using GPU device");
            dnnNet.setPreferableBackend(Dnn.DNN_BACKEND_CUDA);
            dnnNet.setPreferableTarget(Dnn.DNN_TARGET_CUDA);
            }
        }
        
        List<String> layerNames = dnnNet.getLayerNames();

        // 출력 레이어 인덱스 가져오기
        MatOfInt outLayers = dnnNet.getUnconnectedOutLayers();
        int[] outLayersArray = outLayers.toArray();

        // 출력 레이어 이름 자동 탐색
        List<String> outputLayerNames = new ArrayList<>();
        for (int i : outLayersArray) {
            outputLayerNames.add(layerNames.get(i - 1));  // OpenCV는 1-based index를 사용
        }

        // 출력 레이어 자동 출력
//        System.out.println("자동 감지된 Mask R-CNN 출력 레이어: " + outputLayerNames);
//        boolean isMaskShading = false;
//        if (mediaType.equals("Image"))
//        {
//            YoloDetector.detectObjectOnImage(winName, dnnNet, mediaFile, threshold, isSaveOuput, isMaskShading);
//        }
//        else if (mediaType.equals("Video"))
//        {
//            YoloDetector.detectObjectOnVideo(winName, dnnNet, mediaFile, threshold, isSaveOuput, isMaskShading);
//        }
    }
}
